/*
RqRDDFire - A Free FlagShip RDD for Firebird RDBMS
v.0.9.0, September 2007, Brazil
Ricardo Delamar Roque (rroque6428@terra.com.br)

License
=======
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed AS IS in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/licenses/lgpl-2.1.html
*/

#include "dataserv.fh"
#include "dbstruct.fh"
#include "error.fh"
#include "fileio.fh"
#include "rqrdd.h"

#define RDD_DRIVER_RQRDDFB 	RDD_DRIVER_NAME
#define RDD_DRIVER_DBFIDX   "DBFIDX"
#define COPYRIGHT           "Created by " + RDD_DRIVER_RQRDDFB

#define RDD_FIELDTYPES          { ;
        "@@FIELDNAME CHAR(@@FIELDLEN) DEFAULT '' NOT NULL",                ;
        "@@FIELDNAME NUMERIC(@@FIELDLEN,@@FIELDDEC) DEFAULT 0 NOT NULL",   ;
        "@@FIELDNAME TIMESTAMP DEFAULT NULL",                              ;
        "@@RECNO DOUBLE PRECISION NOT NULL",                               ;
        "@@FIELDNAME CHAR(1) DEFAULT 'F' NOT NULL",                        ;
        "@@AREAID VARCHAR(@@FIELDLEN) DEFAULT NULL"                        ; }

#define RDD_UDF_LIBNAME         "rqrddudfs"
#define RDD_UDFS                { "RQBLOBSUBSTR", "RQTRIM", "RQLOWER",        ;
                                  "RQUPPER", "RQSUBSTR", "RQDTOS", "SKIPPAD", ;
                                  "RQSTR", "RQSTRZERO", "RQPADR", "RQPADL" }

#define SIZE_COL_TABLENAME      30
#define SIZE_COL_AREAID         AREAID_FIELD_SIZE

#define TABLE_OF_FLOCKS         "RDDSYS_FLK"
#define FCK_COL_TABLENAME       "FLK_TABLE"
#define FCK_COL_AREAID          "FLK_AREAID"

#define TABLE_OF_RLOCKS         "RDDSYS_RLK"
#define RCK_COL_TABLENAME       "RLK_TABLE"
#define RCK_COL_TABLERECNO      "RLK_RECNO"
#define RCK_COL_AREAID          "RLK_AREAID"

#define TABLE_OF_OPN            "RDDSYS_OPN"
#define OPN_COL_TABLENAME       "OPN_TABLE"
#define OPN_COL_AREAID          "OPN_AREAID"
#define OPN_COL_APPRECNO        "OPN_APPRECNO"
#define OPN_COL_SHARED          "OPN_SHARED"

#define MATRIX_LR_SEPARATOR     "=>"
#define MATRIX_TABLE_SEPARATOR  ">>"
#define MATRIX_STRU_SEPARATOR   "::"
#define MATRIX_FILE_EXTENSION   ".tf"
#define MATRIX_DROPTABLE_TAG    "DROP TABLE"
#define MATRIX_DROPFIELD_TAG    "DROP FIELD"

#define FLD_DOUBLE              "27"       /* See.LangRef RDB$FIELDS */
#define FLD_INTEGER             "8"
#define FLD_SMALLINT            "7"
#define FLD_FLOAT               "10"
#define FLD_TIMESTAMP           "35"
#define FLD_CHAR                "14"
#define FLD_VARCHAR             "37"
#define FLD_INT64               "16"

#define PATH_SEPARATOR          "/"
#define EXTENSION_SEPARATOR     "."
#define FULLPATH_DAT_FILE       TempFileName("/tmp","rqrdd") + ".dat"
#define DAT_FILE_PERMISSIONS    666
#Define CRLF                    Chr(13)+Chr(10)
#define LINEFEED                Chr(10)
#define LF                      LINEFEED
#define TAB                     Chr(9)
#define DESCRIPTION_LOGICAL     "LOGICAL"
#define DESCRIPTION_NUMERIC     "NUMERIC"
#define TIMEOUT_FOPN            5
#define HALF_SECOND             0.5

#define OFFSET_DBF_VERS         0          /* offset of .dbf database type   */
#define OFFSET_1STFIELD         32         /* offset of 1st field descriptor */
#define OFFSET_RECLEN           10         /* offset of record size          */
#define OFFSET_RECCOUNT         4          /* offset of record count         */
#define OFFSET_HDRLEN           8          /* offset of header size          */
#define SIZE_HEADFIELD          32         /* size of field descr in header  */

#Cinline
#include <FSopenc.h>
#include <FSextend.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <openssl/evp.h>
#include <regex.h>
#include <unistd.h>
#include <netdb.h>
#include "align.h"
#include "rqrdd.h"

#define MAXLEN  		8192
#define MAX_LOG_EVENT_SIZE      10000

  EXEC SQL BEGIN DECLARE SECTION;
    EXEC SQL
      SET DATABASE DB1 = STATIC COMPILETIME "employee.fdb" ;
  EXEC SQL END DECLARE SECTION;
  static char _rqfbconn_event_[MAX_LOG_EVENT_SIZE];
  static int  _rqfbconn_loglevel_ = LOG_OFF;
  char        err_buf[256];

  // --------------------------------
  // Not valid in FB 1.5.0 change to: typedef struct vary VARY;
  // In FB 2.0.3 change to vary2 and VARY2 respectively
  typedef struct vary2 
  {
    short       vary_length;
    char        vary_string [1];
  } VARY2;

// - - - F U N C O E S  A U X I L I A R E S 

void rddlog( int level, char *string )
{
  if( level <= _rqfbconn_loglevel_ )
  {
    FILE *fp;
    long time(), now;
    char errfile[100];
    time( &now );
    sprintf( errfile, "/var/log/rqrddfire.log" );
    fp = fopen( errfile, "a" );
    if( fp != NULL )
    {
      fprintf( fp, "\t[%i] %s ", level, getenv("USER") );
      fprintf( fp, "%s ", ctime(&now) );
      fprintf( fp, "%s\n", string );
      fclose( fp );
    }
  }
  return;
}

char *substr( char *buff, char *s, short m, short n )
{
  short i = 0;

  if( *s && m >= 1 )
  {
    if( n < 1 ) n = 0;
    s += m - 1;
    while( *s && (i < n || n == 0) ) buff[i++] = *s++;
  }
  buff[i] = '\0';

  return buff;
}

char *matchregex( char *buff, char *str, const char *ex )
{
  int        iCt;
  regex_t    preg;
  regmatch_t rmatch[10];

  buff[0] = '\0';
  memset( &preg, 0, sizeof(preg) );
  regcomp( &preg, ex, REG_EXTENDED | (ex[0]=='^'?REG_NEWLINE:0) );
  if( regexec(&preg, str, 10, rmatch, 0) == 0 )
  {
    for( iCt = 1; iCt < 10; iCt++ )
    {
      if( rmatch[iCt].rm_so == -1 ) break;
      substr( buff, str, rmatch[iCt].rm_so + 1, rmatch[iCt].rm_eo - 
                    rmatch[iCt].rm_so );
    }
  }
  regfree( &preg );
  return buff;
}

static char *translateType( short fbtype )
{
  short dtype = fbtype & ~1;
  char *tipo = " ";

  switch( dtype )
  {
    case SQL_TEXT:
    case SQL_VARYING:
      tipo = "C";
      break;
    case SQL_LONG:
    case SQL_SHORT:
    case SQL_FLOAT:
    case SQL_DOUBLE:
	case SQL_INT64:
      tipo = "N";
      break;
    case SQL_TIMESTAMP:
      tipo = "D";
      break;
  }
  return tipo;
}
#endCinline

CLASS rqFBConn
  INSTANCE bConn            := .F.                   AS Logical
  INSTANCE sDBName          := ""                    AS Character
  INSTANCE sUsername        := ""                    AS Character
  INSTANCE sRolename        := ""                    AS Character
  HIDDEN   dbhndl           := 0                     AS INTVAR
  HIDDEN   bRDDAtivo        := .T.                   AS Logical
  HIDDEN   previousRDD      := ""                    AS Character
  PROTECT  aAliases         := {}                    AS Array
  PROTECT  sFullPathDATFile := ""                    AS Character
  PROTECT  bAck             := .T.                   AS Logical
  PROTECT  sOwner           := "RqRDDFire Free"      AS Character
  PROTECT  iGlbSkMode       := SKIP_MODE_1           AS IntVar
  PROTECT  iGlbSkChSz       := DEFAULT_SKCACHE_SIZE  AS IntVar
  PROTECT  oldScreen

PROTOTYPE PROTECT METHOD internalSelect(p1) CLASS rqFBConn AS ARRAY
PROTOTYPE METHOD executeUpdate(p1)          CLASS rqFBConn
PROTOTYPE METHOD assembleStatement(p1,p2)   CLASS rqFBConn AS Character
PROTOTYPE PROTECT METHOD writeLog(p1,p2)    CLASS rqFBConn
PROTOTYPE METHOD createTableOfFLocks()      CLASS rqFBConn AS Logical
PROTOTYPE METHOD createTableOfRLocks()      CLASS rqFBConn AS Logical
PROTOTYPE METHOD createTableOfOPN()         CLASS rqFBConn AS Logical
PROTOTYPE METHOD getLockedTables()          CLASS rqFBConn AS Array
PROTOTYPE METHOD getFieldDescription(p1)    CLASS rqFBConn AS Array
PROTOTYPE METHOD getTableDescription(p1)    CLASS rqFBConn AS Array
PROTOTYPE PROTECT METHOD dropField(p1,p2)   CLASS rqFBConn AS Logical
PROTOTYPE PROTECT METHOD fOpnFile(p1,p2)    CLASS rqFBConn AS Numeric
PROTOTYPE PROTECT METHOD fClsFile(p1)       CLASS rqFBConn
PROTOTYPE PROTECT METHOD showGreetScr()     CLASS rqFBConn
PROTOTYPE PROTECT METHOD hideGreetScr()     CLASS rqFBConn
PROTOTYPE PROTECT METHOD isAllNeededUDFsInstalled() CLASS rqFBConn AS Logical
PROTOTYPE METHOD deleteAllRecords(p1)       CLASS rqFBConn AS Logical

ACCESS dbhndl CLASS rqFBConn AS INTVAR
  Return ::dbhndl
ACCESS sDBName CLASS rqFBConn AS Character
  Return ::sDBName
ACCESS sUsername CLASS rqFBConn AS Character
  Return ::sUsername
ACCESS sRolename CLASS rqFBConn AS Character
  Return ::sRolename
ACCESS bConn CLASS rqFBConn AS Logical
  Return ::bConn

ASSIGN dbhndl( vl ) CLASS rqFBConn AS INTVAR
  dbhndl := vl
  Return dbhndl
ASSIGN bConn( vl ) CLASS rqFBConn AS Logical
  bConn := vl
  Return bConn

PROTECT METHOD savePreviousRDD() CLASS rqFBConn
  ::previousRDD := RDDSetDefault()
  RDDSetDefault(RDD_DRIVER_RQRDDFB)
  Return

PROTECT METHOD restorePreviousRDD() CLASS rqFBConn
  RDDSetDefault(::previousRDD)
  Return

METHOD isConnected() CLASS rqFBConn AS Logical
  Return ::bConn

METHOD isAck() CLASS rqFBConn AS Logical
  Return ::bAck

METHOD isRDDActive() CLASS rqFBConn AS Logical
  Return ::bRDDAtivo

METHOD isFireBirdArea( als ) CLASS rqFBConn AS Logical
  Local bRet   := .F.
  Local oDBSvr := DBObject( als ) AS DataServer
  IF Upper(oDBSvr:RDDName) == RDD_DRIVER_RQRDDFB
    bRet := .T.
  Endif 
  Return bRet

METHOD setLogLevel( level ) CLASS rqFBConn
#Cinline
{
  _rqfbconn_loglevel_ = VAR_INT(VAR_NAME_LOCPAR(level));

  // --- Log
  sprintf( _rqfbconn_event_,"[LOG#] rqFBConn loglevel now is: %i", 
           _rqfbconn_loglevel_ );
  rddlog( LOG_OFF,_rqfbconn_event_ );
}
#endCinline
  Return

METHOD setFullPathDATFile( fullpath ) CLASS rqFBConn
  ::sFullPathDATFile := IIF(!Empty(fullpath),fullpath,FULLPATH_DAT_FILE)
  Return ::sFullPathDATFile

METHOD setGlbSkipMode( vl ) CLASS rqFBConn
  IF vl == NIL .or. vl < 0 .or. vl > SKIP_MODE_2
    ::iGlbSkMode := SKIP_MODE_1
  Else
    ::iGlbSkMode := vl
  Endif
  // --- Log
  ::writeLog( "[SET] Global SkipMode now is: " + Alltrim(Str(::iGlbSkMode)) ;
              + " for: " + ::sDBName, LOG_LEVEL_06 )
  Return

METHOD setGlbSkipCacheSize( sz ) CLASS rqFBConn
  IF Empty(sz) .or. sz < 0 
    ::iGlbSkChSz := DEFAULT_SKCACHE_SIZE
  Else
    ::iGlbSkChSz := sz   
  Endif
  // --- Log
  ::writeLog( "[SET] Global SkipCacheSize now is: " ;
              + Alltrim(Str(::iGlbSkChSz)) + " for: " + ::sDBName, ;
              LOG_LEVEL_06 )
  Return

METHOD getFullPathDATFile() CLASS rqFBConn
  Return ::sFullPathDATFile

METHOD getGlbSkipMode() CLASS rqFBConn AS IntVar
  Return ::iGlbSkMode

METHOD getGlbSkipCacheSize() CLASS rqFBConn AS IntVar
  Return ::iGlbSkChSz

METHOD getLogLevel() CLASS rqFBConn AS IntVar
  Local iRet := 0
#Cinline
{
  SET_VAR_INT(VAR_NAME_LOCAL(iret), _rqfbconn_loglevel_);
}
#endCinline
  Return iRet

METHOD getDatabaseName() CLASS rqFBConn AS Character
  Return ::sDBName

METHOD setRDDActive( bol ) CLASS rqFBConn
  ::bRDDAtivo := bol
  Return

PROTECT METHOD setIBRDD() CLASS rqFBConn AS Logical
  Local bRet := .F.

  IF ::isConnected()
    RDDSetDefault( RDD_DRIVER_RQRDDFB )
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- INIT
METHOD Init(db, user, pass, role) CLASS rqFBConn

  IF Empty(db) .and. Empty(user) .and. Empty(pass) .and. Empty(role)
    ::setRDDActive( .F. )
  Else
    ::showGreetScr()
    ::sDBName   := db
    ::sUsername := user
    ::sRolename := role
#Cinline
{
  EXEC SQL BEGIN DECLARE SECTION;
    char *dbname   = VAR_CHR(VAR_NAME_LOCPAR(db));
    char *username = VAR_CHR(VAR_NAME_LOCPAR(user));
    char *passwd   = VAR_CHR(VAR_NAME_LOCPAR(pass));
    char *rolename = VAR_CHR(VAR_NAME_LOCPAR(role));
  EXEC SQL END DECLARE SECTION;

  EXEC SQL 
    CONNECT :dbname AS DB1 USER :username PASSWORD :passwd ROLE :rolename;

  if( SQLCODE == 0 )
  {
    OBJ_ASSEXEC(VAR_NAME_LOCPAR(self), "bconn", -1, SET_VAR_LOG(VAR_NEW, 1));
    OBJ_ASSEXEC(VAR_NAME_LOCPAR(self), "dbhndl", -1, SET_VAR_INT(VAR_NEW, 
                (long)DB1));

    // --- Log
    sprintf( _rqfbconn_event_, "[CONN] Connected to: %s Pointer DB1: %p(%d)",
             dbname, &DB1, DB1 );
    rddlog( LOG_OFF,_rqfbconn_event_ );

    DB1 = 0L;
  }
  else 
  {
    isc_sql_interprete( (short)SQLCODE, err_buf, sizeof(err_buf) );

    // --- Log
    sprintf( _rqfbconn_event_, "[ERROR] SQLCODE=%d (%s)", SQLCODE, err_buf );
    rddlog( LOG_OFF, _rqfbconn_event_ );
  }
}
#endCinline
    ::setFullPathDATFile()
    IF ::isConnected() 
      IF !::isAllNeededUDFsInstalled()

        // --- Log
        ::writeLog( "[FATAL] Missing UDFs in Database.", LOG_OFF )

        raiseError( "The DataBase:; " + ::sDBName + "; Does not have all " + ;
                    "needed UDFs installed!!" )
      Endif
      IF !::createTableOfOPN() .or. !::createTableOfFLocks() .or. ;
         !::createTableOfRLocks()

        // --- Log
        ::writeLog( "[FATAL] Error creating system tables.", LOG_OFF )

        raiseError( "Fatal error on creating system tables. Better to Abort!" )
      Endif
    Endif
    ::hideGreetScr()
  Endif
  Return

METHOD closeAllTables() CLASS rqFBConn
  Local aAux := {}
  Local iCt

  // - - - Force close on All Unclosed Tables
  aAux := AClone(::aAliases)
  For iCt := 1 To Len(aAux)
    IF !Empty(aAux[iCt,1])

      // --- Log
      ::writeLog( "[FORCED] Trying close alias: " + aAux[iCt,1], LOG_LEVEL_05 )

      Close (aAux[iCt,1])
    Endif
  Next
  Return

// ------------------------------------------------------- disconnect
METHOD disconnect() CLASS rqFBConn
  Local Db   := ::sDBName
  Local aAux := {}
  Local iCt

  ::closeAllTables()

#Cinline
{
  char          *dbname   = VAR_CHR(VAR_NAME_LOCAL(db));
  isc_db_handle DB1       = 0L;

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl", 
                               -1));
  if( DB1 > 0 )
  {
    EXEC SQL DISCONNECT DB1;

    // --- Log
    sprintf( _rqfbconn_event_, "[EXIT] Disconnected from: %s", dbname );
    rddlog( LOG_OFF,_rqfbconn_event_ );
  }
  OBJ_ASSEXEC(VAR_NAME_LOCPAR(self), "bconn", -1, SET_VAR_LOG(VAR_NEW, 0));
}
#endCinline
  Return 

PROTECT METHOD execRegex( str, expr ) CLASS rqFBConn
  Local sRet := ""
#Cinline
{
  char buff[1024];
  char *data    = VAR_CHR(VAR_NAME_LOCPAR(str));
  char *regexpr = VAR_CHR(VAR_NAME_LOCPAR(expr));
  matchregex( buff, data, regexpr );
  if( *buff )
  {
    SET_VAR_CHR(VAR_NAME_LOCAL(sret), buff);
  }
}
#endCinline
  Return sRet

METHOD isCursesOn() CLASS rqFBConn
  Local bRet := .T.
  Local iCursesOn AS Int
#Cinline
  icurseson = (FSint) stdscr;
#endCinline
  IF iCursesOn == 0
    bRet := .F.
  Endif
  Return bRet

PROTECT METHOD showGreetScr() CLASS rqFBConn
  IF ::isCursesOn()
    ::oldScreen := SaveScreen( 0, 0, MaxRow(), MaxCol() )
    // --- VFS - Add NoPixel option
    @ MaxRow()-4, MaxCol()-50 Say "      *** Connecting to Firebird ***       ";
      Color "BG/N"  // --- NoPixel
    @ MaxRow()-3, MaxCol()-50 Say "                                           ";
      Color "BG/B"
    @ MaxRow()-2, MaxCol()-50 Say Padl( ::sOwner, 42, " " ) + " ";
      Color "BG/B"
    @ MaxRow()-1, MaxCol()-50 Say "                                  by Roque ";
      Color "BG/B"
    @ MaxRow()-2, MaxCol()-47 Say "Please wait...";
      Color "GR/B"
  Endif
  Return

PROTECT METHOD hideGreetScr() CLASS rqFBConn
  IF ::isCursesOn()
    RestScreen( 0, 0, MaxRow(), MaxCol(), ::oldScreen )
  Endif
  Return

// ------------------------------------------------------- announceAlias
METHOD announceAlias( als ) CLASS rqFBConn
  Aadd(::aAliases, {als})
  Return

// ------------------------------------------------------- removeAlias
METHOD removeAlias( als ) CLASS rqFBConn
  Local iCt 
  Local iAux := Ascan(::aAliases, {|x| x[1] == als})
  Local aAux := {}
  IF iAux > 0
    ::aAliases[iAux,1] := NIL
  Endif
  For iCt := 1 To Len(::aAliases)
    IF !Empty(::aAliases[iCt,1])
      Aadd(aAux, {::aAliases[iCt,1]})
    Endif
  Next
  ::aAliases := aAux 
  Return

// ------------------------------------------------------- axit
METHOD Axit() CLASS rqFBConn
  // - - - Log
  writeLog( "[GONE] Class: rqFBConn no longer exists. DB: " + ::sDBName,;
            LOG_OFF )

  IF ::isConnected()
    ::disconnect()
  Endif
  Return .T.

METHOD NoMethod(meth) CLASS rqFBConn
  // - - - Log
  ::writeLog( "[ERROR] Meth: " + meth + " does not exist inside rqFBConn", ;
              LOG_OFF )

  raiseError( "rqFBConn has no Method named: " + meth )
  Return

// --------------------------------------------------- getInternalTableStruct
PROTECT METHOD getInternalTableStruct( tbl ) CLASS rqFBConn
  Local aRet    := {}
  Local aSelect := ::internalSelect( ;
      "select rf.rdb$field_name, fd.rdb$field_type, fd.rdb$field_sub_type, " ;
    + "fd.rdb$field_length, fd.rdb$field_scale, rf.rdb$null_flag, " ;
    + "fd.rdb$field_precision, rf.rdb$field_position " ;
    + "from rdb$relation_fields rf, rdb$fields fd " ;
    + "where fd.rdb$field_name = rf.rdb$field_source and " ;
    + "rf.rdb$relation_name = '" + Upper(tbl) + "' " ;
    + "order by rf.rdb$field_position" )

  // aSelect
  // [1] Nome do Campo   [2] Tipo        [3] SubTipo      [4] Tamanho
  // [5] Escala          [6] Flag Null   [7] Precisao     [8] Posicao 

  IF Len(aSelect) >= 2
    aRet := ARRAY( Len(aSelect) - 1 )
    aRet := ACopy( aSelect, aRet, 2 )
  Endif
  Return aRet

// ------------------------------------------------------- getTableStruct
// --- ATTENTION!! Structure adapted for DBFIDX manipulation
METHOD getTableStruct( tbl ) CLASS rqFBConn
  Local aStru     := {}
  Local aRealStru := ::getInternalTableStruct( tbl )
  Local sFieldName, sFieldType, sPrecision
  Local iFieldLen, iFieldDec
  Local sFieldDescr
  Local iCt

  For iCt = 1 To Len(aRealStru)
    sFieldName  := Alltrim(Upper(aRealStru[iCt,1]))
    sFieldType  := aRealStru[iCt,2]
    iFieldLen   := Val(aRealStru[iCt,4])
    iFieldDec   := Abs(Val(aRealStru[iCt,5]))
    sPrecision  := aRealStru[iCt,7]
    sFieldDescr := Alltrim((::getFieldDescription(tbl, sFieldName))[1])
    IF sFieldType == FLD_DOUBLE   .or. sFieldType == FLD_INTEGER .or. ;
       sFieldType == FLD_SMALLINT .or. sFieldType == FLD_FLOAT .or. ;
	   sFieldType == FLD_INT64
      sFieldType := "N"                     // Numeric
      IF At(DESCRIPTION_NUMERIC, sFieldDescr) > 0
        iFieldLen := Val(::execRegex(sFieldDescr, ".*\((.*),.*") )
      ElseIF sPrecision == "NULL"
        iFieldLen := 15
      Else
        iFieldLen := Val(sPrecision)
      Endif
    ElseIF sFieldType == FLD_TIMESTAMP
      sFieldType := "D"                     // Date
    ElseIF sFieldType == FLD_CHAR .or. sFieldType == FLD_VARCHAR
      Do Case
        Case At(DESCRIPTION_LOGICAL, sFieldDescr) > 0
          sFieldType := "L"                 // Logical
        Otherwise
          sFieldType := "C"                 // Character
      EndCase
    Else
      sFieldType := "U"                     // Undefined
    Endif
    Aadd( aStru, {sFieldName, sFieldType, iFieldLen, iFieldDec} )
  Next
  Return aStru

// ----------------------------------------------------- executeUDFSkipPad
METHOD executeUDFSkipPad( col, tbl, recno ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect( ;
                     "select skippad(" + col + ") from " + tbl ;
                   + " where recno = " + recno ;
                   + " order by recno" )
  IF Len(aSelect) >= 2
    aRet := aSelect[2]
  Endif
  Return aRet

// ------------------------------------------------------- getRow
METHOD getRow( tbl, cond, order, [plan], sk ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect(                                 ;
                     "select first 1 skip " + Alltrim(Str(sk))       ;
                   + " * from " + tbl + " where " + cond             ;
                   + IIF(!Empty(plan),plan,"")                       ;
                   + " order by " + order )
  IF Len(aSelect) >= 2
    aRet := aSelect[2]
  Endif
  Return aRet

// ------------------------------------------------------- getFirstRows
METHOD getFirstRows( tbl, cond, order, [plan], qti, sk ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect(                                        ;
                     "select first " + Alltrim(Str(qti)) + " skip "         ;
                   + Alltrim(Str(sk)) + " * from " + tbl + " where " + cond ;
                   + IIF(!Empty(plan),plan,"")                              ;
                   + " order by " + order )
  IF Len(aSelect) >= 2
    aRet := aSelect
  Endif
  Return aRet

// ------------------------------------------------------- isUDF
METHOD isUDF( udfname, modname ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$functions " ;
                   + "where rdb$function_name = '" + udfname + "' and " ;
                   + "rdb$module_name = '" + modname + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

PROTECT METHOD isAllNeededUDFsInstalled() CLASS rqFBConn AS Logical
  Local bRet := .T.
  Local iCt
  For iCt := 1 To Len(RDD_UDFS)
    IF !::isUDF(Upper(RDD_UDFS[iCt]), RDD_UDF_LIBNAME)
      bRet := .F.
      Exit
    Endif
  Next
  Return bRet

// ------------------------------------------------------- isView
METHOD isView( vwname ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local viewname := Upper(vwname)
  Local aSelect := ::internalSelect( ;
                      "SELECT COUNT(*) FROM RDB$VIEW_RELATIONS " ;
                    + "WHERE RDB$VIEW_NAME = '" + viewname + "' " ) 
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet
  
// ------------------------------------------------------- isTable
METHOD isTable( tbl ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$relations " ;
                   + "where rdb$relation_name = '" + Upper(tbl) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isFieldOfTable
METHOD isFieldOfTable( tbl, fldname ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "SELECT COUNT(*) FROM RDB$RELATION_FIELDS "        ;
                   + "WHERE RDB$FIELD_NAME = '" + Upper(fldname) + "' AND " ;
                   + "RDB$RELATION_NAME = '" + Upper(tbl) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isFieldInStru
PROTECT METHOD isFieldInStru( fldname, tblstru ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local iAux := AScan(tblstru, {|x| x[DBS_NAME] == Alltrim(Upper(fldname))})
  IF iAux > 0
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isLimboField
METHOD isLimboField( fldname ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF Left(fldname, Len(RDDLIMBO_PREFIX)) == RDDLIMBO_PREFIX
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- filterRDDFields
PROTECT METHOD filterRDDFields( tblstru ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local iCt
  Local sFieldName
  For iCt := 1 To Len(tblstru)
    sFieldName := tblstru[iCt,DBS_NAME]
    // --- But leave Limbo Fields around
    Do Case 
      Case sFieldName == RECNO_FIELD
      Case sFieldName == AREAID_FIELD
      Case Left(sFieldName,Len(RDDIDX_PREFIX))     == RDDIDX_PREFIX
      Case Left(sFieldName,Len(RDDIDX_OLD_PREFIX)) == RDDIDX_OLD_PREFIX
      Otherwise
        Aadd(aRet, {sFieldName, tblstru[iCt,DBS_TYPE], tblstru[iCt,DBS_LEN], ;
                    tblstru[iCt,DBS_DEC]})
    EndCase
  Next
  Return aRet

// ------------------------------------------------------- filterNormalFields
PROTECT METHOD filterNormalFields( tblstru ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local iCt
  Local sFieldName
  For iCt := 1 To Len(tblstru)
    sFieldName := tblstru[iCt,DBS_NAME]
    Do Case 
      Case sFieldName == RECNO_FIELD   ;
      .or. sFieldName == AREAID_FIELD  ;
      .or. Left(sFieldName,Len(RDDLIMBO_PREFIX))   == RDDLIMBO_PREFIX   ;
      .or. Left(sFieldName,Len(RDDIDX_PREFIX))     == RDDIDX_PREFIX     ;
      .or. Left(sFieldName,Len(RDDIDX_OLD_PREFIX)) == RDDIDX_OLD_PREFIX
        Aadd(aRet, {sFieldName, tblstru[iCt,DBS_TYPE], tblstru[iCt,DBS_LEN], ;
                    tblstru[iCt,DBS_DEC]})
    EndCase
  Next
  Return aRet

// ------------------------------------------------------- isRQRDDTable
METHOD isRQRDDTable( tbl ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF Alltrim((::getTableDescription(tbl))[1]) == COPYRIGHT
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isIndex
METHOD isIndex( ind ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$indices " ;
                   + "where rdb$index_name = '" + Upper(ind) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isTrigger
METHOD isTrigger( trig ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$triggers " ;
                   + "where rdb$trigger_name = '" + Upper(trig) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isIndexOfTable
METHOD isIndexOfTable( tbl, ind ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$indices " ;
                   + "where rdb$relation_name = '" + Upper(tbl) + "' and " ;
                   + "rdb$index_name = '" + Upper(ind) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- isTriggerOfTable
METHOD isTriggerOfTable( tbl, trig ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local aSelect := ::internalSelect( ;
                     "select count(*) from rdb$triggers " ;
                   + "where rdb$relation_name = '" + Upper(tbl) + "' and " ;
                   + "rdb$trigger_name = '" + Upper(trig) + "' " )
  IF Len(aSelect) >= 2 .and. Val(aSelect[2,1]) >= 1
    bRet := .T.
  Endif
  Return bRet

// ------------------------------------------------------- countTableRecords
METHOD countTableRecords( tbl ) CLASS rqFBConn AS IntVar
  Local iRet    := 0
  Local aSelect := {}
  Local sStmt   := ""
  Local aParam  := {}
 
  sStmt := ;
    "SELECT COUNT(*) FROM @@TABLENAME"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2 
    iRet := Val(aSelect[2,1])
  Endif
  Return iRet

// --------------------------------------------------- getTableNumberOfInserts
METHOD getTableNumberOfInserts( tbl ) CLASS rqFBConn
  Local iRet    := 0
  Local aSelect := {}
  Local sStmt   := ""
  Local aParam  := {}
 
  sStmt := ;
     "SELECT GEN_ID(G_@@TABLENAME_I, 0) FROM RDB$DATABASE "

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2 
    iRet := Val(aSelect[2,1])
  Endif
  Return iRet

// ------------------------------------------------------- getTableRecCount
METHOD getTableRecCount( tbl ) CLASS rqFBConn
  Local iRet    := 0
  Local aSelect := {}
  Local sStmt   := ""
  Local aParam  := {}
 
  sStmt := ;
     "SELECT GEN_ID(G_@@TABLENAME_I, 0) - GEN_ID(G_@@TABLENAME_D, 0) " ;
   + "FROM RDB$DATABASE "

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2 
    iRet := Val(aSelect[2,1])
  Else
    iRet := ::countTableRecords( tbl )   // Hard count fallback

    // --- Log
    ::writeLog( "[FORCED] Hard Count. Table: " + tbl, LOG_LEVEL_05 )
  Endif
  Return iRet

// ------------------------------------------------------- getIndexFields
METHOD getIndexFields( ind ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect( ;
                     "select rdb$field_name from rdb$index_segments " ;
                   + "where rdb$index_name = '" + Upper(ind) + "' " ;
                   + "order by rdb$field_position" )
  IF Len(aSelect) >= 2 
    aRet := ARRAY( Len(aSelect) - 1 )
    aRet := ACopy( aSelect, aRet, 2 )
  Endif
  Return aRet

// ------------------------------------------------------- getPID
PROTECT METHOD getPID() CLASS rqFBConn
  Local iRet := 0
#Cinline
{
  SET_VAR_INT(VAR_NAME_LOCAL(iret), getpid());
}
#endCinline
  Return iRet

// ------------------------------------------------------- getIPAddress
PROTECT METHOD getIPAddress() CLASS rqFBConn
  Local sRet := ""
#Cinline
{
  char   name[50];
  char   ip[30];
  struct hostent *host;

  if( gethostname(name, sizeof(name)) == 0 )
  { 
    host = gethostbyname( name );
    if( host != NULL )
    { 
      inet_ntop( host->h_addrtype, host->h_addr_list[0], ip, sizeof(ip) );
      SET_VAR_CHR(VAR_NAME_LOCAL(sret), ip);
    }
  }
}
#endCinline
  Return sRet

// ------------------------------------------------------- getHostname
PROTECT METHOD getHostname() CLASS rqFBConn
  Local sRet := ""
#Cinline
{
  char   name[50];
  struct hostent *host;

  if( gethostname(name, sizeof(name)) == 0 )
  { 
    host = gethostbyname( name );
    if( host != NULL )
    { 
      SET_VAR_CHR(VAR_NAME_LOCAL(sret), host->h_name);
    }
  }
}
#endCinline
  Return sRet

// --------------------------------------------------- getRLockAreaID
PROTECT METHOD getRLockAreaID( tbl, recno ) CLASS rqFBConn AS Character
  Local sRet     := ""
  Local sStmt    := ""
  Local aParam   := {}
  Local aSelect  := {}

  sStmt := ;
    "SELECT @@RCK_COL_AREAID FROM @@TABLE_OF_RLOCKS WHERE " ;
  + "@@RCK_COL_TABLENAME = '@@TABLENAME' AND @@RCK_COL_TABLERECNO = @@RECNO"

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_RLOCKS",    TABLE_OF_RLOCKS})
  Aadd(aParam, {"@@RCK_COL_AREAID",     RCK_COL_AREAID})
  Aadd(aParam, {"@@RCK_COL_TABLENAME",  RCK_COL_TABLENAME})
  Aadd(aParam, {"@@RCK_COL_TABLERECNO", RCK_COL_TABLERECNO})
  Aadd(aParam, {"@@TABLENAME",          Upper(tbl)})
  Aadd(aParam, {"@@RECNO",              recno})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    sRet := Trim(aSelect[2,1])
  Endif
  Return sRet

// --------------------------------------------------- getFLockAreaID
METHOD getFLockAreaID( tbl ) CLASS rqFBConn AS Character
  Local sRet     := ""
  Local sStmt    := ""
  Local aParam   := {}
  Local aSelect  := {}

  sStmt := ;
    "SELECT @@FCK_COL_AREAID FROM @@TABLE_OF_FLOCKS " ;
  + "WHERE @@FCK_COL_TABLENAME = '@@TABLENAME' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_FLOCKS",   TABLE_OF_FLOCKS})
  Aadd(aParam, {"@@FCK_COL_TABLENAME", FCK_COL_TABLENAME})
  Aadd(aParam, {"@@FCK_COL_AREAID",    FCK_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    sRet := Trim(aSelect[2,1])
  Endif
  Return sRet

// ------------------------------------------------------- getUniqueAreaID
METHOD getUniqueAreaID( tbl, area ) CLASS rqFBConn AS Character
  Local sRet := tbl + ":" + area + ":" + Alltrim(Str(::getPID())) + ":" ;
                + Alltrim(::getHostname())
  // --- Log
  ::writeLog( "[AREAID] AreaID requested: " + sRet, LOG_LEVEL_08 )

  Return Left(sRet, SIZE_COL_AREAID)

// ------------------------------------------------------- getAPPRecnoFromArea
METHOD getAPPRecnoFromArea( tbl, areaid ) CLASS rqFBConn
  Local nRet    := 0
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}

  sStmt := ;
    "SELECT OPN_APPRECNO FROM @@TABLE_OF_OPN " ;
  + "WHERE @@OPN_COL_TABLENAME = '@@TABLENAME' AND @@OPN_COL_AREAID = " ;
  + "'@@AREAID' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
  Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
  Aadd(aParam, {"@@OPN_COL_AREAID",    OPN_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  Aadd(aParam, {"@@AREAID",            areaid})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    nRet := Val(aSelect[2,1])
    nRet := Val(Alltrim(Str(nRet, RECNO_FIELD_SIZE, 0)))
  Endif
  Return nRet

// ------------------------------------------------------- setTableDescription
METHOD setTableDescription( tbl, descr ) CLASS rqFBConn
  Local sAux := ::executeUpdate( ;
                  "update rdb$relations set rdb$description = '" ;
                + StrTran(descr,"'","''") + "' " ;
                + "where rdb$relation_name = '" + Upper(tbl) + "' " )
  Return sAux

// ------------------------------------------------------- setIndexDescription
METHOD setIndexDescription( ind, descr ) CLASS rqFBConn
  Local sAux := ::executeUpdate( ;
                  "update rdb$indices set rdb$description = '" ;
                + StrTran(descr,"'","''") + "' " ;
                + "where rdb$index_name = '" + Upper(ind) + "' " )
  Return sAux

// ------------------------------------------------------- setFieldDescription
METHOD setFieldDescription( tbl, fldname, descr ) CLASS rqFBConn
  Local sAux := ::executeUpdate( ;
                  "update rdb$relation_fields set rdb$description = '" ;
                + StrTran(descr,"'","''") + "' " ;
                + "where rdb$field_name = '" + Upper(fldname) + "' and " ;
                + "      rdb$relation_name = '" + Upper(tbl) + "' " )
  Return sAux

// --------------------------------------------------- isTableLockedByUser
METHOD isTableLockedByUser( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sLockId := ::getFLockAreaID( tbl )
  IF sLockId == areaid
    bRet := .T.
  Endif
  Return bRet

// --------------------------------------------------- isRecordLocked
METHOD isRecordLocked( tbl, areaid, recno ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF !Empty(::getRLockAreaID(tbl, recno))
    bRet := .T.
  Endif
  Return bRet 

// --------------------------------------------------- isRecordLockedByUser
METHOD isRecordLockedByUser( tbl, areaid, recno ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local sLockId := ::getRLockAreaID( tbl, recno )
  IF sLockId == areaid
    bRet := .T.
  Endif
  Return bRet

// ----------------------------------------------- isRecordLockedByAnotherUser
METHOD isRecordLockedByAnotherUser( tbl, areaid, recno ) CLASS rqFBConn            AS Logical

  Local bRet    := .F.
  Local sLockId := ::getRLockAreaID( tbl, recno )
  IF !Empty(sLockId) .and. sLockId != areaid
    bRet := .T.
  Endif
  Return bRet

// ----------------------------------------------------- isTableInUse
PROTECT METHOD isTableInUse( tbl ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}

  sStmt := ;
    "SELECT FIRST 1 @@OPN_COL_TABLENAME FROM @@TABLE_OF_OPN " ;
  + "WHERE @@OPN_COL_TABLENAME = '@@TABLENAME'"

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
  Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    bRet := .T.   
  Endif
  Return bRet

// ----------------------------------------------------- isTableInExclusiveMode
PROTECT METHOD isTableInExclusiveMode( tbl ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}

  sStmt := ;
    "SELECT FIRST 1 @@OPN_COL_TABLENAME FROM @@TABLE_OF_OPN " ;
  + "WHERE @@OPN_COL_TABLENAME = '@@TABLENAME' AND @@OPN_COL_SHARED = " ;
  + "'@@SHARED' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
  Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
  Aadd(aParam, {"@@OPN_COL_SHARED",    OPN_COL_SHARED})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  Aadd(aParam, {"@@SHARED",            "F"})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    bRet := .T.   
  Endif
  Return bRet

// ------------------------------------------------------- isTableUsedbyArea
PROTECT METHOD isTableUsedbyArea( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}

  sStmt := ;
    "SELECT COUNT(*) FROM @@TABLE_OF_OPN " ;
  + "WHERE @@OPN_COL_TABLENAME = '@@TABLENAME' AND @@OPN_COL_AREAID = " ;
  + "'@@AREAID' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
  Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
  Aadd(aParam, {"@@OPN_COL_AREAID",    OPN_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  Aadd(aParam, {"@@AREAID",            areaid})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    IF Val(aSelect[2,1]) > 0
      bRet := .T.   
    Endif
  Endif
  Return bRet

// ------------------------------------------------------- existsUserRLocks
PROTECT METHOD existsUserRLocks( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}

  sStmt := ;
    "SELECT COUNT(*) FROM @@TABLE_OF_RLOCKS " ;
  + "WHERE @@RCK_COL_TABLENAME = '@@TABLENAME' AND @@RCK_COL_AREAID = " ;
  + "'@@AREAID' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_RLOCKS",   TABLE_OF_RLOCKS})
  Aadd(aParam, {"@@RCK_COL_TABLENAME", RCK_COL_TABLENAME})
  Aadd(aParam, {"@@RCK_COL_AREAID",    RCK_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  Aadd(aParam, {"@@AREAID",            areaid})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    IF Val(aSelect[2,1]) > 0
      bRet := .T.   
    Endif
  Endif
  Return bRet

// --------------------------------------------------- existsAnotherUserRLocks
PROTECT METHOD existsAnotherUserRLocks( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}
  Local sAux

  sStmt := ;
    "SELECT COUNT(*) FROM @@TABLE_OF_RLOCKS WHERE @@RCK_COL_TABLENAME = " ;
  + "'@@TABLENAME' AND @@RCK_COL_AREAID <> '' AND @@RCK_COL_AREAID <> " ;
  + "'@@AREAID' "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_RLOCKS",   TABLE_OF_RLOCKS})
  Aadd(aParam, {"@@RCK_COL_TABLENAME", RCK_COL_TABLENAME})
  Aadd(aParam, {"@@RCK_COL_AREAID",    RCK_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
  Aadd(aParam, {"@@AREAID",            areaid})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    IF Val(aSelect[2,1]) > 0
      bRet := .T. 
    Endif
  Endif
  Return bRet

// --------------------------------------------------- existsAnotherUserFLock
PROTECT METHOD existsAnotherUserFLock( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bRet    := .F.
  Local sLockId := ::getFLockAreaID( tbl )
  IF !Empty(sLockId) .and. sLockId != areaid
    bRet := .T.
  Endif
  Return bRet

// -------------------------------------------------------- setRddFields
METHOD setRddFields( astruct ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local bLimboFields := .F.
  Local iCt
  Local iRddFieldType
  Local sFieldName, sFieldType
  For iCt := 1 To Len(astruct)
    sFieldName := astruct[iCt,DBS_NAME]
    sFieldType := astruct[iCt,DBS_TYPE]
    IF ::isLimboField( sFieldName )
      bLimboFields := .T.
    Endif
    Do Case
      Case sFieldName == RECNO_FIELD .and. sFieldType == "N"
        iRddFieldType := RDDFIELDTYPE_RECNO
      Case sFieldName == AREAID_FIELD .and. sFieldType == "C"
        iRddFieldType := RDDFIELDTYPE_AREAID
      Case Left(sFieldName,Len(RDDIDX_PREFIX)) == RDDIDX_PREFIX
        iRddFieldType := RDDFIELDTYPE_RDDIDX
      Otherwise
        iRddFieldType := RDDFIELDTYPE_NORMAL
    Endcase
    Aadd(aRet, {sFieldName, sFieldType, astruct[iCt,DBS_LEN], ;
                astruct[iCt,DBS_DEC], iRddFieldType} )
  Next
  Return {aRet, bLimboFields}

// -------------------------------------------------------- setTableUsedbyArea
METHOD setTableUsedbyArea( bol, tbl, areaid, shared ) CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local bOk     := .T.
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux

  IF bol
    sStmt := ;
      "INSERT INTO @@TABLE_OF_OPN (@@OPN_COL_TABLENAME, @@OPN_COL_AREAID, " ;
    + "@@OPN_COL_SHARED) VALUES('@@TABLENAME', '@@AREAID', '@@SHARED') " ;

    IF ::isTableUsedbyArea( tbl, areaid )
      bOk := .F.
    Endif
  Else
    sStmt := ;
      "DELETE FROM @@TABLE_OF_OPN WHERE @@OPN_COL_TABLENAME = " ;
    + "'@@TABLENAME' AND @@OPN_COL_AREAID = '@@AREAID' "

    IF !::isTableUsedbyArea( tbl, areaid )
      bOk := .F.
    Endif
  Endif

  IF bOk
    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
    Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
    Aadd(aParam, {"@@OPN_COL_AREAID",    OPN_COL_AREAID})
    Aadd(aParam, {"@@OPN_COL_SHARED",    OPN_COL_SHARED})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    Aadd(aParam, {"@@AREAID",            areaid})
    Aadd(aParam, {"@@SHARED",            IIF(shared,"T","F")})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bRet := .F.
    Endif
  Endif
  Return bRet

// ------------------------------------------------------- setRecordAsLocked
METHOD setRecordAsLocked( tbl, areaid, recno ) CLASS rqFBConn AS Logical
  Local bOk         := .T.
  Local bLockRecord := .T.
  Local sStmt       := ""
  Local aParam      := {}
  Local sAux

  // - - - C H E C K  P O S S I B I L T Y  O F  L O C K I N G
  IF ::isRecordLockedByAnotherUser( tbl, areaid, recno ) .or. ;
     ::existsAnotherUserFLock( tbl, areaid )
    bOk := .F.
  Endif
  IF bOk
    IF ::isRecordLockedByUser( tbl, areaid, recno )
      bLockRecord := .F.      // Ignore new locking operation
    Endif
  Endif

  // - - - L O C K  R E C O R D
  IF bOk .and. bLockRecord
    sStmt := ;
      "INSERT INTO @@TABLE_OF_RLOCKS (@@RCK_COL_TABLENAME, " ;
    + "@@RCK_COL_TABLERECNO, @@RCK_COL_AREAID) VALUES('@@TABLENAME', " ;
    + "@@RECNO, '@@AREAID') " ;

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_RLOCKS",    TABLE_OF_RLOCKS})
    Aadd(aParam, {"@@RCK_COL_TABLENAME",  RCK_COL_TABLENAME})
    Aadd(aParam, {"@@RCK_COL_TABLERECNO", RCK_COL_TABLERECNO})
    Aadd(aParam, {"@@RCK_COL_AREAID",     RCK_COL_AREAID})
    Aadd(aParam, {"@@TABLENAME",          Upper(tbl)})
    Aadd(aParam, {"@@RECNO",              recno})
    Aadd(aParam, {"@@AREAID",             areaid})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- setRecordAsUnLocked
METHOD setRecordAsUnLocked( tbl, areaid, recno ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local sStmt  := ""
  Local aParam := {}
  Local sAux

  // - - - C H E C K  P O S S I B I L T Y  O F  U N L O C K I N G
  IF !::isRecordLockedByUser( tbl, areaid, recno )
    bOk := .F.
  Endif

  // - - - U N L O C K  R E C O R D
  IF bOk
    sStmt := ;
      "DELETE FROM @@TABLE_OF_RLOCKS WHERE @@RCK_COL_TABLENAME = " ;
    + "'@@TABLENAME' AND @@RCK_COL_TABLERECNO = @@RECNO "

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_RLOCKS",    TABLE_OF_RLOCKS})
    Aadd(aParam, {"@@RCK_COL_TABLENAME",  RCK_COL_TABLENAME})
    Aadd(aParam, {"@@RCK_COL_TABLERECNO", RCK_COL_TABLERECNO})
    Aadd(aParam, {"@@TABLENAME",          Upper(tbl)})
    Aadd(aParam, {"@@RECNO",              recno})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- unlockAllRecords
METHOD unlockAllRecords( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local sStmt  := ""
  Local aParam := {}
  Local sAux

  // - - - C H E C K  P O S S I B I L I T Y  O F  G L O B A L  U N L O C K I N G
  IF !::existsUserRLocks( tbl, areaid )
    bOk := .F.
  Endif

  // - - - G L O B A L  U N L O C K
  IF bOk
    sStmt := ;
      "DELETE FROM @@TABLE_OF_RLOCKS WHERE @@RCK_COL_TABLENAME = " ;
    + "'@@TABLENAME' AND @@RCK_COL_AREAID = '@@AREAID' "

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_RLOCKS",   TABLE_OF_RLOCKS})
    Aadd(aParam, {"@@RCK_COL_TABLENAME", RCK_COL_TABLENAME})
    Aadd(aParam, {"@@RCK_COL_AREAID",    RCK_COL_AREAID})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    Aadd(aParam, {"@@AREAID",            areaid})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- setTableAsLocked
METHOD setTableAsLocked( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bOk        := .T.
  Local bLockTable := .T.
  Local sStmt      := ""
  Local aParam     := {}
  Local sAux

  // - - - C H E C K  P O S S I B I L I T Y  O F  T A B L E  L O C K I N G
  IF ::existsAnotherUserRLocks( tbl, areaid ) .or. ;
     ::existsAnotherUserFLock( tbl, areaid )
    bOk := .F.
  Endif
  IF bOk
    IF ::isTableLockedByUser( tbl, areaid )
      bLockTable := .F.       // Ignore new locking operation
    Endif
  Endif

  // - - - L O C K  T A B L E
  IF bOk .and. bLockTable
    sStmt := ;
      "INSERT INTO @@TABLE_OF_FLOCKS (@@FCK_COL_TABLENAME, @@FCK_COL_AREAID) " ;
    + "VALUES('@@TABLENAME', '@@AREAID')"

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_FLOCKS",   TABLE_OF_FLOCKS})
    Aadd(aParam, {"@@FCK_COL_TABLENAME", FCK_COL_TABLENAME})
    Aadd(aParam, {"@@FCK_COL_AREAID",    FCK_COL_AREAID})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    Aadd(aParam, {"@@AREAID",            areaid})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- setTableAsUnLocked
METHOD setTableAsUnLocked( tbl, areaid ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local sStmt  := ""
  Local aParam := {}
  Local sAux

  // - - - C H E C K  P O S S I B I L I T Y  O F  T A B L E  U N L O C K I N G
  IF !::isTableLockedByUser( tbl, areaid )
    bOk := .F.
  Endif
    
  // - - - U N L O C K  T A B L E
  IF bOk
    sStmt := ;
      "DELETE FROM @@TABLE_OF_FLOCKS WHERE @@FCK_COL_TABLENAME = " ;
    + "'@@TABLENAME' AND @@FCK_COL_AREAID = '@@AREAID'"

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_FLOCKS",   TABLE_OF_FLOCKS})
    Aadd(aParam, {"@@FCK_COL_TABLENAME", FCK_COL_TABLENAME})
    Aadd(aParam, {"@@FCK_COL_AREAID",    FCK_COL_AREAID})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    Aadd(aParam, {"@@AREAID",            areaid})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- releaseLocks
METHOD releaseLocks( tbl ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local sStmt  := ""
  Local aParam := {}
  Local sAux

  // - - - G L O B A L  R E L E A S E  O F  R L O C K S
  IF bOk
    sStmt := ;
      "DELETE FROM @@TABLE_OF_RLOCKS WHERE @@RCK_COL_TABLENAME = '@@TABLENAME'"

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_RLOCKS",   TABLE_OF_RLOCKS})
    Aadd(aParam, {"@@RCK_COL_TABLENAME", RCK_COL_TABLENAME})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif

  // - - - G L O B A L  R E L E A S E  O F  F L O C K S
  IF bOk
    sStmt := ;
      "DELETE FROM @@TABLE_OF_FLOCKS WHERE @@FCK_COL_TABLENAME = '@@TABLENAME'"

    aParam := {}
    Aadd(aParam, {"@@TABLE_OF_FLOCKS",   TABLE_OF_FLOCKS})
    Aadd(aParam, {"@@FCK_COL_TABLENAME", FCK_COL_TABLENAME})
    Aadd(aParam, {"@@TABLENAME",         Upper(tbl)})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
    Endif
  Endif
  Return bOk

// ------------------------------------------------------- getLockedRecords
METHOD getLockedRecords( tbl, areaid ) CLASS rqFBConn AS Array
  Local aRet    := {}
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}
  Local sAux
  Local iCt

  sStmt := ;
    "SELECT @@RCK_COL_TABLERECNO FROM @@TABLE_OF_RLOCKS " ;
  + "WHERE @@RCK_COL_TABLENAME = '@@TABLENAME' AND @@RCK_COL_AREAID = " ;
  + "'@@AREAID'"

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_RLOCKS",    TABLE_OF_RLOCKS})
  Aadd(aParam, {"@@RCK_COL_TABLERECNO", RCK_COL_TABLERECNO})
  Aadd(aParam, {"@@RCK_COL_TABLENAME",  RCK_COL_TABLENAME})
  Aadd(aParam, {"@@RCK_COL_AREAID",     RCK_COL_AREAID})
  Aadd(aParam, {"@@TABLENAME",          Upper(tbl)})
  Aadd(aParam, {"@@AREAID",             areaid})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    aRet := ARRAY( Len(aSelect) - 1 ) 
    For iCt := 2 To Len(aSelect)
      aRet[iCt-1] := Val(aSelect[iCt,1])
    Next
  Endif
  Return aRet

// ------------------------------------------------------- getLockedTables
METHOD getLockedTables() CLASS rqFBConn AS Array
  Local aRet    := {}
  Local sStmt   := ""
  Local aParam  := {}
  Local aSelect := {}
  Local sAux
  Local iCt

  sStmt := ;
    "SELECT @@FCK_COL_TABLENAME, @@FCK_COL_AREAID FROM @@TABLE_OF_FLOCKS "

  aParam := {}
  Aadd(aParam, {"@@TABLE_OF_FLOCKS",    TABLE_OF_FLOCKS})
  Aadd(aParam, {"@@FCK_COL_TABLENAME",  FCK_COL_TABLENAME})
  Aadd(aParam, {"@@FCK_COL_AREAID",     FCK_COL_AREAID})
  aSelect := ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    aRet := ARRAY( Len(aSelect) - 1 ) 
    aRet := ACopy( aSelect, aRet, 2 )
  Endif
  Return aRet

// ------------------------------------------------------- getTableDescription
METHOD getTableDescription( tbl ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect( ;
                     "select rqBlobSubstr(rdb$description,1,255) " ;
                   + "from rdb$relations " ;
                   + "where rdb$relation_name = '" + Upper(tbl) + "' " )
  IF Len(aSelect) >= 2 
    aRet := aSelect[2]
  Else 
    aRet := { "" }
  Endif
  Return aRet

// ------------------------------------------------------- getIndexDescription
METHOD getIndexDescription( ind ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect( ;
                     "select rqBlobSubstr(rdb$description,1,255) " ;
                   + "from rdb$indices " ;
                   + "where rdb$index_name = '" + Upper(ind) + "' " )
  IF Len(aSelect) >= 2 
    aRet := aSelect[2]
  Else 
    aRet := { "" }
  Endif
  Return aRet

// ------------------------------------------------------- getFieldDescription
METHOD getFieldDescription( tbl, fldname ) CLASS rqFBConn AS Array
  Local aRet := {}
  Local aSelect := ::internalSelect( ;
                     "select rqBlobSubstr(rdb$description,1,255) " ;
                   + "from rdb$relation_fields " ;
                   + "where rdb$field_name = '" + Upper(fldname) + "' and " ;
                   + "rdb$relation_name = '" + Upper(tbl) + "' " )
  IF Len(aSelect) >= 2 
    aRet := aSelect[2]
  Else 
    aRet := { "" }
  Endif
  Return aRet

// ------------------------------------------------------- getTablesName
METHOD getTablesName() CLASS rqFBConn AS Array
  Local aRet := {}
  Local iCt
  Local aSelect := ::internalSelect( ;
                     "select rdb$relation_name from rdb$relations " ;
                   + "where rdb$system_flag = 0 order by rdb$relation_name" )
  IF Len(aSelect) >= 2
    aRet := ARRAY( Len(aSelect) - 1 )
    For iCt := 2 To Len(aSelect)
      aRet[iCt-1] := Alltrim(aSelect[iCt,1])
    Next
  Endif
  Return aRet

// ------------------------------------------------------- executeQuery
METHOD executeQuery( stmt ) CLASS rqFBConn AS Array
  Return ::internalSelect( stmt )

// ------------------------------------------------------- executeUpdate
METHOD executeUpdate( stmt ) CLASS rqFBConn
  Local sStmt  := stmt
  Local sRet   := ""

  // --- Log
  ::writeLog( "[UPDT] Statement: " + sStmt, LOG_LEVEL_07 )

#Cinline
{
  char          *stmt_str = VAR_CHR(VAR_NAME_LOCAL(sstmt));
  isc_db_handle DB1 = 0L;

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl",
                               -1));

  if( DB1 > 0)
  {
    EXEC SQL WHENEVER SQLERROR GO TO Error;
    EXEC SQL SET TRANSACTION WAIT ISOLATION LEVEL READ COMMITTED;
    EXEC SQL EXECUTE IMMEDIATE :stmt_str;
    EXEC SQL COMMIT;

Error:

    EXEC SQL WHENEVER SQLERROR CONTINUE;
    if( SQLCODE != 0 )
    {
      isc_sql_interprete( (short)SQLCODE, err_buf, sizeof(err_buf) );

      // --- Log
      sprintf( _rqfbconn_event_, "[ERROR] SQLCODE=%d (%s)", SQLCODE, err_buf );
      rddlog( LOG_LEVEL_01, _rqfbconn_event_ );

      SET_VAR_CHR(VAR_NAME_LOCAL(sret), _rqfbconn_event_);

      EXEC SQL ROLLBACK;
    }
  }
}
#endCinline
  Return sRet

// ------------------------------------------------------- cachedUpdate
METHOD cachedUpdate( stmt ) CLASS rqFBConn
  Local sStmt  := stmt
  Local sRet   := ""
#Cinline
{
  char          *stmt_str = VAR_CHR(VAR_NAME_LOCAL(sstmt));
  isc_db_handle DB1 = 0L;

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl",
                               -1));
  // --- Log
  sprintf( _rqfbconn_event_, "[CACHE] Statement: %s", stmt_str );
  rddlog( LOG_LEVEL_07, _rqfbconn_event_ );

  if( DB1 > 0)
  {
    EXEC SQL WHENEVER SQLERROR GO TO Error;
    EXEC SQL EXECUTE IMMEDIATE :stmt_str;

Error:

    EXEC SQL WHENEVER SQLERROR CONTINUE;
    if( SQLCODE != 0 )
    {
      isc_sql_interprete( (short)SQLCODE, err_buf, sizeof(err_buf) );

      // --- Log
      sprintf( _rqfbconn_event_, "[ERROR] SQLCODE=%d (%s)", SQLCODE, err_buf );
      rddlog( LOG_LEVEL_01, _rqfbconn_event_ );

      SET_VAR_CHR(VAR_NAME_LOCAL(sret), _rqfbconn_event_);
    }
  }
}
#endCinline
  Return sRet

// ------------------------------------------------------- startTransaction
METHOD startTransaction() CLASS rqFBConn
#Cinline
{
  isc_db_handle DB1 = 0L;

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl",
                               -1));
  // --- Log
  sprintf( _rqfbconn_event_, "[TRANS] Starting Transaction" );
  rddlog( LOG_LEVEL_07, _rqfbconn_event_ );

  if( DB1 > 0)
  {
    EXEC SQL SET TRANSACTION;
  }
}
#endCinline
  Return

// ------------------------------------------------------- commitChanges
METHOD commitChanges() CLASS rqFBConn
#Cinline
{
  isc_db_handle DB1 = 0L;

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl",
                               -1));
  // --- Log
  sprintf( _rqfbconn_event_, "[COMMIT] Commit" );
  rddlog( LOG_LEVEL_07, _rqfbconn_event_ );

  if( DB1 > 0)
  {
    EXEC SQL COMMIT;
  }
}
#endCinline
  Return

// ------------------------------------------------------- internalSelect
PROTECT METHOD internalSelect( stmt ) CLASS rqFBConn AS ARRAY
  Local aRange := {}
  Local sStmt  := stmt

  // --- Log
  ::writeLog( "[QURY] Statement: " + sStmt, LOG_LEVEL_08 )

#Cinline
{
  int     ict;
  char    *aux = "";
  short   nullind = 0;
  FSvar   *auxPtr, *param[2];
  FSvar   *struPtr, *columnPtr, *elemPtr;
  char    *sel_str = VAR_CHR(VAR_NAME_LOCAL(sstmt));
  XSQLDA  *out_sqlda;
  XSQLVAR *var;
  long    buffer[MAXLEN];
  short   num_cols, i;
  short   length, alignment, offset;

  isc_db_handle DB1 = 0L;

  UDF_PROT(aadd);
  VAR_NEW_ARGS(param, 2);

  DB1 = (isc_db_handle)VAR_INT(OBJ_ACCEXEC(VAR_NAME_LOCPAR(self), "dbhndl", 
                               -1));

  if( DB1 > 0 )
  {
    EXEC SQL WHENEVER SQLERROR GO TO Error;
    EXEC SQL SET TRANSACTION READ ONLY;

    out_sqlda = (XSQLDA *) malloc(XSQLDA_LENGTH(1));
    out_sqlda->sqln = 1;
    out_sqlda->version = SQLDA_VERSION1;

    EXEC SQL PREPARE sql_stmt FROM :sel_str;
    EXEC SQL DESCRIBE OUTPUT sql_stmt INTO SQL DESCRIPTOR out_sqlda;
    if( out_sqlda->sqld > out_sqlda->sqln )
    {
      int n = out_sqlda->sqld;
      free( out_sqlda );
      out_sqlda = (XSQLDA *) malloc(XSQLDA_LENGTH(n));
      out_sqlda->sqln = n;
      out_sqlda->version = SQLDA_VERSION1;
      EXEC SQL DESCRIBE OUTPUT sql_stmt INTO SQL DESCRIPTOR out_sqlda;
    }

    columnPtr = VAR_NEW_ARRAY(columnPtr, out_sqlda->sqld);
    for( ict = 0, var = out_sqlda->sqlvar; ict < out_sqlda->sqld; ict++, var++ )
    {
      struPtr = VAR_NEW_ARRAY(struPtr, 4);
      if( *var->sqlname ) {
                            SET_VAR_CHR(VAR_ARRELEM(struPtr,1), var->sqlname);
                          }
      else                {
                            SET_VAR_CHR(VAR_ARRELEM(struPtr,1), var->aliasname);
                          }

      SET_VAR_CHR(VAR_ARRELEM(struPtr,2), translateType(var->sqltype));
      SET_VAR_INT(VAR_ARRELEM(struPtr,3), var->sqllen);
      SET_VAR_INT(VAR_ARRELEM(struPtr,4), -var->sqlscale);
      SET_VAR_COPY(VAR_ARRELEM(columnPtr,ict + 1), struPtr);
    }

    param[0] = VAR_NAME_LOCAL(arange);
    param[1] = columnPtr;
    UDF_EXEC(aadd)(2,param);

    if( _rqfbconn_loglevel_ >= LOG_LEVEL_10 )
      for( ict = 0; ict < out_sqlda->sqld; ict++ )
      {
        // --- Log
        sprintf( _rqfbconn_event_, "[INFO] Column: %s Type: %d Length: %d "
                                   "Alias: %s", 
                                   out_sqlda->sqlvar[ict].sqlname, 
                                   out_sqlda->sqlvar[ict].sqltype, 
                                   out_sqlda->sqlvar[ict].sqllen,
                                   out_sqlda->sqlvar[ict].aliasname );
        rddlog( LOG_LEVEL_10, _rqfbconn_event_ );
      }

    /*
     *	 Set up SQLDA.
     */
    num_cols = out_sqlda->sqld;
    for( var = out_sqlda->sqlvar, offset = 0, ict = 0; ict < num_cols; 
         var++, ict++ )
    {
      short dtype = var->sqltype & ~1;
      length = alignment = var->sqllen;
      
      if (dtype == SQL_TEXT)
      	alignment = 1;
      else if (dtype == SQL_VARYING)
      {   
      	length += sizeof (short) + 1;
      	alignment = sizeof (short);
      }   
      offset = FB_ALIGN(offset, alignment);
      var->sqldata = (char*) buffer + offset;
      offset += length;
      offset = FB_ALIGN(offset, sizeof(short));
      var->sqlind = (short*) ((char*) buffer + offset);

      offset += sizeof(short);
    }

    EXEC SQL DECLARE DYN_CURSOR CURSOR FOR sql_stmt;
    EXEC SQL OPEN DYN_CURSOR;

    while( SQLCODE == 0 )
    { 
      EXEC SQL FETCH DYN_CURSOR USING SQL DESCRIPTOR out_sqlda;
      if( SQLCODE == 100 ) break;

      elemPtr = VAR_NEW_ARRAY(auxPtr, out_sqlda->sqld);
      for( ict = 0, var = out_sqlda->sqlvar; ict < out_sqlda->sqld; 
           ict++, var++ )
      {
        VARY2           *vary;
        short sqlind  = *(short *) var->sqlind;
        short sqllen  = var->sqllen;
        char *sqldata = var->sqldata;
        short dscale  = var->sqlscale;
        short dtype   = var->sqltype & ~1;
        struct tm times;
        char data[MAXLEN], *aux;
        
        aux = data;

        if ((var->sqltype & 1) && (sqlind < 0))
        {
          sprintf( aux, "NULL" );
        }
        else
        {
          switch( dtype )
          {
            case SQL_TEXT:
              sprintf( aux, "%.*s", sqllen, sqldata );
              break;
            case SQL_VARYING:
              vary = (VARY2 *)sqldata;
              vary->vary_string[vary->vary_length] = '\0';
              sprintf( aux, "%-*s", sqllen, vary->vary_string );
              break;
            case SQL_LONG:
			case SQL_INT64:
              if( dscale < 0 )
              {
                short ict;
                int   tens = 1;
                long  value = *(long *) sqldata;
                for( ict = 0; ict > dscale; ict-- ) tens *= 10;
                sprintf( aux, "%d.%0*d", value / tens, -dscale, value % tens );
              }
              else sprintf( aux, "%i", *(long *) sqldata );
              break;
            case SQL_SHORT:
              sprintf( aux, "%i", *(short *) sqldata );
              break;
            case SQL_FLOAT:
              sprintf( aux, "%g", *(float *) sqldata );
              break;
            case SQL_DOUBLE:
              sprintf( aux, "%f", *(double *) sqldata );
              break;
            case SQL_TIMESTAMP:
              isc_decode_timestamp((ISC_TIMESTAMP *)sqldata, &times);
              sprintf( aux, "%04d-%02d-%02d %02d:%02d:%02d.%04d",
                       times.tm_year + 1900,
                       times.tm_mon+1,
                       times.tm_mday,
                       times.tm_hour,
                       times.tm_min,
                       times.tm_sec,
                       ((ISC_TIMESTAMP *)sqldata)->timestamp_time % 10000 );
              break;
          }
        }
        SET_VAR_CHR(VAR_ARRELEM(elemPtr,ict + 1), aux);
      }
      param[0] = VAR_NAME_LOCAL(arange);
      param[1] = elemPtr;
      UDF_EXEC(aadd)(2,param);
    }

    EXEC SQL CLOSE DYN_CURSOR;
    EXEC SQL COMMIT;

Error:

    EXEC SQL WHENEVER SQLERROR CONTINUE;
    if( SQLCODE != 0 )
    {
      isc_sql_interprete( (short)SQLCODE, err_buf, sizeof(err_buf) );

      // --- Log
      sprintf( _rqfbconn_event_, "[ERROR] SQLCODE=%d (%s)", SQLCODE, err_buf );
      rddlog( LOG_LEVEL_01, _rqfbconn_event_ );

      EXEC SQL ROLLBACK;
    }

    free( out_sqlda );
  }
}
#endCinline
  Return aRange

// ----------------------------------------------------- loopTest
METHOD loopTest( tbl, cond, order, cols ) CLASS rqFBConn AS Array
  Local aRet := {}

  raiseError( tbl )

  Return aRet

// ----------------------------------------------------- assembleStatement
METHOD assembleStatement( stmt, param ) CLASS rqFBConn AS Character
  Local iCt
  For iCt := 1 To Len(param)
    stmt := StrTran(stmt, param[iCt,1], param[iCt,2])
  Next
  Return stmt   

// --------------------------------------------------- getFieldsOnTrigger
PROTECT METHOD getFieldsOnTrigger( trigname ) CLASS rqFBConn AS Array
  Local aRet    := {}
  Local aSelect := {}
  Local aParam  := {}
  Local sStmt   := ""
  
  sStmt := ;
     "SELECT RDB$FIELD_NAME FROM RDB$DEPENDENCIES " ;
   + "WHERE RDB$DEPENDENT_NAME = '@@TRIGGERNAME' AND RDB$FIELD_NAME IS NOT NULL"

  aParam := {}
  Aadd(aParam, {"@@TRIGGERNAME", Alltrim(Upper(trigname))})
  aSelect = ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    aRet := ARRAY( Len(aSelect) - 1 )
    aRet := ACopy( aSelect, aRet, 2 )
  Endif
  Return aRet

METHOD clearSystemTables() CLASS rqFBConn AS Logical
  Local bRet := .F.
  // --- Precaution: no opened tables around to clear system tables
  IF Empty(::aAliases) 
    bRet := ::deleteAllRecords( TABLE_OF_OPN )   &&
            ::deleteAllRecords( TABLE_OF_FLOCKS) &&
            ::deleteAllRecords( TABLE_OF_RLOCKS )
    IF bRet
      // --- Log
	  ::writeLog( "[WARN] System Tables were cleared. This is a DANGEROUS ";
                  + "procedure. Make sure you were the only person using " ;
                  + "the system. Unexpected results could happen.", ;
                  LOG_OFF )
    Else
      // --- Log
      ::writeLog( "[ERROR] Failed clearing System Tables. Try again.", ;
                  LOG_OFF )
    Endif
  Else
    // --- Log
    ::writeLog( "[ERROR] Impossible to clear System Tables with opened ";
                + "tables", LOG_OFF )

    raiseError( "[ERROR] Impossible to clear System Tables with opened ";
                + "tables" )
  Endif
  Return bRet

// ------------------------------------------------------- deleteRecord
METHOD deleteRecord( tbl, rec ) CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux
 
  sStmt := ;
    "DELETE FROM @@TABLENAME WHERE RECNO = @@RECNUMBER"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  Aadd(aParam, {"@@RECNUMBER", rec})
  sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    raiseError( "Erro ao deletar registro; Tabela: " + tbl + " Recno: " + rec ;
                + ";FB Error: " + sAux )
    bRet := .F.
  Endif
  Return bRet

// ------------------------------------------------------- deleteAllRecords
METHOD deleteAllRecords( tbl ) CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux
 
  sStmt := ;
    "DELETE FROM @@TABLENAME"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    raiseError( "Erro ao zerar tabela; Tabela: " + tbl + ";FB Error: " + sAux )
    bRet := .F.
  Endif
  Return bRet

// ----------------------------------------------------- dropGenerator
METHOD dropGenerator( genname ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sGenName := ""
  IF !Empty(genname) 
    sGenName := Alltrim(Upper(genname))

    // - - - D R O P  G E N E R A T O R
    sStmt := ;
      "DROP GENERATOR @@GENNAME"

    aParam := {}
    Aadd(aParam, {"@@GENNAME", sGenName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Impossvel eliminar generator; Generator: " + sGenName ;
                   + "; FB Error: " + sAux)
      Endif
    Endif
  Else
    bOk := .F.
  Endif
  Return bOk

// ----------------------------------------------------- dropTrigger
METHOD dropTrigger( trigname ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sTrigName := ""
  IF !Empty(trigname) 
    sTrigName := Alltrim(Upper(trigname))

    // - - - D R O P  T R I G G E R
    sStmt := ;
      "DROP TRIGGER @@TRIGGERNAME"

    aParam := {}
    Aadd(aParam, {"@@TRIGGERNAME", sTrigName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Impossvel eliminar trigger; Trigger: " + sTrigName ;
                   + "; FB Error: " + sAux)
      Endif
    Endif
  Else
    bOk := .F.
  Endif
  Return bOk

// ----------------------------------------------------- dropIndex
METHOD dropIndex( idxname ) CLASS rqFBConn AS Logical
  Local aParam   := {}
  Local sStmt    := ""
  Local sAux     := ""
  Local bOk      := .T.
  Local sIdxName := ""
  IF !Empty(idxname) 
    sIdxName := Alltrim(Upper(idxname))

    // - - - D R O P  I N D E X
    sStmt := ;
      "DROP INDEX @@INDEXNAME"

    aParam := {}
    Aadd(aParam, {"@@INDEXNAME", sIdxName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Impossvel eliminar ndice; ndice: " + sIdxName ;
                   + ";FB Error: " + sAux)
      Endif
    Endif
  Else
    bOk := .F.
  Endif
  Return bOk

// ----------------------------------------------------- dropOnlyTable
METHOD dropOnlyTable( tblname ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sTableName := ""
  IF !Empty(tblname) 
    sTableName := Alltrim(Upper(tblname))

    // - - - D R O P  T A B L E
    sStmt := ;
      "DROP TABLE @@TABLENAME"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Impossvel eliminar a Tabela: " + sTableName ;
                   + ";FB Error: " + sAux)
      Endif
    Endif
  Else
    bOk := .F.
  Endif
  Return bOk

// ----------------------------------------------------- dropView
METHOD dropView( vwname ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sViewName  := ""
  IF !Empty(vwname) 
    sViewName := Alltrim(Upper(vwname))

    // - - - D R O P  V I E W
    sStmt := ;
      "DROP VIEW @@VIEWNAME"

    aParam := {}
    Aadd(aParam, {"@@VIEWNAME", sViewName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Impossvel eliminar View: " + sViewName ;
                   + ";FB Error: " + sAux)
      Endif
    Endif
  Else
    bOk := .F.
  Endif
  Return bOk

// ----------------------------------------------------- dropViewsFromTable
METHOD dropViewsFromTable( tblname ) CLASS rqFBConn AS Logical
  Local bOk     := .F.
  Local aParam  := {}
  Local aSelect := {}
  Local sStmt   := ""
  Local sTableName
  Local iCt

  sStmt := ;
        "SELECT RDB$VIEW_NAME FROM RDB$VIEW_RELATIONS " ;
      + "WHERE RDB$RELATION_NAME = '@@TABLENAME' " 

  IF !Empty(tblname) 
    sTableName := Alltrim(Upper(tblname))
    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    aSelect = ::internalSelect( ::assembleStatement(sStmt, aParam) )
    IF Len(aSelect) >= 2
	  bOk := .T.
      For iCt := 2 To Len(aSelect)
	    IF !::dropView(aSelect[iCt,1])
          bOk := .F.
          Exit
        Endif
      Next
	Else 
	  bOk := .T.
    Endif
  Endif
  Return bOk

// ----------------------------------------------------- dropTable
METHOD dropTable( tblname ) CLASS rqFBConn AS Logical
  Local bOk     := .F.
  Local sTableName
  Local iCt

  IF !Empty(tblname) 
    sTableName := Alltrim(Upper(tblname))
    IF ::dropViewsFromTable(sTableName)
      IF ::dropOnlyTable(sTableName)
        bOk := ::dropGenerator( "G_" + sTableName + "_I" )
        bOk := ::dropGenerator( "G_" + sTableName + "_D" )
      Endif
    Endif
  Endif
  Return bOk

// --------------------------------------------- dropTriggerBasedOnField
PROTECT METHOD dropTriggerBasedOnField( tbl, fldname ) CLASS rqFBConn AS Logical
  Local bOk            := .T.
  Local aSelect        := {}
  Local aParam         := {}
  Local aTriggerFields := {}
  Local aDropIDX       := {}
  Local sStmt          := ""
  Local iCt, iCt2
  Local sFieldName
  
  sStmt := ;
     "SELECT RDB$DEPENDENT_NAME FROM RDB$DEPENDENCIES " ;
   + "WHERE RDB$DEPENDED_ON_NAME = '@@TABLENAME' AND RDB$DEPENDENT_TYPE = 2 " ;
   + "AND RDB$FIELD_NAME = '@@FIELDNAME'"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Alltrim(Upper(tbl))})
  Aadd(aParam, {"@@FIELDNAME", Alltrim(Upper(fldname))})
  aSelect = ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    For iCt := 2 To Len(aSelect)
      aTriggerFields := ::getFieldsOnTrigger(aSelect[iCt,1])
      IF ::dropTrigger(aSelect[iCt,1])
        For iCt2 := 1 To Len(aTriggerFields)
          sFieldName := Alltrim(aTriggerFields[iCt2,1])
          IF Left(sFieldName,Len(RDDIDX_PREFIX)) == RDDIDX_PREFIX 
            IF Empty(aDropIDX) .or. ;
               Ascan(aDropIDX, {|x| x[1] == sFieldName}) == 0
              Aadd(aDropIDX, {sFieldName})
            Endif
          Endif
        Next
      Else
        bOk := .F.
        Exit
      Endif
    Next
  Endif
  // - - - D R O P  R D D I D X  F I E L D S
  //IF bOk
  //  For iCt := 1 To Len(aDropIDX)
  //    IF !::dropField(tbl, {aDropIDX[iCt,1], "", 0, 0})
  //      bOk := .F.
  //      Exit
  //    Endif
  //  Next
  //Endif
  Return bOk

// --------------------------------------------- dropIndexBasedOnField
PROTECT METHOD dropIndexBasedOnField( tbl, fldname ) CLASS rqFBConn AS Logical
  Local bOk            := .T.
  Local aSelect        := {}
  Local aParam         := {}
  Local aTriggerFields := {}
  Local aDropIDX       := {}
  Local sStmt          := ""
  Local iCt, iCt2
  Local sFieldName
  
  sStmt := ;
     "SELECT IDX.RDB$INDEX_NAME FROM RDB$INDICES IDX " ;
   + "LEFT JOIN RDB$INDEX_SEGMENTS SEG ON (SEG.RDB$INDEX_NAME = " ;
   + "IDX.RDB$INDEX_NAME) WHERE IDX.RDB$RELATION_NAME = '@@TABLENAME' AND ";
   + "SEG.RDB$FIELD_NAME = '@@FIELDNAME'"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Alltrim(Upper(tbl))})
  Aadd(aParam, {"@@FIELDNAME", Alltrim(Upper(fldname))})
  aSelect = ::internalSelect( ::assembleStatement(sStmt, aParam) )
  IF Len(aSelect) >= 2
    For iCt := 2 To Len(aSelect)
      IF !::dropIndex(aSelect[iCt,1])
        bOk := .F.
        Exit
      Endif
    Next
  Endif
  Return bOk

// ----------------------------------------------------- dropField
PROTECT METHOD dropField( tbl, fldstru ) CLASS rqFBConn AS Logical
  Local bOk     := .F.
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux    := ""
  Local sFieldName

  sStmt := ;
     "ALTER TABLE @@TABLENAME DROP @@FIELDNAME"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Alltrim(Upper(tbl))})
  sFieldName := fldstru[DBS_NAME]
  IF ::dropIndexBasedOnField(tbl, sFieldName)
    IF ::dropTriggerBasedOnField(tbl, sFieldName)
      Aadd(aParam, {"@@FIELDNAME", sFieldName})
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF Empty(sAux)
        bOk := .T.
      Else
        raiseError( "Erro ao deletar Campo: " + sFieldName + " Tabela: " ;
                    + tbl + ";FB Error: " + sAux )
      Endif
    Endif
  Endif
  Return bOk

PROTECT METHOD getFileName( f ) CLASS rqFBConn AS Character
  Return Substr(f, Rat(PATH_SEPARATOR, f) + 1)

PROTECT METHOD removeFileNameExt( f ) CLASS rqFBConn AS Character
  Local sRet := ""
  IF At(EXTENSION_SEPARATOR, f) > 0
    sRet := Left(f, At(EXTENSION_SEPARATOR, f) - 1)
  Else
    sRet := f
  Endif
  Return sRet

PROTECT METHOD removeZeroBytes( ln ) CLASS rqFBConn AS Character
  Local sRet := ln
  Local iCt
  For iCt := 1 To Len(ln)
    IF Asc(Substr(ln,iCt,1)) == 0
      sRet := Substr(ln,1,iCt - 1)
      Exit
    Endif
  Next
  Return sRet

PROTECT METHOD writeLog( ln, level ) CLASS rqFBConn
  writeLog( ln, level )
  Return

// ----------------------------------------------------- addField
PROTECT METHOD addField( tbl, fldstru ) CLASS rqFBConn AS Logical
  Local bOk     := .T.
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux    := ""
  Local sFieldName, sFieldType, sFieldLen, sFieldDec
 
  sStmt := ;
     "ALTER TABLE @@TABLENAME ADD @@FIELDDEF"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Upper(tbl)})
  sFieldName := fldstru[DBS_NAME]
  sFieldType := fldstru[DBS_TYPE]
  sFieldLen  := Alltrim(Str(fldstru[DBS_LEN]))
  sFieldDec  := Alltrim(Str(fldstru[DBS_DEC]))
  Do Case
    Case sFieldType == "C"
      Aadd(aParam, {"@@FIELDDEF",  RDD_FIELDTYPES[1]})
      Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
    Case sFieldType == "N"
      Aadd(aParam, {"@@FIELDDEF",  RDD_FIELDTYPES[2]})
      Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
      Aadd(aParam, {"@@FIELDDEC",  sFieldDec})  
    Case sFieldType == "D"
      Aadd(aParam, {"@@FIELDDEF",  RDD_FIELDTYPES[3]})
    Case sFieldType == "L"
      Aadd(aParam, {"@@FIELDDEF",  RDD_FIELDTYPES[5]})
    Otherwise
      bOk := .F.
  EndCase
  Aadd(aParam, {"@@FIELDNAME", sFieldName})  
  IF bOk
    sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
      raiseError( "Erro ao adicionar Campo: " + sFieldName + " Tabela: " ;
                  + tbl + ";FB Error: " + sAux )
    Endif
  Endif

  // - - - S E T  F I E L D S  D E S C R I P T I O N
  IF bOk
    Do Case
      Case sFieldType == "N"
        sAux := ::setFieldDescription( tbl, sFieldName, "NUMERIC(" + sFieldLen ;
                                       + "," + sFieldDec + ")" ) 
      Case sFieldType == "L"
        sAux := ::setFieldDescription( tbl, sFieldName, DESCRIPTION_LOGICAL )
    EndCase
    IF !Empty(sAux)
      bOk := .F.
      raiseError( "Erro na descrio do Campo: " + sFieldName + " Tabela: " ;
                  + tbl + ";FB Error: " + sAux )
    Endif
  Endif
  Return bOk

// ----------------------------------------------------- addLimboField
PROTECT METHOD addLimboField( tbl, tblstru, fldpos ) CLASS rqFBConn AS Logical
  Local bOk       := .T.
  Local aFieldAux := AClone(tblstru[fldpos])

  IF !::isLimboField( aFieldAux[DBS_NAME] )
    aFieldAux[DBS_NAME] := RDDLIMBO_PREFIX + tblstru[fldpos][DBS_NAME]
    IF !::isFieldInStru( aFieldAux[DBS_NAME], tblstru )  
      IF !::addField( tbl, aFieldAux )
        bOk := .F.
      Endif
    Endif
  Endif
  Return bOk

// ----------------------------------------------------- renameField
PROTECT METHOD renameField( tbl, oldnm, newnm ) CLASS rqFBConn AS Logical
  Local bOk     := .T. 
  Local sStmt   := ""
  Local aParam  := {}
  Local sAux    := ""

  sStmt := ;
    "ALTER TABLE @@TABLENAME ALTER @@OLDFIELDNAME TO @@NEWFIELDNAME"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME",    Alltrim(Upper(tbl))})
  Aadd(aParam, {"@@OLDFIELDNAME", Alltrim(Upper(oldnm))})
  Aadd(aParam, {"@@NEWFIELDNAME", Alltrim(Upper(newnm))})
  sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    bOk := .F.
    raiseError( "Erro ao renomear Campo: " + oldnm + " Tabela: " ;
                + tbl + ";FB Error: " + sAux )
  Endif
  Return bOk

// ----------------------------------------------------- setFieldPosition
PROTECT METHOD setFieldPosition( tbl, fstru, pos ) CLASS rqFBConn AS Logical
  Local bOk        := .T.
  Local sStmt      := ""
  Local aParam     := {}
  Local sAux       := ""
  Local sFieldName := Alltrim(Upper(fstru[DBS_NAME]))

  sStmt := ;
     "ALTER TABLE @@TABLENAME ALTER @@FIELDNAME POSITION @@FIELDPOS"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Alltrim(Upper(tbl))})
  Aadd(aParam, {"@@FIELDNAME", sFieldName})
  Aadd(aParam, {"@@FIELDPOS",  Strzero(pos,3)})
  sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    bOk := .F.
    raiseError( "Erro ao reposicionar Campo: " + sFieldName + " Tabela: " ;
                + tbl + ";FB Error: " + sAux )
  Endif
  Return bOk

// ----------------------------------------------------- alterField
PROTECT METHOD alterField( tbl, fstru_r, fstru_l ) CLASS rqFBConn AS Logical
  Local bOk       := .F. 
  Local sStmt     := ""
  Local aParam    := {}
  Local sAux      := ""
  Local aFieldAux := {}
  Local sOldFldName

  sStmt := ;
     "UPDATE @@TABLENAME SET @@NEWFIELDNAME = @@OLDFIELDNAME"

  aParam := {}
  Aadd(aParam, {"@@TABLENAME", Alltrim(Upper(tbl))})

  // - - - C R E A T E / U P D A T E  A U X  F I E L D
  aFieldAux := AClone(fstru_r)
  aFieldAux[DBS_NAME] := RDDFLDAUX_PREFIX + Alltrim(Upper(fstru_r[DBS_NAME]))
  IF ::addField(tbl, aFieldAux)
    sOldFldName := Alltrim(Upper(fstru_l[DBS_NAME]))
    Aadd(aParam, {"@@NEWFIELDNAME", aFieldAux[DBS_NAME]})
    Aadd(aParam, {"@@OLDFIELDNAME", sOldFldName})
    sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF Empty(sAux)
      bOk := .T.
    Else
      raiseError( "Erro ao popular AuxCampo: " + aFieldAux[DBS_NAME] ;
                  + " Tabela: " + tbl + ";FB Error: " + sAux )
      ::dropField( tbl, aFieldAux )
    Endif
  Endif
    
  // - - - P R O M O T E S  A U X  F I E L D
  IF bOk
    IF ::dropField(tbl, fstru_l)
      IF !::renameField(tbl, aFieldAux[DBS_NAME], fstru_r[DBS_NAME])
        bOk := .F.
      Endif
    Else
      bOk := .F.
    Endif
  Endif
  Return bOk

METHOD grantAccess( priv, tbl, user ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local aParam := {}
  Local sStmt  := ""
  Local sAux   := ""

  // - - - G R A N T  P R I V I L E G E S
  sStmt := ;
    "GRANT @@PRIVILEGES ON @@TABLENAME TO @@USER"

  aParam := {}
  Aadd(aParam, {"@@PRIVILEGES", priv})
  Aadd(aParam, {"@@TABLENAME",  Alltrim(Upper(tbl))})
  Aadd(aParam, {"@@USER",       user})
  sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    bOk := .F.
    raiseError("Error granting Privileges To:" + user + " Target: " + tbl)
  Endif
  Return bOk

// ----------------------------------------------------- createView
METHOD createView( vwname, fldsalias, stmt ) CLASS rqFBConn AS Logical
  Local bOk    := .T.
  Local aParam := {}
  Local sStmt  := ""
  Local sAux   := ""

  // - - - V I E W  C R E A T I O N
  sStmt := ;
     "CREATE VIEW @@VIEWNAME (@@FIELDS) AS @@SELECT_EXPR"

  aParam := {}
  Aadd(aParam, {"@@VIEWNAME",    Alltrim(Upper(vwname))})
  Aadd(aParam, {"@@FIELDS",      fldsalias})
  Aadd(aParam, {"@@SELECT_EXPR", stmt})
  sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
  IF !Empty(sAux)
    bOk := .F.
    raiseError("Error on creating VIEW; FB Error: " + sAux)
  Endif
  Return bOk

// ----------------------------------------------------- createTable
METHOD createTable( tblname, struct ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sTableName := ""
  Local sFieldName, sFieldType, sFieldLen, sFieldDec
  Local iCt

  IF ::isTable(tblname)
    bOk := .F.
    raiseError("Table already exists!; Tbl: " + tblname)
  Endif

  IF !Empty(tblname) .and. ValType(struct) == "A" .and. !Empty(struct) ;
     .and. bOk
    sTableName := Alltrim(Upper(tblname))

    // - - - T A B L E  C R E A T I O N
    sStmt  := ;
      "CREATE TABLE @@TABLENAME( @@FIELDS, PRIMARY KEY(@@RECNO) )"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    For iCt := 1 To Len(struct)
      sFieldName := Alltrim(Upper(struct[iCt,DBS_NAME]))
      sFieldType := Left(Alltrim(Upper(struct[iCt,DBS_TYPE])),1)
      sFieldLen  := Alltrim(Str(struct[iCt,DBS_LEN]))
      sFieldDec  := Alltrim(Str(struct[iCt,DBS_DEC]))
      Do Case
        Case sFieldType == "C"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[1] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
          Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
        Case sFieldType == "N"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[2] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
          Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
          Aadd(aParam, {"@@FIELDDEC",  sFieldDec})  
        Case sFieldType == "D"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[3] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
        Case sFieldType == "L"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[5] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
        Otherwise
#Cinline
{
          // --- Log
          sprintf( _rqfbconn_event_, "[ERROR] FieldType not supported on table "
                   "creation. Table %s Field: %s Type %s", 
                   VAR_CHR(VAR_NAME_LOCAL(stablename)), 
                   VAR_CHR(VAR_NAME_LOCAL(sfieldname)), 
                   VAR_CHR(VAR_NAME_LOCAL(sfieldtype)) );
          rddlog( LOG_OFF, _rqfbconn_event_ );

}
#endCinline
          bOk := .F.
          raiseError("FieldType not supported; Field: " + sFieldName ;
                     + " Type: " + sFieldType)
      EndCase
    Next
    Aadd(aParam, {"@@FIELDS",   RDD_FIELDTYPES[6] + ",@@FIELDS"}) // AreaID auto
    Aadd(aParam, {"@@FIELDS",   RDD_FIELDTYPES[4]}) // Recno auto added
    Aadd(aParam, {"@@RECNO",    RECNO_FIELD})  
    Aadd(aParam, {"@@AREAID",   AREAID_FIELD})  
    Aadd(aParam, {"@@FIELDLEN", Alltrim(Str(AREAID_FIELD_SIZE))})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Error on creating Table; FB Error: " + sAux)
      Endif
    Endif

    // - - - S E T  T A B L E  D E S C R I P T I O N
    IF bOk
      sAux := ::setTableDescription( tblname, COPYRIGHT )
      IF !Empty(sAux)
        bOk := .F.
        raiseError( "Erro na descrio da tabela; Tabela: " + tblname )
      Endif
    Endif

    // - - - S E T  F I E L D S  D E S C R I P T I O N
    IF bOk
      sAux := ""
      For iCt := 1 To Len(struct)
        sFieldName := Alltrim(Upper(struct[iCt,DBS_NAME]))
        sFieldType := Left(Alltrim(Upper(struct[iCt,DBS_TYPE])),1)
        sFieldLen  := Alltrim(Str(struct[iCt,DBS_LEN]))
        sFieldDec  := Alltrim(Str(struct[iCt,DBS_DEC]))
        Do Case
          Case sFieldType == "N"
            sAux := ::setFieldDescription( sTableName, sFieldName, "NUMERIC(" ;
                                           + sFieldLen + "," + sFieldDec + ")") 
          Case sFieldType == "L"
            sAux := ::setFieldDescription( sTableName, sFieldName, ;
                                           DESCRIPTION_LOGICAL )
        EndCase
        IF !Empty(sAux)
          bOk := .F.
          raiseError( "Erro na descrio do campo; Tabela: " + tblname ;
                      + " Campo: " + sFieldName )
        Endif
      Next
    Endif

    // - - - I N D E X  R E C N O  D E S C 
    sStmt := ;
      "CREATE UNIQUE DESC INDEX @@TABLENAME_RECNO_DESC ON @@TABLENAME(RECNO)"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Erro na criao do ndice recno; FB Error: " + sAux)
      Endif
    Endif

    // - - - G E N E R A T O R  O F  I N S E R T S
    sStmt := ;
      "CREATE GENERATOR G_@@TABLENAME_I"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Erro na criao do generator I; FB Error: " + sAux)
      Endif
    Endif

    // - - - G E N E R A T O R  O F  D E L E T E S 
    sStmt := ;
      "CREATE GENERATOR G_@@TABLENAME_D"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Erro na criao do generator D; FB Error: " + sAux)
      Endif
    Endif

    // - - - T R I G G E R  R E C N O 
    sStmt := ;
       "CREATE TRIGGER T_@@TRIGGERNAME FOR @@TABLENAME "    ;
     + "ACTIVE BEFORE INSERT POSITION 0 AS "                ;
     + "DECLARE VARIABLE CT INTEGER; "                      ;
     + "BEGIN "                                             ;
     + "  NEW.@@RECNO = GEN_ID( G_@@TABLENAME_I, 1 ); "     ;
     + "  IF( NEW.@@AREAID IS NOT NULL ) THEN "             ;
     + "  BEGIN "                                           ;
     + "    SELECT COUNT(*) FROM RDDSYS_OPN "               ;
     + "           WHERE OPN_TABLE = '@@TABLENAME' AND "    ;
     + "                 OPN_AREAID = NEW.@@AREAID INTO :CT; "      ;
     + "    IF( CT > 0 ) THEN "                                     ;
     + "    BEGIN "                                                 ;
     + "      UPDATE RDDSYS_OPN SET OPN_APPRECNO = NEW.@@RECNO "    ;
     + "             WHERE OPN_TABLE = '@@TABLENAME' AND "          ;
     + "                   OPN_AREAID = NEW.@@AREAID; "             ;
     + "      NEW.@@AREAID = NULL; "                        ;
     + "    END "   ;
     + "  END "     ;
     + "END "

    aParam := {}
    Aadd(aParam, {"@@TRIGGERNAME", sTableName + "_RECNO"})
    Aadd(aParam, {"@@TABLENAME",   sTableName})
    Aadd(aParam, {"@@RECNO",       RECNO_FIELD})
    Aadd(aParam, {"@@AREAID",      AREAID_FIELD})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Erro na criao da trigger Recno; FB Error: " + sAux)
      Endif
    Endif

    // - - - T R I G G E R  D E L E T E S
    sStmt := ;
       "CREATE TRIGGER T_@@TRIGGERNAME FOR @@TABLENAME " ;
     + "ACTIVE AFTER DELETE POSITION 1 AS " ;
     + "DECLARE VARIABLE IAUX INTEGER;" ;
     + "BEGIN " ;
     + "  IAUX = GEN_ID( G_@@TABLENAME_D, 1 ); " ;
     + "END "

    aParam := {}
    Aadd(aParam, {"@@TRIGGERNAME", sTableName + "_DELETES"})
    Aadd(aParam, {"@@TABLENAME", sTableName})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Erro na criao da trigger Deletes; FB Error: " + sAux)
      Endif
    Endif
  Endif
  Return bOk

// ----------------------------------------------------- createTableExternal
METHOD createTableExternal( tblname, struct, datfile ) CLASS rqFBConn AS Logical
  Local aParam     := {}
  Local sStmt      := ""
  Local sAux       := ""
  Local bOk        := .T.
  Local sTableName := ""
  Local sFieldName, sFieldType, sFieldLen, sFieldDec
  Local iCt

  IF ::isTable(tblname)
    bOk := .F.
    raiseError("A tabela j existe; Tabela: " + tblname)
  Endif

  IF !Empty(tblname) .and. ValType(struct) == "A" .and. !Empty(struct) ;
     .and. bOk
    sTableName := Alltrim(Upper(tblname))

    // - - - E X T E R N A L  T A B L E  C R E A T I O N
    sStmt  := ;
      "CREATE TABLE @@TABLENAME EXTERNAL FILE '@@DATFILE' ( @@FIELDS )"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    Aadd(aParam, {"@@DATFILE",   datfile})
    For iCt := 1 To Len(struct)
      sFieldName := Alltrim(Upper(struct[iCt,DBS_NAME]))
      sFieldType := Left(Alltrim(Upper(struct[iCt,DBS_TYPE])),1)
      sFieldLen  := Alltrim(Str(struct[iCt,DBS_LEN]))
      sFieldDec  := Alltrim(Str(struct[iCt,DBS_DEC]))
      Do Case
        Case sFieldType == "C"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[1] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
          Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
        Case sFieldType == "N"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[2] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
          Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
          Aadd(aParam, {"@@FIELDDEC",  sFieldDec})  
        Case sFieldType == "D"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[3] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
        Case sFieldType == "L"
          Aadd(aParam, {"@@FIELDS",    RDD_FIELDTYPES[5] + ",@@FIELDS"})
          Aadd(aParam, {"@@FIELDNAME", sFieldName})  
          Aadd(aParam, {"@@FIELDLEN",  sFieldLen})  
        Otherwise
#Cinline
{
          // --- Log
          sprintf( _rqfbconn_event_, "[ERROR] FieldType not supported on table "
                   "creation. Table %s Field: %s Type %s", 
                   VAR_CHR(VAR_NAME_LOCAL(stablename)), 
                   VAR_CHR(VAR_NAME_LOCAL(sfieldname)), 
                   VAR_CHR(VAR_NAME_LOCAL(sfieldtype)) );
          rddlog( LOG_OFF, _rqfbconn_event_ );

}
#endCinline
          bOk := .F.
          raiseError("FieldType not supported; Field: " + sFieldName ;
                     + " Type: " + sFieldType)
      EndCase
    Next
    Aadd(aParam, {",@@FIELDS", ""})
    IF bOk
      sAux = ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
        raiseError("Error on creating External Table; FB Error: " ;
                   + sAux)
      Endif
    Endif
  Endif
  Return bOk

// ----------------------------------------------------- createTableOfOPN
METHOD createTableOfOPN() CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local aParam  := {}
  Local sStmt   := ""
  Local sAux    := ""
  Local aStruct := { {OPN_COL_TABLENAME, "C", SIZE_COL_TABLENAME, 0}, ;
                     {OPN_COL_AREAID,    "C", SIZE_COL_AREAID, 0}, ;
                     {OPN_COL_APPRECNO,  "N", 18, 0},          ;
                     {OPN_COL_SHARED,    "C",  1, 0} }

  IF !::isTable( TABLE_OF_OPN )
    IF ::createTable( TABLE_OF_OPN, aStruct )

      sStmt := ;
        "CREATE UNIQUE INDEX I_@@TABLE_OF_OPN ON @@TABLE_OF_OPN " ;
      + "(@@OPN_COL_TABLENAME, @@OPN_COL_AREAID)"

      aParam := {}
      Aadd(aParam, {"@@TABLE_OF_OPN",      TABLE_OF_OPN})
      Aadd(aParam, {"@@OPN_COL_TABLENAME", OPN_COL_TABLENAME})
      Aadd(aParam, {"@@OPN_COL_AREAID",    OPN_COL_AREAID})
      sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bRet := .F.
      Endif 
    Else
      bRet := .F.
    Endif
  Endif
  Return bRet

// ----------------------------------------------------- createTableOfFLocks
METHOD createTableOfFLocks() CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local aParam  := {}
  Local sStmt   := ""
  Local sAux    := ""
  Local aStruct := { {FCK_COL_TABLENAME, "C", SIZE_COL_TABLENAME, 0}, ;
                     {FCK_COL_AREAID,    "C", SIZE_COL_AREAID, 0} }

  IF !::isTable( TABLE_OF_FLOCKS )
    IF ::createTable( TABLE_OF_FLOCKS, aStruct )

      sStmt := ;
        "CREATE UNIQUE INDEX I_@@TABLE_OF_FLOCKS ON @@TABLE_OF_FLOCKS " ;
      + "(@@FCK_COL_TABLENAME)"

      aParam := {}
      Aadd(aParam, {"@@TABLE_OF_FLOCKS",   TABLE_OF_FLOCKS})
      Aadd(aParam, {"@@FCK_COL_TABLENAME", FCK_COL_TABLENAME})
      sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bRet := .F.
      Endif 
    Else
      bRet := .F.
    Endif
  Endif
  Return bRet

// ----------------------------------------------------- createTableOfRLocks
METHOD createTableOfRLocks() CLASS rqFBConn AS Logical
  Local bRet    := .T.
  Local aParam  := {}
  Local sStmt   := ""
  Local sAux    := ""
  Local aStruct := { {RCK_COL_TABLENAME,  "C", SIZE_COL_TABLENAME, 0}, ;
                     {RCK_COL_TABLERECNO, "N", 18, 0}, ;
                     {RCK_COL_AREAID,     "C", SIZE_COL_AREAID, 0} }

  IF !::isTable( TABLE_OF_RLOCKS )
    IF ::createTable( TABLE_OF_RLOCKS, aStruct )
      sStmt := ;
        "CREATE UNIQUE INDEX I_@@TABLE_OF_RLOCKS ON @@TABLE_OF_RLOCKS " ;
      + "(@@RCK_COL_TABLENAME, @@RCK_COL_TABLERECNO)"

      aParam := {}
      Aadd(aParam, {"@@TABLE_OF_RLOCKS",    TABLE_OF_RLOCKS})
      Aadd(aParam, {"@@RCK_COL_TABLENAME",  RCK_COL_TABLENAME})
      Aadd(aParam, {"@@RCK_COL_TABLERECNO", RCK_COL_TABLERECNO})
      sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bRet := .F.
      Endif 
    Else
      bRet := .F.
    Endif
  Endif
  Return bRet

METHOD bringDBF( dbfname ) CLASS rqFBConn AS Logical
  Local hndlDBF   := 0
  Local hndlDAT   := 0
  Local cBuff     := Space(OFFSET_1STFIELD)
  Local bOk       := .F.
  Local aStru     := {}
  Local aStruExt  := {}
  Local aParam    := {}
  Local iFileType, iRecLen, iQtiRec, iStartPos, iEndPos 
  Local iQtiFld, iBuff, iPos
  Local sAux
  Local sTableName
  Local sStmt, sIFS
  Local sFieldName, sFieldType, sValue
  Local iFieldType, iFieldLen
  Local sNewBuff
  Local bHasDates
  Local iCt, iCt2, iCt3

  // - - - O P E N  D B F 
  IF !Empty(dbfname)
    hndlDBF := FOpen(dbfname, FO_READ)
    IF hndlDBF > 0
      FSeek(hndlDBF, 0, FS_START) 
      FRead(hndlDBF, @cBuff, 16) 
      iFileType := StrPeek(@cBuff, OFFSET_DBF_VERS + 1)
      IF iFileType == 3                        // 0x03 = regular .dbf
        bOk := .T.        
      Else
        IF iFileType == 131                    // 0x83 = .dbf and .dbt
          raiseError("Files .dbt and memo not supported")
        Else
          raiseError("Not a DBF File; Filename: " + dbfname)
          FClose (hndlDBF)
        Endif
      Endif
    Else
      raiseError("File not found; Filename: " + dbfname)
    Endif
  Endif

  // - - - R E A D  D B F  S T R U C T U R E
  IF bOk
    iRecLen := Bin2I(Substr(@cBuff, OFFSET_RECLEN + 1, 2))
    IF iRecLen < 0 
      iRecLen := 65535 + iRecLen 
    Endif 
    iQtiRec   := Bin2L(Substr(@cBuff, OFFSET_RECCOUNT + 1, 4))
    iStartPos := Bin2I(Substr(@cBuff, OFFSET_HDRLEN   + 1, 2))
    iEndPos   := iStartPos + (iQtiRec * iRecLen)
    iQtiFld   := Num2Int((iStartPos - 1 - OFFSET_1STFIELD) / SIZE_HEADFIELD)
    iBuff     := (iQtiFld * SIZE_HEADFIELD) + 1   // Include FS header check
    cBuff     := Space(iBuff)
    FSeek(hndlDBF, OFFSET_1STFIELD, FS_START)
    FRead(hndlDBF, @cBuff, iBuff)
#Cinline
{
    // --- Log
    sprintf( _rqfbconn_event_,"[BRING] Bringing to Heaven DBF: %s #Records: %i"
             " #Fields: %i FileSz: %i bytes RecSz: %i bytes Starting at %i",
             VAR_CHR(VAR_NAME_LOCPAR(dbfname)),
             VAR_INT(VAR_NAME_LOCAL(iqtirec)),
             VAR_INT(VAR_NAME_LOCAL(iqtifld)),
             VAR_INT(VAR_NAME_LOCAL(iendpos)),
             VAR_INT(VAR_NAME_LOCAL(ireclen)),
             VAR_INT(VAR_NAME_LOCAL(istartpos)) );
    rddlog( LOG_LEVEL_04,_rqfbconn_event_ );
}
#endCinline
    // - - FlagShip uses the last byte in header for consistency check
    IF StrPeek(@cBuff, iBuff) # 13
      // --- Log
      ::writeLog( "[WARN] Possible DBF header corruption or not a FlagShip " ;
                  + "native DBF! File:" + dbfname, LOG_LEVEL_04 )
    Endif
    iPos := 1
    For iCt := 1 To iQtiFld
      Aadd( aStru, {::removeZeroBytes(Upper(Substr(@cBuff, iPos, 10))), ;   
                    Upper(substr(@cBuff, iPos + 11, 1)), ;
                    StrPeek(@cBuff, iPos + 16), ;
                    StrPeek(@cBuff, iPos + 17)} ) 
      iPos += SIZE_HEADFIELD
    Next
    // - - Normalize Character Size
    For iCt := 1 To Len(aStru)
      IF aStru[iCt,DBS_TYPE] == "C" .and. aStru[iCt,4] > 0 
        aStru[iCt,DBS_LEN] := aStru[iCt,DBS_DEC] * 256 + aStru[iCt,DBS_LEN]
        aStru[iCt,DBS_DEC] := 0
      Endif
    Next
  Endif 

  // - - - C R E A T E  T A B L E  
  IF bOk
    sTableName := Upper(::removeFileNameExt(::getFileName(dbfname)))
    IF !::isTable(sTableName)
      bOk := ::createTable( sTableName, aStru )
    Else 
      bOk := .F.

      // --- Log
      ::writeLog( "[ERROR] Table already exists. Table: " + sTableName, ;
                  LOG_LEVEL_04 )

      raiseError( "Table already exists; Table: " + sTableName )
    Endif
  Endif

  // - - - D A T  F I L E
  IF bOk
    FSeek(hndlDBF, iStartPos, FS_START)
    hndlDAT = FCreate(::getFullPathDATFile(), DAT_FILE_PERMISSIONS) 
    IF hndlDAT > 0 
      For iCt := 1 To iQtiRec
        cBuff    := Space(iRecLen)
        FRead(hndlDBF, @cBuff, iRecLen)
        cBuff    := Substr(cBuff,2)      // Elimina o Lock de Registro
        sNewBuff := ""
        iCt3     := 1
        For iCt2 := 1 To Len(aStru)
          sFieldName := ::removeZeroBytes(aStru[iCt2,DBS_NAME])
          sFieldType := aStru[iCt2,DBS_TYPE]
          iFieldLen  := aStru[iCt2,DBS_LEN]
          sValue     := Substr(cBuff,iCt3,iFieldLen)
          Do Case
            Case sFieldType == "D"
              IF !Empty(StoD(sValue))
                sValue := Substr(sValue,1,4) + "-" + Substr(sValue,5,2) ;
                          + "-" + Substr(sValue,7,2)
              Else
                sValue := "0001-01-01"  // Null Date
              Endif
            Case sFieldType == "N"
              IF Empty(sValue)
                sValue := Padr("0", iFieldLen)
              Endif
          EndCase
          sNewBuff += sValue
          iCt3 += iFieldLen
        Next
        FWrite(hndlDAT, sNewBuff)
      Next
      FClose(hndlDAT)
    Else
      bOk := .F.
      raiseError("Impossible creating temporary file during importing")
    Endif
  Endif

  // - - - C R E A T E  E X T E R N A L  T A B L E
  IF bOk
    For iCt2 := 1 To Len(aStru)
      sFieldName := ::removeZeroBytes(aStru[iCt2,DBS_NAME])
      sFieldType := aStru[iCt2,DBS_TYPE]
      iFieldLen  := aStru[iCt2,DBS_LEN]
      Do Case
        Case sFieldType == "D"
          iFieldLen := 10
      EndCase
      Aadd(aStruExt, {sFieldName, "C", iFieldLen, 0} )
    Next
    IF !::isTable("EXT_" + sTableName)
      bOk := ::createTableExternal( "EXT_" + sTableName, aStruExt, ;
                                    ::getFullPathDATFile() )
    Else
      bOk := .F.
      raiseError( "An external table with the same name already exists!" )
    Endif
  Endif

  // - - - C R E A T E  T R I G G E R  N U L L  D A T E S 
  IF bOk
    sStmt := ;
      "CREATE TRIGGER T_@@TRIGGERNAME FOR @@TABLENAME " ;
    + "ACTIVE BEFORE INSERT POSITION 0 AS " ;
    + "BEGIN " ;
    + "  @@IFS " ;
    + "END "

    sIFS := ;
      "IF ( NEW.@@DATEFIELD = '0001-01-01' ) THEN " ;
    + "  NEW.@@DATEFIELD = NULL; "
   
    aParam := {}
    Aadd(aParam, {"@@TRIGGERNAME", sTableName + "_NULLDATES"})
    Aadd(aParam, {"@@TABLENAME",   sTableName})
    bHasDates := .F.
    For iCt := 1 To Len(aStru)
      sFieldName := ::removeZeroBytes(aStru[iCt,DBS_NAME])
      sFieldType := aStru[iCt,DBS_TYPE]
      IF sFieldType == "D"
        Aadd(aParam, {"@@IFS", sIFS + " @@IFS"})
        Aadd(aParam, {"@@DATEFIELD", sFieldName})
        bHasDates := .T.
      Endif
    Next
    Aadd(aParam, {"@@IFS", ""})
    IF bHasDates
      sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
      IF !Empty(sAux)
        bOk := .F.
      Endif
    Endif
  Endif

  // - - - P O P U L A T E  T A B L E
  IF bOk
    sStmt := ;
      "INSERT INTO @@TABLENAME SELECT @@COLUMNS FROM @@EXTTABLE"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", sTableName})
    Aadd(aParam, {"@@EXTTABLE",  "EXT_" + sTableName})
    For iCt := 1 To Len(aStruExt)
      sFieldName := ::removeZeroBytes(aStruExt[iCt,DBS_NAME])
      Aadd(aParam, {"@@COLUMNS",  sFieldName + ",@@COLUMNS"})
    Next
    Aadd(aParam, {"@@COLUMNS", "null,@@COLUMNS"}) // Fill AreaID
    Aadd(aParam, {"@@COLUMNS", "0"})              // Fill Recno
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
      raiseError( "Error transfering data from External Table " ;
                  + "to the official one; FB Error: " + sAux )
    Endif

    // - - - D R O P  T R I G G E R  N U L L  D A T E S
    IF bHasDates
      ::dropTrigger( "T_" + sTableName + "_NULLDATES" )
    Endif
    // - - - D R O P  E X T E R N A L  T A B L E
    ::dropOnlyTable( "EXT_" + sTableName )

    FClose(hndlDBF)
  Endif
  Return bOk

METHOD exileTable( tblname ) CLASS rqFBConn AS Logical
  Local aAux     := {}
  Local aStru    := {}
  Local aStruExt := {}
  Local aParam   := {}
  Local sStmt    
  Local hndlDAT  := 0
  Local bOk      := .F.
  Local iCt
  Local sFieldName, iFieldLen, sFieldType
  Local sAux

  // - - - C R E A T E  D B F
  aStru := ::filterRDDFields(::getTableStruct(tblname)) 
  IF !Empty(aStru)
    bOk := DbCreate(tblname + ".dbf", aStru, RDD_DRIVER_DBFIDX)
  Endif

  // - - - E R A S E  D A T  F I L E
  IF bOk      
    hndlDAT = FCreate(::getFullPathDATFile(), DAT_FILE_PERMISSIONS)
    IF hndlDAT > 0
      FClose(hndlDAT)
    Else
      bOk := .F.
    Endif
  Endif
     
  // - - - C R E A T E  E X T E R N A L  T A B L E
  IF bOk
    For iCt := 1 To Len(aStru)
      Aadd(aStruExt, {aStru[iCt,DBS_NAME], "C", aStru[iCt,DBS_LEN], 0} )
    Next
    Aadd(aStruExt, {"NEWLINE", "C", 1, 0})
    IF !::isTable("EXT_" + tblname)
      bOk := ::createTableExternal( "EXT_" + tblname, aStruExt, ;
                                    ::getFullPathDATFile() )
    Else
      bOk := .F.
      raiseError( "An external table with the same name already exists!" )
    Endif
  Endif

  // - - - P O P U L A T E  E X T E R N A L  T A B L E
  IF bOk
    sStmt := ;
      "INSERT INTO @@EXTTABLE SELECT @@COLUMNS FROM @@TABLENAME"

    aParam := {}
    Aadd(aParam, {"@@TABLENAME", tblname})
    Aadd(aParam, {"@@EXTTABLE",  "EXT_" + tblname})
    For iCt := 1 To Len(aStruExt) - 1  // exclude newline
      sFieldName := aStruExt[iCt,DBS_NAME]
      sFieldType := aStru[iCt,DBS_TYPE]
      Do Case
        Case sFieldType == "D"
          Aadd(aParam, {"@@COLUMNS", "SkipPad('{{D}}'||" + sFieldName ;
                                     + "),@@COLUMNS"})
        Otherwise
          Aadd(aParam, {"@@COLUMNS", sFieldName + ",@@COLUMNS"})
      Endcase
    Next
    Aadd(aParam, {"@@COLUMNS", "'" + LINEFEED + "'"})
    sAux := ::executeUpdate( ::assembleStatement(sStmt, aParam) )
    IF !Empty(sAux)
      bOk := .F.
      raiseError( "Error transfering data from the official table " ;
                  + "to the external one; FB Error: " + sAux )
    Endif
  Endif

  // - - - P O P U L A T E  D B F
  IF bOk
    Use (tblname) Alias dbffile Via RDD_DRIVER_DBFIDX New
    Append From (::getFullPathDATFile()) SDF
    Close dbffile
  Endif

  // - - - D R O P  E X T E R N A L  T A B L E
  IF ::isTable("EXT_" + tblname)
    ::dropOnlyTable( "EXT_" + tblname )
  Endif

  Return bOk

PROTECT METHOD fCrtFile( arq ) CLASS rqFBConn AS Numeric
  Return FCreate( arq, FC_NORMAL )

PROTECT METHOD fOpnFile(arq, [acesso]) CLASS rqFBConn AS Numeric
  Local iRetry := 0
  Local hndl   := -1
  Local modo   := acesso
  IF Empty(modo)
    modo := "r"
  Endif
  Do Case
    Case modo == "rw"
      hndl := FOpen(arq, FO_READWRITE)
      IF hndl >= 0
        Do While FLockf(hndl, FF_READWRITE) # 0 .and. iRetry <= TIMEOUT_FOPN
          Inkey(HALF_SECOND)
          iRetry++
        Enddo
        IF iRetry >= TIMEOUT_FOPN
          FClose(hndl)
        Endif
      Endif
    Case modo == "r"
      hndl := FOpen(arq, FO_READ)
  Endcase
  Return hndl

PROTECT METHOD fClsFile( hndl ) CLASS rqFBConn
  IF hndl >= 0
    FLockf(hndl, FF_UNLOCK)
    FClose(hndl)
  Endif
  Return

PROTECT METHOD toArray( ln, separador ) CLASS rqFBConn AS Array
  Local aRet    := {}
  Local bAddNIL := .F.
  Local sVl
  Local iPos
  IF ValType(ln) == VALTYPE_CHARACTER .and. ;
     ValType(separador) == VALTYPE_CHARACTER

    IF Right(ln, Len(separador)) == separador
      bAddNIL := .T.
    Endif
    Do While At(separador,ln) > 0
      iPos := At(separador,ln)
      sVl  := Substr(ln,1,iPos - 1)
      AAdd(aRet, sVl)
      ln := Substr(ln, iPos + Len(separador), Len(ln))
    Enddo
    IF Len(ln) > 0
      AAdd(aRet, ln)
    Endif
    IF bAddNIL
      AAdd(aRet, "")
    Endif
  Endif
  Return aRet

PROTECT METHOD extracMatrixOfTable( tblname ) CLASS rqFBConn AS Array
  Local aRet   := {}
  Local aStru  := {}
  Local bOk    := .F.
  Local iCt
  Local sFieldName, sFieldType, sPosition, sType, sDescr
  Local iFieldLen, iFieldDec
  Local sAux

  tblname := Upper(tblname)
  // - - - C H E C K  F O R  S Y S T E M  T A B L E S
  Do Case
    Case tblname == TABLE_OF_FLOCKS
    Case tblname == TABLE_OF_RLOCKS
    Case tblname == TABLE_OF_OPN
    Otherwise
      bOk := .T.
  EndCase

  // - - - A S S E M B L E S  M A T R I X  S T R U C T U R E  C O N T E N T
  IF bOk 
    IF ::isTable( tblname )
      IF ::isRQRDDTable( tblname )
        Aadd(aRet,{ Replicate("- ", 5) + COPYRIGHT + Replicate(" -", 5)  })
        Aadd(aRet,{ "-" + Space(17) + "v2.0" + Space(17) + "-"           })
        Aadd(aRet,{ "   Data: " + DtoC(Date()) + " " + Time()            })
        Aadd(aRet,{ "     DB: " + ::sDBName                              })
        Aadd(aRet,{ " Tabela: " + tblname                                })
        Aadd(aRet,{ Space(46) + "!!--- Alterar Aqui ---!!"               })
        aStru := ::filterRDDFields(::getTableStruct(tblname))
        For iCt := 1 To Len(aStru)
          sFieldName  := aStru[iCt,DBS_NAME]
          sFieldType  := aStru[iCt,DBS_TYPE] 
          iFieldLen   := aStru[iCt,DBS_LEN]
          iFieldDec   := aStru[iCt,DBS_DEC]
          sPosition   := StrZero(iCt,3)
          sType       := sFieldType + MATRIX_STRU_SEPARATOR + ;
                         Alltrim(Str(iFieldLen)) + MATRIX_STRU_SEPARATOR + ;
                         Alltrim(Str(iFieldDec))
          sDescr      := sPosition + MATRIX_STRU_SEPARATOR + sFieldName + ;
                         MATRIX_STRU_SEPARATOR + sType
          Aadd(aRet,{ " " + tblname + MATRIX_TABLE_SEPARATOR + sDescr + ;
                      Space(40 - Len(sDescr) - Len(tblname)) + ;
                      MATRIX_LR_SEPARATOR + " " + sDescr })
        Next
      Endif
    Else
      Aadd(aRet,{ " >> Does not exists!" })
    Endif
  Endif
  Return aRet

METHOD extractMatrix( atables, [filename] ) CLASS rqFBConn AS Array
  Local aRet   := {}
  Local aTable := {}
  Local bOk    := .F.
  Local iCt, iCt2
  Local handle

  Aadd(aRet, {{{"  :   :  .::::. .:::::. .:     .: ::::::"}, ;
               {" ::: ::.     ::   ::    ::.::: ::    .:'"}, ;
               {":: ::::: .:::::   ::    :::    ::   .:' "}, ;
               {"::  : :: ::  ::.  ::    ::     ::  .:'  "}, ;
               {":     `: '::::::. `:.   :      :  ::::::"}}})

  // - - - W R I T E  M A T R I X  A R R A Y
  IF Empty(atables) 
    atables := ::getTablesName()    // All database
  Endif
  IF ValType(atables) == VALTYPE_ARRAY
    For iCt := 1 To Len(atables)
      Aadd(aRet,{ ::extracMatrixOfTable(atables[iCt]) })
    Next
    bOk := .T.
  Else
    raiseError( "No foi possvel extrair a matriz.; Coloque os nomes das " +
                "tabelas em um array unidimensional" )
  Endif

  // - - - W R I T E  M A T R I X  F I L E
  IF bOk .and. !Empty(filename)
    handle := ::fCrtFile( filename + MATRIX_FILE_EXTENSION )
    IF FError() == 0
      For iCt := 1 To Len(aRet)
        aTable := aRet[iCt,1]
        For iCt2 := 1 To Len(aTable)
          FWrite(handle, aTable[iCt2,1] + CRLF)
          IF FError() != 0
            Exit
          Endif
        Next
        IF FError() != 0
          raiseError( "Erro na gravao do arquivo matriz" )
          Exit
        Endif
      Next
      ::fClsFile(handle)
    Endif
  Endif
  Return aRet

PROTECT METHOD isValidFieldDescriptor( adescr ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF Len(adescr) >= 5  .and. Val(adescr[1]) > 0 .and. !Empty(adescr[2]) ;
     .and. Upper(adescr[3]) $ "CNDL" .and. Val(adescr[4]) > 0 
    bRet := .T.
  Endif
  Return bRet

PROTECT METHOD isFieldStructureDifferent( astru1, astru2, arrpos )                             CLASS rqFBConn AS Logical
  Local bRet := .T.
  IF arrpos <= Len(astru1) .and. arrpos <= Len(astru2)
    IF astru1[arrpos,DBS_NAME] == astru2[arrpos,DBS_NAME] .and. ;
       astru1[arrpos,DBS_TYPE] == astru2[arrpos,DBS_TYPE] .and. ;
       astru1[arrpos,DBS_LEN]  == astru2[arrpos,DBS_LEN]  .and. ;
       astru1[arrpos,DBS_DEC]  == astru2[arrpos,DBS_DEC]
      bRet := .F.
    Endif
  Endif
  Return bRet

PROTECT METHOD sameFieldName( astru1, astru2, arrpos )                                         CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF arrpos <= Len(astru1) .and. arrpos <= Len(astru2)
    IF astru1[arrpos,DBS_NAME] == astru2[arrpos,DBS_NAME]
      bRet := .T.
    Endif
  Endif
  Return bRet

PROTECT METHOD hasFieldNameChanged( fldstru_r, fldstru_l )                                     CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF !(fldstru_r[DBS_NAME] == fldstru_l[DBS_NAME])
    bRet := .T.
  Endif
  Return bRet

PROTECT METHOD isFieldStructureChanged( fldstru, tblstru )                                     CLASS rqFBConn AS Logical
  Local bRet := .F.
  Local iAux := AScan(tblstru, {|x| x[DBS_NAME] == fldstru[DBS_NAME]})
  IF iAux > 0
    IF tblstru[iAux,DBS_TYPE] != fldstru[DBS_TYPE] .or. ;
       tblstru[iAux,DBS_LEN]  != fldstru[DBS_LEN]  .or. ;
       tblstru[iAux,DBS_DEC]  != fldstru[DBS_DEC]
      bRet := .T.
    Endif
  Endif
  Return bRet

PROTECT METHOD isFieldDropped( fldstru ) CLASS rqFBConn AS Logical
  Local bRet := .F.
  IF fldstru[DBS_NAME] == MATRIX_DROPFIELD_TAG
    bRet := .T.
  Endif
  Return bRet

METHOD applyMatrix( filearray, [simulate] ) CLASS rqFBConn AS Array
  Local bOk         := .T.
  Local aLeft       := {}
  Local aRight      := {}
  Local aLog        := {}
  Local aStruActual := {}
  Local aMatrix     := {}
  Local aAux        := {}
  Local aLine       := {}
  Local aTable      := {}
  Local aRddFields  := {}
  Local bNewTable   := .F.
  Local bDropTable  := .F.
  Local bSemErros   := .T.
  Local bSimulation := IIF(Empty(simulate), .F., .T.)
  Local filename    := ""
  Local iCt, iCt2, iCt3, iCt4
  Local sLine, sLeft, sRight, sTableName, sFieldName
  Local handle
  Local bReOrder, bAdd, bAlter, bDrop
  Local iAux

  IF ValType(filearray) == VALTYPE_CHARACTER
    filename := filearray
  ElseIF ValType(filearray) == VALTYPE_ARRAY
    aMatrix := filearray
  Else
    bOk := .F.
  Endif

  // - - - R E A D  M A T R I X  F R O M  F I L E
  IF bOk
    IF !Empty(filename)
      handle := ::fOpnFile( filename )
      Do While FError() == 0
        sLine := FReadTxt(handle)
        IF At(COPYRIGHT, sLine) > 0
          IF !Empty(aAux)
            Aadd(aMatrix, {aAux})
            aAux := {}
          Endif
        Endif
        Aadd(aAux, {sLine})
      Enddo
      IF !Empty(aAux)
        Aadd(aMatrix, {aAux})
      Endif
    Endif
  Endif
  IF Empty(aMatrix)
    raiseError( "Voc deseja que eu processe uma matriz vazia?" )
    bOk       := .F.
    bSemErros := .F.
  Endif

  IF bOk
    For iCt := 1 To Len(aMatrix)
      aStruActual := {}
      aLeft       := {}
      aRight      := {}
      sTableName  := ""
      bNewTable   := .F.
      bDropTable  := .F.
      bOk         := .T.
      aTable      := aMatrix[iCt,1] 

      // - - - T H E  T H R E E  S T R U C T U R E S

      For iCt2 := 1 To Len(aTable)
        IF At(MATRIX_LR_SEPARATOR, aTable[iCt2,1]) > 0
          sLeft := sRight := ""
          aLine := ::toArray( aTable[iCt2,1], MATRIX_LR_SEPARATOR )
          IF At(MATRIX_TABLE_SEPARATOR, aLine[1]) > 0
            aAux := ::toArray( aLine[1], MATRIX_TABLE_SEPARATOR )
            IF Empty(sTableName)
              sTableName := Alltrim(aAux[1])
            Else
              IF !(Alltrim(aAux[1]) == sTableName) 
                raiseError( "Os dados para esta tabela parecem corrompidos;" ;
                            + "Tabela: " + sTableName )
                bOk       := .F.
                bSemErros := .F.
                Exit
              Endif
            Endif
            sLeft  := aAux[2]
            sRight := aLine[2]
          Else
            raiseError( "Os dados para esta tabela parecem corrompidos;" ;
                        + "Registro: " + aLine[1] )
            bOk       := .F.
            bSemErros := .F.
            Exit
          Endif
          IF !Empty(sTableName)
 
            // - - - L E F T
            IF At(MATRIX_STRU_SEPARATOR, sLeft) > 0
              aAux := ::toArray( sLeft, MATRIX_STRU_SEPARATOR )
              IF ::isValidFieldDescriptor(aAux)
                Aadd(aLeft, { Alltrim(Upper(aAux[2])), Upper(aAux[3]), ;
                              Val(aAux[4]), Val(aAux[5]) })
              Else
                bOk := .F.
              Endif
            ElseIF At(MATRIX_DROPTABLE_TAG, sLeft) > 0
              bDropTable := .T.
              Exit
            Else
              bOk := .F.
            Endif
            IF !bOk
              raiseError( "Os dados para esta tabela parecem corrompidos;" ;
                          + "Campo: " + sLeft + ";Tabela: " + sTableName )
              bSemErros := .F.
              Exit
            Endif

            // - - - R I G H T
            IF At(MATRIX_STRU_SEPARATOR, sRight) > 0
              aAux := ::toArray( sRight, MATRIX_STRU_SEPARATOR )
              IF ::isValidFieldDescriptor(aAux)
                Aadd(aRight, { Alltrim(Upper(aAux[2])), Upper(aAux[3]), ;
                               Val(aAux[4]), Val(aAux[5]) })
              Else
                bOk := .F.
              Endif
            ElseIF At(MATRIX_DROPFIELD_TAG, sRight) > 0
              Aadd(aRight, {MATRIX_DROPFIELD_TAG, "", 0, 0})
            Else
              bOk := .F.
            Endif
            IF !bOk
              raiseError( "Os dados para esta tabela parecem corrompidos;" ;
                          + "Campo: " + sRight + ";Tabela: " + sTableName )
              bSemErros := .F.
              Exit
            Endif

          Else
            raiseError( "No foi possvel determinar a tabela alvo" )
            bOk       := .F.
            bSemErros := .F.
            Exit
          Endif 
        Endif
      Next
      IF bOk .and. !Empty(sTableName)
        IF ::isTable(sTableName)
          IF ::isRQRDDTable(sTableName)
            aStruActual := ::filterRDDFields(::getTableStruct(sTableName)) 
          Else
            raiseError( "Impossvel modificar tabelas aliengenas!!;" ;
                        + "Tabela: " + sTableName )
            bOk       := .F.
            bSemErros := .F.
          Endif
        Else
          bNewTable := .T.
        Endif
      Endif

      // - - - A P P L Y  N O W
      IF bOk
        Aadd(aLog, { Replicate("- ",25) + "[ " + sTableName + " ]" })
        IF bDropTable                                // D R O P  T A B L E
          IF ::isRQRDDTable(sTableName)
            IF (IIF(bSimulation, .T., ::dropTable(sTableName)))
              Aadd(aLog, { "[OK] Table dropped" })
            Else
              Aadd(aLog, { "[ERROR] Dropping table" })
              bSemErros := .F.
            Endif
          Else
            Aadd(aLog, { "[WARN] Table already dropped" })
          Endif
        ElseIF bNewTable                             // C R E A T E  T A B L E
          IF (IIF(bSimulation, .T., ::createTable(sTableName, aRight)))
            Aadd(aLog, { "[OK] Table created!" })
          Else
            Aadd(aLog, { "[ERROR] Creating table" })
            bSemErros := .F.
          Endif
        Else                                         // A L T E R  T A B L E
          // - - - Alter, Add or do Nothing
          bReOrder := .F.
          For iCt3 := 1 To Len(aRight)
            bAdd := bAlter := bDrop := .F.
            IF ::isFieldDropped( aRight[iCt3] ) 
              IF ::isFieldInStru( aLeft[iCt3,DBS_NAME], aStruActual )
                bDrop := .T.
              Else
                Aadd(aLog, ;
                     {"[WARN] Field already dropped: " + aLeft[iCt3,DBS_NAME]})
              Endif
            Else
              IF ::isFieldStructureDifferent( aRight, aStruActual, iCt3 )
                IF ::sameFieldName( aRight, aStruActual, iCt3 ) .or. ;
                   ::hasFieldNameChanged( aRight[iCt3], aLeft[iCt3] )
                  IF ::isFieldInStru( aLeft[iCt3,DBS_NAME], aStruActual )
                    bReOrder := .T.
                    bAlter   := .T.
                  Endif
                Else
                  bReOrder := .T.
                  IF ::isFieldInStru( aLeft[iCt3,DBS_NAME], aStruActual )
                    IF ::isFieldStructureChanged( aRight[iCt3], aStruActual )
                      bAlter := .T.
                    Endif
                  Else
                    bAdd := .T.
                  Endif
                Endif
              Endif       
            Endif
            IF bAdd                                     // A D D  F I E L D
              IF (IIF(bSimulation, .T., ::addField(sTableName, aRight[iCt3])))
                Aadd(aLog, { "[OK] Added Field: " + aRight[iCt3,DBS_NAME] })
              Else
                Aadd(aLog, { "[ERROR] Add Field: " + aRight[iCt3,DBS_NAME] })
                bSemErros := .F.
              Endif
            ElseIF bAlter                               // A L T E R  F I E L D
              IF (IIF(bSimulation, .T., ;
                      ::alterField(sTableName, aRight[iCt3], aLeft[iCt3])))
                Aadd(aLog, { "[OK] Alter Field: " + aLeft[iCt3,DBS_NAME] })
              Else
                Aadd(aLog, { "[ERROR] Alter Field: " + aLeft[iCt3,DBS_NAME] })
                bSemErros := .F.
              Endif
            ElseIF bDrop                                // D R O P  F I E L D
              IF (IIF(bSimulation, .T., ::dropField(sTableName, aLeft[iCt3])))
                Aadd(aLog, { "[OK] Drop Field: " + aLeft[iCt3,DBS_NAME] })
              Else
                Aadd(aLog, { "[ERROR] Drop Field: " + aLeft[iCt3,DBS_NAME] })
                bSemErros := .F.
              Endif
            Endif
          Next
          // - - - Add Limbo Fields
          For iCt3 := 1 To Len(aStruActual)
            IF !::isFieldInStru( aStruActual[iCt3,DBS_NAME], aLeft ) .and. ;
               !::isLimboField( aStruActual[iCt3,DBS_NAME] )
              IF (IIF(bSimulation, .T., ;
                  ::addLimboField(sTableName, aStruActual, iCt3)))
                Aadd(aLog, { "[WARN] Add LimboField: " ;
                             + aStruActual[iCt3,DBS_NAME]} )
              Else
                Aadd(aLog, { "[ERROR] Add LimboField: ";
                             + aStruActual[iCt3,DBS_NAME] })
                bSemErros := .F.
              Endif
            Endif
          Next
          // - - - ReOrder
          IF bReOrder
            For iCt3 := 1 To Len(aRight)
              sFieldName := aRight[iCt3,DBS_NAME]
              IF !::isFieldDropped(aRight[iCt3])
                IF (IIF(bSimulation, .T., ;
                    ::setFieldPosition(sTableName, aRight[iCt3], iCt3)))
                  Aadd(aLog, ;
                       { "[OK] Reorder: " + Strzero(iCt3,3) + " Field: " ;
                         + sFieldName })
                Else
                  Aadd(aLog, ;
                       { "[ERROR] Reorder: " + Strzero(iCt3,3) + " Field: " ;
                         + sFieldName })
                  bSemErros := .F.
                Endif
              Endif
            Next
            aRddFields := ::filterNormalFields(::getTableStruct(sTableName))
            For iCt4 := 1 To Len(aRddFields)
              sFieldName := aRddFields[iCt4,DBS_NAME]
              IF (IIF(bSimulation, .T., ;
                  ::setFieldPosition(sTableName, aRddFields[iCt4], iCt3)))
                iCt3++
              Else
                Aadd(aLog, ;
                     { "[ERROR] Reorder SysField: " + sFieldName + " At: " ;
                       + Strzero(iCt3,3) })
                bSemErros := .F.
              Endif
            Next
          Endif
        Endif
      Endif
    Next 
  Endif
  Return { bSemErros, aLog }

METHOD checkMatrix( filearray ) CLASS rqFBConn AS Array
  Return ::applyMatrix( filearray, .T. )

METHOD EvalMeth( met, param ) CLASS rqFBConn
  Local iLen := 0
  IF Empty(param)
    param := {}
  ElseIF ValType(param) == VALTYPE_ARRAY
    iLen := Len(param)
  Endif
  Return ::&met(IIF(iLen >= 1, param[1], ""), ;
                IIF(iLen >= 2, param[2], ""), ;
                IIF(iLen >= 3, param[3], ""), ;
                IIF(iLen >= 4, param[4], ""), ;
                IIF(iLen >= 5, param[5], ""))

Static Function writeLog( ln, level )
#Cinline
{
    snprintf( _rqfbconn_event_, 
            MAX_LOG_EVENT_SIZE, "%s", VAR_CHR(VAR_NAME_LOCPAR(ln)) );
    rddlog( VAR_INT(VAR_NAME_LOCPAR(level)), _rqfbconn_event_ );
}
#endCinline
  Return

Static Function raiseError( msg )
  Local oErr := Error{} AS Error
  oErr:DESCRIPTION := "[FBCONN Error];" + msg
  oErr:GENCODE     := -1
  oErr:SEVERITY    := ES_RTERR
  Eval(ErrorBlock(),oErr)
  Return

