/*
RqRDDFire - A Free FlagShip RDD for Firebird RDBMS
v.0.9.0, September 2007, Brazil
Ricardo Delamar Roque (rroque6428@terra.com.br)

License
=======
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed AS IS in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/licenses/lgpl-2.1.html
*/

#Cinline
#include <FSopenc.h>
#endCinline

#include "rqfbconn.fh"
#include "dbstruct.fh" 
#include "dataserv.fh"
#include "error.fh"
#include "rqrdd.h"

#define NULL_COLUMN             "NULL"

#ifdef FlagShip5
#define NOPIXEL NoPixel
#else
#define NOPIXEL
#endif

CLASS rqFBRS INHERIT DataServer
  PROTECT sStatement      := ""                    AS Character
  PROTECT sAlias          := ""                    AS Character
  PROTECT iLogLevel       := LOG_OFF               AS IntVar
  PROTECT iActualRow      := 0                     AS IntVar
  PROTECT aStru           := {}                    AS Array
  PROTECT aFieldVars      := {}                    AS Array
  PROTECT aFetchedRS      := {}                    AS Array
  PROTECT bActive         := .F.                   AS Logical
  PROTECT bBeforeFirst    := .T.                   AS Logical
  PROTECT bConn           := .F.                   AS Logical
  PROTECT bShared         := .T.                   AS Logical
  PROTECT bReadOnly       := .T.                   AS Logical
  PROTECT bEof            := .T.                   AS Logical
  PROTECT bBof            := .T.                   AS Logical
  PROTECT bLimboFields    := .F.                   AS Logical
  PROTECT oConn          

PROTOTYPE PROTECT METHOD writeLog(p1,p2)     CLASS rqFBRS
PROTOTYPE PROTECT METHOD openResultSet(p1)   CLASS rqFBRS
PROTOTYPE PROTECT METHOD closeRS()           CLASS rqFBRS
PROTOTYPE METHOD GoTop()                     CLASS rqFBRS
PROTOTYPE METHOD getFetchedSize()            CLASS rqFBRS
PROTOTYPE METHOD isRSpromotedToAlias()       CLASS rqFBRS AS Logical
PROTOTYPE METHOD DbStruct()                  CLASS rqFBRS

ACCESS Used CLASS rqFBRS
  Return ::bActive
ACCESS Shared CLASS rqFBRS
  Return ::bShared
ACCESS ReadOnly CLASS rqFBRS
  Return ::bReadOnly
ACCESS Alias CLASS rqFBRS
  Return ::sAlias
ACCESS Eof CLASS rqFBRS
  Return ::bEof
ACCESS Bof CLASS rqFBRS
  Return ::bBof
ACCESS LastRec CLASS rqFBRS
  Return ::getFetchedSize()
ACCESS Deleted CLASS rqFBRS
  Return .F.
ACCESS FCount CLASS rqFBRS
  Return Len(::DbStruct())
ACCESS Recno CLASS rqFBRS
  Local nRet := 0
  IF ::bActive
    IF ::bEof
      nRet := ::getFetchedSize() + 1
    ElseIF ::bBof
      nRet := 1
    Else
      nRet := ::iActualRow
    Endif
  Endif
  Return Val(Alltrim(Str(nRet, RECNO_FIELD_SIZE, 0)))

METHOD getLogLevel() CLASS rqFBRS AS IntVar
  Return ::iLogLevel

PROTECT METHOD getConnection() CLASS rqFBRS
  Return ::oConn

METHOD setLogLevel( level ) CLASS rqFBRS
  IF ::iLogLevel <> level
    ::iLogLevel := level
    // - - - Log
    ::writeLog( LOG_OFF, "[LOG#] rqFBRS loglevel now: " ;
                         + Alltrim(Str(::iLogLevel)))
  Endif
  Return

PROTECT METHOD setConnection( conn AS rqFBConn ) CLASS rqFBRS
  IF conn:isConnected()
    ::oConn := conn
    ::bConn := .T.
  Endif
  Return

METHOD isConnected() CLASS rqFBRS AS Logical
  Return ::bConn

PROTECT METHOD writeLog( level, msg ) CLASS rqFBRS
  wrtLog( level, msg, ::getLogLevel() )
  Return

METHOD Init( sStmt, cnx ) CLASS rqFBRS
  Local conn AS rqFBConn

  IF !Empty(cnx)
    ::setConnection( cnx )
  ElseIF ValType(m->oConn1) == "O"
    ::setConnection( m->oConn1 )
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] Connection to database not defined" )

    raiseError( "Connection to database not defined!" )
  Endif
  ::sStatement := sStmt
  IF ::isConnected()
    conn := ::getConnection()
    ::setLogLevel( conn:getLogLevel() ) // LogLevel da cnx
    ::bActive := ::openResultSet( ::sStatement )
    IF ::bActive

      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[RS] ResultSet created: " + ::sStatement ;
                  + " DB: " + conn:getDatabaseName() )
    Endif
  Endif
  Return Self

METHOD refresh() CLASS rqFBRS AS Logical
  IF ::isConnected()
    ::bActive := ::openResultSet( ::sStatement )
  Endif
  Return ::bActive

PROTECT METHOD openResultSet( stmt ) CLASS rqFBRS
  Local bRet    := .F.
  Local conn    := ::getConnection() AS rqFBConn
  Local aSelect := conn:executeQuery( stmt )
  Local aAux    := {}
  IF Len(aSelect) > 0
    aAux := conn:setRddFields( aSelect[1] )
    ::aStru := aAux[1]
    ::bLimboFields := aAux[2]
    IF ::bLimboFields 
      // - - - Log
      ::writeLog( LOG_OFF, "[WARN] Resultset with Limbo Fields!" )

      raiseError( "Resultset with Limbo Fields, probably created by an;" ;
                  + "unsuccessful; structure update" )
    Endif
    IF Len(aSelect) >= 2
      ::aFetchedRS := Array( Len(aSelect) - 1 )
      ::aFetchedRS := ACopy( aSelect, ::aFetchedRS, 2 )
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[WARN] Empty ResultSet: " + stmt + " DB: " ;
                  + conn:getDatabaseName() )
    Endif
    bRet := .T.
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] ResultSet failed: " + stmt + " DB: " ;
                + conn:getDatabaseName() )

    raiseError( "Erro no resultset: " + stmt + " ; DB: " ;
                + conn:getDatabaseName() )
  Endif
  Return bRet

METHOD promoteRStoAlias( als ) CLASS rqFBRS As Logical
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local nLen := 0
  Local iCt
  IF ::bActive
    Select 0
    ::sAlias := RDDannAlias( SELF, als )
    IF Empty( ::sAlias )
      // - - - Log
      ::writeLog( LOG_OFF, "[ERROR] Could not register Alias: " + ::sAlias )

      raiseError( "Could not register Alias: " + ::sAlias )
    Else
      ::aFieldVars := Array(Len(::aStru))
      For iCt := 1 to Len(::aStru) 
        IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
          Do Case
            Case ::aStru[iCt,DBS_TYPE] == "N"
              nLen := ::aStru[iCt,DBS_LEN] + ::aStru[iCt,DBS_DEC]
            Otherwise
              nLen := ::aStru[iCt,DBS_LEN]
          EndCase
          RDDannField( ::aStru[iCt,DBS_NAME], iCt, ; 
                       ::aStru[iCt,DBS_TYPE], 1, ;
                       nLen, ;
                       ::aStru[iCt,DBS_DEC ], ;
                       ::aFieldVars )
        Endif
      Next
      conn:announceAlias( ::sAlias )
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[PROMOTION] RS promoted to Als: " + ::sAlias;
                  + " Area: " + Alltrim(Str(Select())) )

      ::GoTop()
      bRet := .T.
    Endif
  Endif
  Return bRet

METHOD createTable( nm ) CLASS rqFBRS
  Local bRet   := .F.
  Local conn   := ::getConnection() AS rqFBConn
  Local aParam := {}
  Local sStmt
  Return .F.

PROTECT METHOD isFBDateTime( ln ) CLASS rqFBRS
  Return At(":", ln)

PROTECT METHOD translateFBDateTime( ln ) CLASS rqFBRS
  Local dtRet := NIL

  // - - - Log
  ::writeLog( LOG_LEVEL_09, "[DATE] FB DateTime translated: " + ln )

  ln := Substr(ln,1,At(" ",ln))
  ln := StrTran(ln,"-")
  dtRet := StoD(ln)
  Return dtRet

METHOD getFetchedSize() CLASS rqFBRS
  Return Len(::aFetchedRS)

METHOD beforeFirst() CLASS rqFBRS
  ::bBeforeFirst := .T.
  ::iActualRow   := 0
  Return

METHOD next() CLASS rqFBRS AS Logical
  Local bRet := .F.
  IF (::iActualRow + 1) <= Len(::aFetchedRS)
    ::iActualRow   += 1
    ::bBeforeFirst := .F.
    bRet := .T.
  Endif
  Return bRet

METHOD previous() CLASS rqFBRS AS Logical
  Local bRet := .F.
  IF (::iActualRow - 1) >= 1
    ::iActualRow   -= 1
    ::bBeforeFirst := .F.
    bRet := .T.
  Else
    ::bBeforeFirst := .T.
  Endif
  Return bRet

METHOD first() CLASS rqFBRS AS Logical
  Local bRet := .F.
  IF Len(::aFetchedRS) > 0
    ::iActualRow   := 1
    ::bBeforeFirst := .F.
    bRet := .T.
  Endif
  Return bRet

METHOD last() CLASS rqFBRS AS Logical
  Local bRet := .F.
  IF Len(::aFetchedRS) > 0
    ::iActualRow   := Len(::aFetchedRS)
    ::bBeforeFirst := .F.
    bRet := .T.
  Endif
  Return bRet

PROTECT METHOD getColumnName( colpos ) CLASS rqFBRS
  Return ::aStru[colpos,DBS_NAME]

PROTECT METHOD getColumnPos( colname ) CLASS rqFBRS
  Local sColName := Alltrim(Upper(colname))
  Return AScan(::aStru, {|x| sColName == Alltrim(Upper(x[DBS_NAME]))})

PROTECT METHOD getColumnType( colpos ) CLASS rqFBRS AS Character
  Local sRet := ""
  IF colpos > 0
    sRet := ::aStru[colpos][DBS_TYPE]
  Endif
  Return sRet

PROTECT METHOD getColumnLength( colpos ) CLASS rqFBRS
  Local nRet := 0
  IF colpos > 0
    Do Case
      Case ::aStru[colpos][DBS_TYPE] == "N"
        nRet := ::aStru[colpos][DBS_LEN] + ::aStru[colpos][DBS_DEC]
      Otherwise
        nRet := ::aStru[colpos][DBS_LEN]
    EndCase
  Endif
  Return nRet

PROTECT METHOD getColumnValue( colpos ) CLASS rqFBRS As Character
  Local sRet := ""
  IF colpos > 0 .and. !::bBeforeFirst
    sRet := ::aFetchedRS[::iActualRow][colpos]
  Endif
  Return sRet
 
PROTECT METHOD translateValue( colname, typeout ) CLASS rqFBRS
  Local uRet     := NIL
  Local iColPos  := 0
  Local sColValue
  Local iTypeIn
  IF !::bBeforeFirst
    iColPos   := ::getColumnPos( colname )
    iTypeIn   := ::getColumnType( iColPos )
    sColValue := ::getColumnValue( iColPos )
    IF sColValue != NULL_COLUMN
      Do Case
        Case (iTypeIn == "C" .and. typeout == "D") .or. ;
             (iTypeIn == "D" .and. typeout == "D")
          IF ::isFBDateTime(sColValue)
            uRet := ::translateFBDateTime(sColValue)
          Else
            uRet := CtoD(sColValue)
          Endif
        Case (iTypeIn == "C" .and. typeout == "C") .or. ;
             (iTypeIn == "D" .and. typeout == "C") .or. ;
             (iTypeIn == "N" .and. typeout == "C")
          uRet := ::getColumnValue( iColPos )
        Case (iTypeIn == "C" .and. typeout == "N") .or. ;
             (iTypeIn == "N" .and. typeout == "N")
          uRet := Val(::getColumnValue(iColPos))
      EndCase
    Endif
  Endif
  Return uRet

METHOD getString( colname ) CLASS rqFBRS AS Character
  Local sRet := Space(::getColumnLength(::getColumnPos(colname)))
  Local uAux := ::translateValue( colname, "C" )
  IF !Empty(uAux)
    sRet := uAux
  Endif
  Return sRet

METHOD getDate( colname ) CLASS rqFBRS
  Local dtRet := CtoD("")
  Local uAux  := ::translateValue( colname, "D" )
  IF !Empty(uAux)
    dtRet := uAux
  Endif
  Return dtRet

METHOD getNumber( colname ) CLASS rqFBRS
  Local fRet := 0
  Local uAux := ::translateValue( colname, "N" )
  IF !Empty(uAux)
    fRet := uAux
  Endif
  Return fRet

METHOD getValue( colname ) CLASS rqFBRS
  Local uRet     := NIL
  Local iColPos  := ::getColumnPos( colname )
  Local sColType := ::getColumnType( iColPos )
  IF ::isRSpromotedToAlias() .and. ::bEof
    Do Case
      Case sColType == "C"
        uRet := Space(::getColumnLength(iColPos))
      Case sColType == "D"
        uRet := CtoD("")
      Case sColType == "N"
        uRet := 0
      Otherwise
        uRet := NIL
    EndCase
  Else
    Do Case
      Case sColType == "C"
        uRet := ::getString(colname)
      Case sColType == "D"
        uRet := ::getDate(colname)
      Case sColType == "N"
        uRet := ::getNumber(colname)
      Otherwise
        uRet := NIL
    EndCase
  Endif
  Return uRet

METHOD isNull( colname ) CLASS rqFBRS AS Logical
  Local bRet := .F.
  IF Empty(::getValue(colname)) 
    bRet := .T.
  Endif
  Return bRet

METHOD isActive() CLASS rqFBRS AS Logical
  Return ::bActive

METHOD isRSpromotedToAlias() CLASS rqFBRS AS Logical
  Return !Empty(::sAlias)

METHOD DbStruct() CLASS rqFBRS
  Local aRet := {}
  Local iCt
  Local nLen
  For iCt := 1 To Len(::aStru)
    IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
      Do Case
        Case ::aStru[iCt,DBS_TYPE] == "N"
          nLen := ::aStru[iCt,DBS_LEN] + ::aStru[iCt,DBS_DEC]
        Otherwise
          nLen := ::aStru[iCt,DBS_LEN]
      EndCase
      Aadd(aRet, {::aStru[iCt,DBS_NAME], ::aStru[iCt,DBS_TYPE], ;
                  nLen,  ::aStru[iCt,DBS_DEC]} )
    Endif
  Next
  Return aRet

METHOD GoTop() CLASS rqFBRS
  Local bRet := .F.
  IF ::first()
    ::bEof := ::bBof := .F.
    bRet := .T.
  Else
    ::bEof := ::bBof := .T.
  Endif
  Return bRet 

METHOD GoBottom() CLASS rqFBRS
  Local bRet := .F.
  IF ::last()
    ::bEof := ::bBof := .F.
    bRet := .T.
  Else
    ::bEof := ::bBof := .T.
  Endif
  Return bRet 

METHOD Goto( vl ) CLASS rqFBRS
  Local bRet := .F.
  IF vl >= 1 .and. vl <= Len(::aFetchedRS)
    ::bEof := ::bBof := .F.
    ::iActualRow   := vl
    ::bBeforeFirst := .F.
    bRet := .T.
  Else
    ::bEof := ::bBof := .T.
  Endif
  Return bRet

METHOD RLock( vl ) CLASS rqFBRS AS Logical
  Return .T.
METHOD FLock() CLASS rqFBRS AS Logical
  Return .T.
METHOD UnLock( vl ) CLASS rqFBRS AS Logical
  Return .T.
METHOD Append( p1 ) CLASS rqFBRS AS Logical
  Return .T.

METHOD FieldName( fld_nro ) CLASS rqFBRS
  Return ::getColumnName( fld_nro )

METHOD FieldPos( fld_name ) CLASS rqFBRS
  Return ::getColumnPos( fld_name )

METHOD FieldInfo( info, fld_nroname ) CLASS rqFBRS
  Local uValue   := NIL
  Local iFieldNo := 0
  IF ::bActive
    IF ValType(fld_nroname) == "C"
      iFieldNo := ::FieldPos( fld_nroname )
    ElseIF ValType(fld_nroname) $ "NI"
      iFieldNo := fld_nroname
    Endif
    IF iFieldNo > 0
      Do Case
        Case info == DBS_LEN
          uValue := ::getColumnLength(iFieldNo)
        Otherwise 
          uValue := ::aStru[iFieldNo][info]
      EndCase
    Endif
  Endif
  Return uValue

METHOD FieldPut( fld_nroname, value ) CLASS rqFBRS
  // - - - Log
  ::writeLog( LOG_OFF, "[ERROR] Trying to modify a readonly ResultSet! Als:" ;
                       + ::sAlias )

  raiseError( "Cannot modify a readonly ResultSet. Als: " + ::sAlias )
  Return .F.

METHOD FieldGet( fld_nroname ) CLASS rqFBRS
  Local uValue
  Local iFieldNo := 0
  IF ::bActive
    IF ValType(fld_nroname) == "C"
      iFieldNo := ::FieldPos( fld_nroname )
    ElseIF ValType(fld_nroname) $ "NI"
      iFieldNo := fld_nroname
    Endif
    // - - - 
    IF iFieldNo > 0
      uValue := ::getValue( ::FieldName(iFieldNo) )
      AssignFldValue(@aFieldVars[iFieldNo], uValue)

      // - - - Log
      ::writeLog( LOG_LEVEL_09, "[FGET] FieldGet: " + ::FieldName(iFieldNo) ;
                    + " T: " + ValType(uValue) + " V: " ;
                    + IIF(ValType(uValue)=="N",Alltrim(Str(uValue)),"") ;
                    + IIF(ValType(uValue)=="D",DtoC(uValue),"") ;
                    + IIF(ValType(uValue)=="C",uValue,"") )

      Return @aFieldVars[iFieldNo]
    Endif
  Endif
  Return NIL

METHOD Commit() CLASS rqFBRS AS Logical
  Return .T.

METHOD Skip( vl ) CLASS rqFBRS
  Local iCount := IIF(vl == NIL, 1, vl)
  Local conn   := ::getConnection() AS rqFBConn
  Local iCt

  IF ::bEof .and. ::bBof
    Return .F.
  Endif
  IF ::bEof .and. iCount < 0
    ::GoBottom()
    iCount += 1
  Endif

  IF iCount > 0 .and. !::bEof                          // - - - Forward
    ::bEof := ::bBof := .F.
    For iCt := 1 To iCount
      IF !::next()
        ::bEof := .T.
        Exit
      Endif
    Next
  ElseIF iCount < 0 .and. !::bBof                      // - - - Backwards
    ::bEof := ::bBof := .F.
    For iCt := 1 To Abs(iCount)
      IF !::previous()
        ::bBof := .T.
        Exit
      Endif
    Next
  Endif
  Return .T.

PROTECT METHOD closeRS() CLASS rqFBRS
  Local conn := ::getConnection() AS rqFBConn
  Local iCt
  IF ::bActive
    ::bActive := .F.
    ::bConn   := .F.

    // - - - Log
    ::writeLog( LOG_LEVEL_07, "[CLOS] ResultSet closed. Als: " + ::sAlias )
  
    IF ::isRSpromotedToAlias()
      conn:removeAlias( ::sAlias )
      For iCt := 1 to Len(::aStru)
        IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
          RDDretField( SELF, ::aStru[iCt,DBS_NAME] )
        Endif
      Next
      RDDretAlias( SELF )
    Endif
  Endif
  Return .T.

METHOD Axit() CLASS rqFBRS
  Return .T.

METHOD NoMethod( meth ) CLASS rqFBRS
  // - - - Log
  ::writeLog( LOG_OFF, "[ERROR] Meth: " + meth + " does not exist inside: ";
              + "rqFBRS class" )

  raiseError( "This class has no Method named: " + meth )
  Return

METHOD Close() CLASS rqFBRS
  ::closeRS()
  Return .T.

Static Function raiseError( msg )
  Local oErr := Error{} AS Error
  oErr:DESCRIPTION := "[FBRS Error] ;" + msg
  oErr:GENCODE     := -1
  oErr:SEVERITY    := ES_RTERR
  Eval(ErrorBlock(),oErr)
  Return

Static Function wrtLog( level, msg, actuallevel )
  Local iAux := actuallevel
#Cinline
{
  int level    = VAR_INT(VAR_NAME_LOCPAR(level));
  char *msg    = VAR_CHR(VAR_NAME_LOCPAR(msg));
  int loglevel = VAR_INT(VAR_NAME_LOCAL(iaux));

  if( level <= loglevel )
  {
    FILE *fp;
    long time(), now;
    char errfile[100];
    time( &now );
    sprintf( errfile, "/var/log/rqrddfire.log" );
    fp = fopen( errfile, "a" );
    if( fp != NULL )
    {
      fprintf( fp, "\t[%i] %s ", level, getenv("USER") );
      fprintf( fp, "%s ", ctime(&now) );
      fprintf( fp, "%s\n", msg );
      fclose( fp );
    }
  }
}
#endCinline
  Return
