/*
RqRDDFire - A Free FlagShip RDD for Firebird RDBMS
v.0.9.0, September 2007, Brazil
Ricardo Delamar Roque (rroque6428@terra.com.br)

License
=======
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed AS IS in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/licenses/lgpl-2.1.html
*/

#Cinline
#include <FSopenc.h>
#endCinline

#include "rqfbconn.fh"
#include "dbstruct.fh" 
#include "dataserv.fh"
#include "error.fh"
#include "rddsys.fh"
#include "rqrdd.h"

#define PATH_SEPARATOR          "/"
#define EXTENSION_SEPARATOR     "."
#define NULL_FB_FIELD           "__NULL_FB_FIELD__" 

#ifdef FlagShip5
#define NOPIXEL NoPixel
#else
#define NOPIXEL
#endif

#define RDDIDX_NAME             1
#define RDDIDX_FIELDS           2
#define RDDIDX_EXPRESSION       3

memvar oConn1

CLASS rqRDDFire INHERIT DataServer
  PROTECT sAlias           := ""                       AS Character
  PROTECT sTableName       := ""                       AS Character
  PROTECT sAreaID          := ""                       AS Character
  PROTECT sFilter          := ""                       AS Character
  PROTECT blkFilter        := {|| .T.}                 AS Codeblock
  PROTECT bFiltered        := .F.                      AS Logical
  PROTECT bUsed            := .F.                      AS Logical
  PROTECT bShared          := .T.                      AS Logical
  PROTECT bReadOnly        := .F.                      AS Logical
  PROTECT bEof             := .T.                      AS Logical
  PROTECT bBof             := .T.                      AS Logical
  PROTECT bFound           := .F.                      AS Logical
  PROTECT bAppendPending   := .F.                      AS Logical
  PROTECT bDeleted         := .F.                      AS Logical
  PROTECT bConn            := .F.                      AS Logical
  PROTECT bCachingAppends  := .F.                      AS Logical
  PROTECT bOwnerOfLocks    := .F.                      AS Logical
  PROTECT aStru            := {}                       AS Array
  PROTECT aFieldVars       := {}                       AS Array
  PROTECT aActualRecord    := {}                       AS Array
  PROTECT sNextRecord      := ""                       AS Character
  PROTECT aFieldChanged    := {}                       AS Array
  PROTECT aIndexes         := {}                       AS Array
  PROTECT aCachedAppends   := {}                       AS Array
  PROTECT aSkCacheAsc      := {}                       AS Array
  PROTECT aSkCacheDesc     := {}                       AS Array
  PROTECT iSkipMode        := SKIP_MODE_1              AS IntVar
  PROTECT iSkCacheSize     := DEFAULT_SKCACHE_SIZE     AS IntVar
  PROTECT iSkCacheAscAge   := 0.0                      AS Numeric
  PROTECT iSkCacheDescAge  := 0.0                      AS Numeric
  PROTECT iSkCacheTimeOut  := DEFAULT_SKCACHE_TMOUT    AS Numeric
  PROTECT aSeekCache       := {}                       AS Array
  PROTECT iSeekChTmOut     := DEFAULT_SEEKCACHE_TMOUT  AS Numeric
  PROTECT iSeekChAge       := 0.0                      AS Numeric
  PROTECT iRecCount        := 0                        AS IntVar
  PROTECT iRecCountTimeOut := DEFAULT_RECCOUNT_TMOUT   AS Numeric
  PROTECT iRecCountAge     := 0.0                      AS Numeric
  PROTECT iActualOrder     := 0                        AS IntVar
  PROTECT iLogLevel        := LOG_OFF                  AS IntVar
  PROTECT bLimboFields     := .F.                      AS Logical
  PROTECT bTmpTable        := .F.                      AS Logical
  PROTECT iFlushCounter    := 0                        AS IntVar
  PROTECT oConn          

PROTOTYPE PROTECT METHOD getRecordBuffer(p1)        CLASS rqRDDFire
PROTOTYPE PROTECT METHOD writeLog(p1,p2)            CLASS rqRDDFire
PROTOTYPE METHOD FieldPos(p1)                       CLASS rqRDDFire 
PROTOTYPE PROTECT METHOD closeRdd()                 CLASS rqRDDFire
PROTOTYPE PROTECT METHOD openTable()                CLASS rqRDDFire
PROTOTYPE METHOD GoTop()                            CLASS rqRDDFire
PROTOTYPE METHOD IndexOrd()                         CLASS rqRDDFire
PROTOTYPE METHOD IndexKey(p1)                       CLASS rqRDDFire
PROTOTYPE METHOD DbStruct()                         CLASS rqRDDFire
PROTOTYPE PROTECT METHOD reccountTable()            CLASS rqRDDFire
PROTOTYPE PROTECT METHOD getConnection()            CLASS rqRDDFire
PROTOTYPE PROTECT METHOD clearRecordBuffer()        CLASS rqRDDFire
PROTOTYPE PROTECT METHOD getBufferRecno()           CLASS rqRDDFire
PROTOTYPE PROTECT METHOD setTableUsedbyThisArea(p1) CLASS rqRDDFire
PROTOTYPE PROTECT METHOD clearAllCaches()           CLASS rqRDDFire

INIT Procedure _rqRDDFire  
   // - - - Log
   wrtLog( LOG_OFF, "[RDD] A new Rdd named: " + RDD_DRIVER_NAME + " have ";
           + "been announced.", LOG_OFF )
   AnnNewRdd( RDD_DRIVER_NAME ) 
   Return

ACCESS Used CLASS rqRDDFire
  Return ::bUsed
ACCESS Shared CLASS rqRDDFire
  Return ::bShared
ACCESS ReadOnly CLASS rqRDDFire
  Return ::bReadOnly
ACCESS Alias CLASS rqRDDFire
  Return ::sAlias
ACCESS Eof CLASS rqRDDFire
  Return ::bEof
ACCESS Bof CLASS rqRDDFire
  Return ::bBof
ACCESS Found CLASS rqRDDFire
  Return ::bFound
ACCESS FCount CLASS rqRDDFire
  Return Len(::DbStruct()) 
ACCESS IndexExt CLASS rqRDDFire
  Return ".fbx"
ACCESS IndexKey CLASS rqRDDFire
  Return ::IndexKey(0)
ACCESS IndexCount CLASS rqRDDFire
  Return Len(::aIndexes)
ACCESS Driver CLASS rqRDDFire
  Return RDD_DRIVER_NAME 
ACCESS RddName CLASS rqRDDFire
  Return ::Driver 
ACCESS Name CLASS rqRDDFire
  Return ::sTableName
ACCESS LastRec CLASS rqRDDFire
  Return ::reccountTable()
ACCESS Deleted CLASS rqRDDFire
  Return ::bDeleted
ACCESS Recno CLASS rqRDDFire
  Local nRet := 0
  IF ::bUsed
    IF ::bEof
      nRet := ::reccountTable() + 1
    ElseIF ::bBof
      nRet := 1
    Else
      nRet := ::getBufferRecno()
    Endif
  Endif
  Return Val(Alltrim(Str(nRet, RECNO_FIELD_SIZE, 0)))
ACCESS RlockList() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Return conn:getLockedRecords( ::sTableName, ::sAreaID )
ACCESS Filter CLASS rqRDDFire
  Return ::sFilter

// Dummy, unused access methods. Needed in FS6
ACCESS IsRelation CLASS rqRDDFire 
  Return .F.
ACCESS FileSpec CLASS rqRDDFire
  Local aux := ::Name
  Return ::Name
ACCESS AliasSym CLASS rqRDDFire
   return NIL
ACCESS Concurrenc CLASS rqRDDFire
   return NIL
ACCESS ErrInfo CLASS rqRDDFire
   return NIL
ACCESS ForBlock CLASS rqRDDFire
   return NIL
ACCESS GetLocate CLASS rqRDDFire
   return NIL
ACCESS Header CLASS rqRDDFire
   return NIL
ACCESS IndexCheck CLASS rqRDDFire
   return 0
ACCESS IndexLock CLASS rqRDDFire
   return NIL
ACCESS Lupdate CLASS rqRDDFire
   return .F.
ACCESS OrderBottom CLASS rqRDDFire
   return .F.
ACCESS OrderKeyNo CLASS rqRDDFire
   return 0
ACCESS OrderKeyVal CLASS rqRDDFire
   return NIL
ACCESS OrderTopSc CLASS rqRDDFire
   return NIL
ACCESS Scope CLASS rqRDDFire
   return NIL
ACCESS Status CLASS rqRDDFire
   return NIL
ACCESS WhileBlock CLASS rqRDDFire
   return NIL

ASSIGN IsRelation( vl ) CLASS rqRDDFire    // ??? Not Documented
  Return .F.
ASSIGN Filter( vl ) CLASS rqRDDFire
  ::sFilter := vl 
  Return ::sFilter
ASSIGN Alias( vl ) CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local cAlias

  IF Valtype( vl ) == "C"
    cAlias := RDDannAlias( SELF, vl ) 

    // - - - Log
    ::writeLog( LOG_LEVEL_07, "[ALIAS] Alias change from: " + ::sAlias ; 
                + " to: " + cAlias )

    IF Empty( cAlias )
      // - - - Log
      ::writeLog( LOG_OFF, "[ERROR] Could not register Alias: " + ::sAlias )

      raiseError( "Could not register Alias: " + ::sAlias )
    Else
      conn:removeAlias( ::sAlias )
      ::sAlias := cAlias
      conn:announceAlias( ::sAlias )
    Endif
  Endif
  Return ::sAlias

METHOD setLogLevel( level ) CLASS rqRDDFire 
  IF ::iLogLevel <> level
    ::iLogLevel := level
    IF !Empty(::sAlias) 
      // - - - Log
      ::writeLog( LOG_OFF, "[LOG#] rqRDD loglevel now: " ;
                  + Alltrim(Str(::iLogLevel)) + " for Alias: " + ::sAlias )
    Endif
  Endif
  Return
METHOD setSkipCacheSize( sz ) CLASS rqRDDFire 
  IF Empty(sz) .or. sz < 0
    ::iSkCacheSize := DEFAULT_SKCACHE_SIZE
  Else
    ::iSkCacheSize := sz
  Endif
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] SKIPCache Size: " ;
                            + Alltrim(Str(::iSkCacheSize)) + " Als: " ;
                            + ::sAlias )
  Return
METHOD setSkipCacheTimeOut( vl ) CLASS rqRDDFire 
  IF Empty(vl) .or. vl < 0 
    ::iSkCacheTimeOut := DEFAULT_SKCACHE_TMOUT
  Else
    ::iSkCacheTimeOut := vl
  Endif
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] SKIPCache Timeout: " ;
                            + Alltrim(Str(::iSkCacheTimeOut)) + "s Als: ";
                            + ::sAlias )
  Return
METHOD setSeekCacheTimeOut( vl ) CLASS rqRDDFire 
  IF Empty(vl) .or. vl < 0 
    ::iSeekChTmOut := DEFAULT_SKCACHE_TMOUT
  Else
    ::iSeekChTmOut := vl
  Endif
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] SEEKCache Timeout: " ;
                            + Alltrim(Str(::iSeekChTmOut)) + "s Als: " ;
                            + ::sAlias )
  Return
METHOD setRecCountTimeOut( vl ) CLASS rqRDDFire 
  IF Empty(vl) .or. vl < 0 
    ::iRecCountTimeOut := DEFAULT_RECCOUNT_TMOUT
  Else
    ::iRecCountTimeOut := vl
  Endif
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] RecCount Timeout: " ;
                            + Alltrim(Str(::iRecCountTimeOut)) + "s Als: ";
                            + ::sAlias )
  Return
METHOD setSkipMode( vl ) CLASS rqRDDFire 
  IF vl == NIL .or. vl < 0 .or. vl > SKIP_MODE_2
    ::iSkipMode := SKIP_MODE_1
  Else
    ::iSkipMode := vl
  Endif
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] SkipMode: " ;
                            + Alltrim(Str(::iSkipMode)) + " Als: " ;
                            + ::sAlias )
  Return
METHOD startCachedAppends() CLASS rqRDDFire 
  ::bCachingAppends := .T.
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[SET] Caching Appends. Als: " + ::sAlias )
  Return

METHOD getLogLevel() CLASS rqRDDFire AS IntVar
  Return ::iLogLevel
METHOD getSkipMode() CLASS rqRDDFire AS IntVar
  Return ::iSkipMode
METHOD getSkipCacheSize() CLASS rqRDDFire AS IntVar
  Return ::iSkCacheSize
METHOD getSkipCacheTimeOut() CLASS rqRDDFire
  Return ::iSkCacheTimeOut
METHOD getSeekCacheTimeOut() CLASS rqRDDFire
  Return ::iSeekChTmOut
METHOD getRecCountTimeOut() CLASS rqRDDFire
  Return ::iRecCountTimeOut

PROTECT METHOD writeLog( level, msg ) CLASS rqRDDFire
  wrtLog( level, msg, ::getLogLevel() )
  Return

PROTECT METHOD translateLineToArray( ln, separador ) CLASS rqRDDFire
  Local aRet    := {}
  Local bAddNIL := .F.
  Local sVl
  Local iPos
  IF ValType(ln) == "C" .and. ValType(separador) == "C"
    IF Right(ln, Len(separador)) == separador
      bAddNIL := .T.
    Endif
    Do While At(separador,ln) > 0
      iPos := At(separador,ln)
      sVl  := Substr(ln,1,iPos - 1)
      AAdd(aRet, sVl)
      ln := Substr(ln, iPos + Len(separador), Len(ln))
    Enddo
    IF Len(ln) > 0
      AAdd(aRet, ln)
    Endif
    IF bAddNIL
      AAdd(aRet, "")
    Endif
  Endif
  Return aRet

METHOD setDropTableOnClose( vl ) CLASS rqRDDFire
  IF !Empty(vl) .and. ValType(vl) == "L"
    ::bTmpTable := vl
  Endif
  Return

PROTECT METHOD setConnection( conn AS rqFBConn ) CLASS rqRDDFire
  IF conn:isConnected()
    ::oConn := conn
    ::bConn := .T.
  Endif
  Return

PROTECT METHOD setRecordChanged( fld_nro, bol ) CLASS rqRDDFire
  Local iCt
  IF fld_nro > 0 
    ::aFieldChanged[fld_nro] := bol
  Else
    // - - - Inicializa matriz que aponta mudancas nos campos
    IF Empty(::aFieldChanged)
      ::aFieldChange := Array(::FCount)
    Endif
    For iCt := 1 To ::FCount
      ::aFieldChanged[iCt] := bol
    Next
  Endif
  Return

PROTECT METHOD isRecordChanged() CLASS rqRDDFire AS Logical
  Return AScan(::aFieldChanged, {|x| x == .T.}) != 0

PROTECT METHOD isTemporaryTable() CLASS rqRDDFire
  Return ::bTmpTable

PROTECT METHOD getConnection() CLASS rqRDDFire
  Return ::oConn

METHOD isConnected() CLASS rqRDDFire AS Logical
  Local conn := ::getConnection() AS rqFBConn
  IF conn:isConnected()
    ::bConn := .T.
  Else
    ::bConn := .F.
  Endif
  Return ::bConn

METHOD getDatabaseName() CLASS rqRDDFire AS Character
  Local conn := ::getConnection() AS rqFBConn
  Local sRet := ""
  IF ::bUsed
    sRet := conn:getDatabaseName()
  Endif
  Return sRet

PROTECT METHOD getFileName( f ) CLASS rqRDDFire AS Character
  Return Substr(f, Rat(PATH_SEPARATOR, f) + 1)

PROTECT METHOD removeFileNameExt( f ) CLASS rqRDDFire AS Character
  Local sRet := ""
  IF At(EXTENSION_SEPARATOR, f) > 0
    sRet := Left(f, At(EXTENSION_SEPARATOR, f) - 1)
  Else
    sRet := f
  Endif
  Return sRet

METHOD Init( sTbl, bShare, bReadOnly, any1, any2, bNew, cnx ) CLASS rqRDDFire
  Local iCt
  Local conn AS rqFBConn

  bReadOnly := IIF(Empty(bReadOnly),.F.,bReadOnly)

  IF !Empty(cnx) 
    ::setConnection( cnx )
  ElseIF ValType(m->oConn1) == "O"
    ::setConnection( m->oConn1 )
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] Connection to database not defined" )
    
    raiseError( "Choose a Firebird connection to use with the RDD." )
  Endif
  IF ::isConnected()
    conn := ::getConnection()
    ::setLogLevel( conn:getLogLevel() ) // LogLevel da cnx

    // - - - Log
    ::writeLog( LOG_LEVEL_06, "[INIT] Incoming RQRDD object request creation.";
                + " Table:" + IIF(Empty(sTbl),"NIL",sTbl) + " New:" ;
                + IIF(bNew,".T.",".F.") )
    IF bNew
      Select 0
    Else
      IF ::bUsed 
        ::closeRDD()
      Endif
    Endif
    IF ValType(sTbl) != "C" .or. Empty(sTbl)
      ::bUsed := .F.
    Else
      sTbl := ::removeFileNameExt(::getFileName(sTbl))
      IF bShare == NIL
        ::bShared := !Set(_SET_EXCLUSIVE)
        // - - - Log
        ::writeLog( LOG_LEVEL_06, "[WARN] Undefined Open Table Mode. Assuming";
                    + IIF(::bShared," SHARED MODE"," EXCLUSIVE MODE") ;
                    + " from SET EXCLUSIVE command." )
      Else
        ::bShared := bShare
      Endif
      ::bReadOnly  := bReadOnly  // vd. TODO list
      ::sTableName := sTbl
      ::sAreaID := conn:getUniqueAreaID(::sTableName, Alltrim(Str(Select())))
      ::bUsed   := ::openTable() .and. ::setTableUsedbyThisArea( .T. )
      IF ::bUsed
        ::SetRecordChanged( 0, .F. )
        ::sAlias := RDDannAlias( SELF, ::sTableName )
        IF Empty( ::sAlias )

          // - - - Log
          ::writeLog( LOG_OFF, "[ERROR] Could not register Alias: " + ::sAlias )

          raiseError( "Could not register Alias: " + ::sAlias )
          ::closeRDD()
        Else
          ::aFieldVars := Array(Len(::aStru))
          For iCt := 1 to Len(::aStru) 
            IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
              RDDannField( ::aStru[iCt,DBS_NAME], iCt, ; 
                           ::aStru[iCt,DBS_TYPE], 1, ;
                           ::aStru[iCt,DBS_LEN ], ;
                           ::aStru[iCt,DBS_DEC ], ;
                           ::aFieldVars )
            Endif
          Next
          conn:announceAlias( ::sAlias )
          ::setSkipMode( conn:getGlbSkipMode() )
          ::setSkipCacheSize( conn:getGlbSkipCacheSize() )
          ::clearRecordBuffer()
        Endif
      Endif

      IF ::bUsed
        NetErr( .F. )

        // - - - Log
        ::writeLog( LOG_LEVEL_07, "[TABL] Table opened. Table:" + ::sTableName;
                    + " Area:" + Alltrim(Str(Select())) + " Alias:" ;
                    + ::sAlias + IIF(::bShared," SHARED"," EXCLUSIVE") ;
                    + " DB:" + conn:getDatabaseName() )
        ::GoTop()
      Else
        NetErr( .T. )

        // - - - Log
        ::writeLog( LOG_OFF, "[ERROR] Open Failed Table:" + ::sTableName ;
                    + " ReadOnly:" + IIF(bReadOnly,".T.",".F.") + " New:" ;
                    + IIF(bNew,".T.",".F.") + " Shared:" ;
                    + IIF(::bShared,".T.",".F.") + " DB: " ;
                    + conn:getDatabaseName() )
      Endif
    Endif
  Endif
  Return

METHOD IndexOrd() CLASS rqRDDFire
  Return ::iActualOrder

METHOD IndexKey( ind_nro ) CLASS rqRDDFire
  Local sRet := ""
  IF Empty(ind_nro)
    ind_nro := 0
  Endif
  IF ind_nro > 0 .and. ind_nro <= Len(::aIndexes)
    sRet := ::aIndexes[ind_nro,RDDIDX_EXPRESSION]
  Else
    IF ::IndexOrd() > 0
      sRet := ::aIndexes[::IndexOrd(),RDDIDX_EXPRESSION]
    Endif
  Endif
  Return sRet  

METHOD FieldPos( fld_name ) CLASS rqRDDFire 
  Return AScan(::aStru, {|x| Alltrim(Upper(fld_name)) == x[DBS_NAME]})
 
METHOD FieldName( fld_nro ) CLASS rqRDDFire  
  Return ::aStru[fld_nro,DBS_NAME]

METHOD FieldInfo( info, fld_nroname ) CLASS rqRDDFire
  Local uValue   := NIL
  Local iFieldNo := 0
  IF ::bUsed
    IF ValType(fld_nroname) == "C"
      //iFieldNo := ::FieldPos( Left(Upper(Alltrim(fld_nroname)), 10) )
      iFieldNo := ::FieldPos( fld_nroname )
    ElseIF ValType(fld_nroname) $ "NI"
      iFieldNo := fld_nroname
    Endif
    IF iFieldNo > 0
      uValue := ::aStru[iFieldNo,info]
    Endif
  Endif
  Return uValue    

PROTECT METHOD setRecordBuffer( fld_nro, value ) CLASS rqRDDFire AS Logical
  Local bRet := .F.
  Local sFieldType
  Local sValue
  sFieldType := ValType(value) 
  IF fld_nro > 0 .and. sFieldType == ::aStru[fld_nro,DBS_TYPE]
    Do Case 
      Case sFieldType == "C"
        sValue := Padr(value, ::aStru[fld_nro,DBS_LEN])
      Case sFieldType == "N"
        sValue := Str(value, ::aStru[fld_nro,DBS_LEN], ::aStru[fld_nro,DBS_DEC])
      Case sFieldType == "D"
        sValue := DtoC(value)
      Case sFieldType == "L"
        sValue := IIF(value, "T", "F")
      Otherwise
        sValue := NIL
    EndCase
    IF sValue != NIL
      bRet := .T.
      IF ::aActualRecord[fld_nro] != sValue
        ::aActualRecord[fld_nro] := sValue
        ::setRecordChanged( fld_nro, .T. )
      Endif
    Endif
  Endif
  Return bRet

PROTECT METHOD isFBDateTime( ln ) CLASS rqRDDFire
  Return At(":", ln )

PROTECT METHOD translateFBDateTime( ln ) CLASS rqRDDFire
  Local dtRet := NIL

  // - - - Log
  ::writeLog( LOG_LEVEL_09, "[DATE] FB DateTime translated: " + ln )

  ln := Substr(ln,1,At(" ",ln))
  ln := StrTran(ln,"-")
  dtRet := StoD(ln)
  Return dtRet

PROTECT METHOD clearRecordBuffer() CLASS rqRDDFire
  Local sFieldType
  Local sValue
  Local iCt

  IF Empty(::aActualRecord)  
    ::aActualRecord := Array(Len(::aStru))
  Endif
  //For iCt := 1 To Len(::aStru)
  For iCt := 1 To ::FCount
    sFieldType := ::aStru[iCt,DBS_TYPE]
    Do Case 
      Case sFieldType == "C"
        sValue := Space(::aStru[iCt,DBS_LEN])
      Case sFieldType == "N"
        sValue := Str(0, ::aStru[iCt,DBS_LEN], ::aStru[iCt,DBS_DEC])
      Case sFieldType == "D"
        sValue := ""
      Case sFieldType == "L"
        sValue := "F"
      Otherwise
        sValue := NIL
    EndCase
    ::aActualRecord[iCt] := sValue 
  Next
  Return

PROTECT METHOD getRecordBuffer( fld_nro ) CLASS rqRDDFire
  Local uRet := NIL
  Local sFieldType
  Local sValue
  Local uValue
  IF fld_nro > 0
    IF ::bEof
      sFieldType := ::aStru[fld_nro,DBS_TYPE]
      Do Case 
        Case sFieldType == "C"
          uValue := Space(::aStru[fld_nro,DBS_LEN])
        Case sFieldType == "N"
          uValue := 0
        Case sFieldType == "D"
          uValue := CtoD(Space(8))
        Case sFieldType == "L"
          uValue := .F.
        Otherwise 
          uValue := NIL
      EndCase
      uRet := uValue
    Else
      sFieldType := ::aStru[fld_nro,DBS_TYPE]
      sValue     := ::aActualRecord[fld_nro]
      Do Case 
        Case sFieldType == "C"
          uValue := Padr(sValue, ::aStru[fld_nro,DBS_LEN])
        Case sFieldType == "N"
          uValue := Val(sValue)
        Case sFieldType == "D"
          IF ::isFBDateTime(sValue)
            uValue := ::translateFBDateTime(sValue)
          Else
            uValue := CtoD(sValue)
          Endif
        Case sFieldType == "L"
          uValue := IIF(sValue == "T", .T., .F.)
        Otherwise 
          uValue := NIL
      EndCase
      uRet := uValue
    Endif
  Endif
  Return uRet

PROTECT METHOD getBufferRecno() CLASS rqRDDFire
  Local nRet := ::getRecordBuffer( ::FieldPos(RECNO_FIELD) )
  Return Val(Alltrim(Str(nRet, RECNO_FIELD_SIZE, 0)))

METHOD FieldPut( fld_nroname, value ) CLASS rqRDDFire
  Local bRet     := .F.
  Local conn     := ::getConnection() AS rqFBConn
  Local iFieldNo := 0
  IF ::bUsed
    IF ValType(fld_nroname) == "C"
      iFieldNo := ::FieldPos( Left(fld_nroname,10) )
    ElseIF ValType(fld_nroname) $ "NI"
      iFieldNo := fld_nroname
    Endif
    // - - - 
    IF iFieldNo > 0 ;
       .and. ::aStru[iFieldNo,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL

      // - - - Log
      ::writeLog( LOG_LEVEL_09, "[FPUT] FieldPut: " + ::FieldName(iFieldNo) ;
                  + " T: " + ValType(value) + " V: " ;
                  + IIF(ValType(value)=="N",Alltrim(Str(value)),"") ;
                  + IIF(ValType(value)=="D",DtoC(value),"") ;
                  + IIF(ValType(value)=="C",value,"") )

      IF ::setRecordBuffer( iFieldNo, value )
        AssignFldValue( @aFieldVars[iFieldNo], value )
        bRet := .T.
      Endif
    Endif
  Endif
  Return bRet

METHOD FieldGet( fld_nroname ) CLASS rqRDDFire
  Local uValue
  Local iFieldNo := 0
  IF ::bUsed
    IF ValType(fld_nroname) == "C"
      iFieldNo := ::FieldPos( Left(fld_nroname,10) )
    ElseIF ValType(fld_nroname) $ "NI"
      iFieldNo := fld_nroname
    Endif
    // - - - 
    IF iFieldNo > 0
      uValue := ::getRecordBuffer(iFieldNo)
      AssignFldValue(@aFieldVars[iFieldNo], uValue)

      // - - - Log
      ::writeLog( LOG_LEVEL_09, "[FGET] FieldGet: " + ::FieldName(iFieldNo) ;
                    + " T: " + ValType(uValue) + " V: " ;
                    + IIF(ValType(uValue)=="N",Alltrim(Str(uValue)),"") ;
                    + IIF(ValType(uValue)=="D",DtoC(uValue),"") ;
                    + IIF(ValType(uValue)=="C",uValue,"") )

      Return @aFieldVars[iFieldNo]
    Endif
  Endif
  Return NIL

METHOD Axit() CLASS rqRDDFire
  // - - - Log
  wrtLog( LOG_LEVEL_06, "[GONE] Class: rqRDDFire no longer exists. Alias: " ;
          + ::sAlias, ::iLogLevel )
  Return .T.

METHOD NoMethod( meth ) CLASS rqRDDFire
  // - - - Log
  ::writeLog( LOG_OFF, "[ERROR] Meth: " + meth + " does not exist inside: ";
              + "rqRDDFire class" )

  raiseError( "This class has no Method named: " + meth )
  Return

METHOD Error( obj AS ERROR, vl ) CLASS rqRDDFire
  // - - - Log
  ::writeLog( LOG_OFF, "[ERROR] Unknown:" + obj:DESCRIPTION + " (" + vl + ")" )
  raiseError( "Unknown: " + obj:DESCRIPTION + " (" + vl + ")" )
  Return NIL

METHOD NoIVarGet( p1 ) CLASS rqRDDFire
  Return "Not suported"

PROTECT METHOD flushRecordBuffer() CLASS rqRDDFire
  Local conn         := ::getConnection() AS rqFBConn
  Local sValues      := ""
  Local sStmt        := ""
  Local aParam       := {}
  Local bOk          := .F.
  Local recno        := Alltrim(Str(::getBufferRecno()))
  Local bPartOfIndex := .F.
  Local sColValue 
  Local sFieldName, sFieldType, iRddFieldType
  Local iCt, sAux

  IF ::bAppendPending
    sStmt := ;
      'INSERT INTO @@TABLENAME ( @@FIELDS ) VALUES ( @@VALUES )'

    bOk := .T.
  Else
    IF conn:existsAnotherUserFLock(::sTableName, ::sAreaID) .or. ;
       conn:isRecordLockedByAnotherUser(::sTableName, ::sAreaID, recno )

      // - - - Log
      ::writeLog( LOG_OFF, "[ERROR] Flush on Locked Record or Table.;Alias: " ;
                  + ::sAlias + " Rec #" + recno ) 
      raiseError( "Table or Record locked. Impossible to update!; " ;
                  + "Alias: " + ::sAlias + " Rec #" + recno ) 
    Else
      sStmt := ;
        'UPDATE @@TABLENAME SET @@UPDTVALUES WHERE @@FIELDRECNO = @@UPDTRECNO'

      bOk := .T.
    Endif
  Endif

  IF bOk
    bOk    := .F.
    aParam := {}
    Aadd(aParam, {"@@TABLENAME",  ::sTableName})
    Aadd(aParam, {"@@FIELDRECNO", RECNO_FIELD})
    Aadd(aParam, {"@@UPDTRECNO",  recno})
    IF !::bAppendPending .and. ::getBufferRecno() == 0
      raiseError( "Updating phantom record; Table: " + ::sTableName )

      // - - - Log
      ::writeLog( LOG_OFF, "[PHANTOM] Updating phantom record. Table: " + ;
                           ::sTableName)
    Endif
    For iCt := 1 To ::FCount
      sFieldName    := ::aStru[iCt,DBS_NAME]
      sFieldType    := ::aStru[iCt,DBS_TYPE]
      iRddFieldType := ::aStru[iCt,DBS_RDDFIELDTYPE]
      IF ::aFieldChanged[iCt] .and. iRddFieldType == RDDFIELDTYPE_NORMAL
        IF sFieldType == "D"
          sAux := DtoS(CtoD(::aActualRecord[iCt]))
          IF !Empty(sAux)
            sColValue := Substr(sAux,1,4) + "-" + Substr(sAux,5,2) + "-" ;
                         + Substr(sAux,7,2)
          Else
            sColValue := NULL_FB_FIELD
          Endif
        Else
          sColValue := ::aActualRecord[iCt]
        Endif
        //sColValue := Alltrim(StrTran(sColValue,"'", "''"))
        sColValue := StrTran(sColValue,"'", "''") // Norm.aspas simples
        Aadd(aParam, {"@@FIELDS", sFieldName + ",@@FIELDS"})
        IF sColValue == NULL_FB_FIELD
          Aadd(aParam, {"@@VALUES", "null,@@VALUES"})
          Aadd(aParam, {"@@UPDTVALUES", sFieldName + " = null,@@UPDTVALUES"})
        Else
          Aadd(aParam, {"@@VALUES", "'" + sColValue + "',@@VALUES"})
          Aadd(aParam, {"@@UPDTVALUES", sFieldName + " = '" + sColValue ;
                        + "',@@UPDTVALUES"})
        Endif
        bPartOfIndex := At(sFieldName, Upper(::IndexKey)) > 0
        bOk := .T. 
      Endif
    Next
    IF ::bAppendPending
      Aadd(aParam, {"@@FIELDS", AREAID_FIELD + ",@@FIELDS"})
      Aadd(aParam, {"@@VALUES", "'" + ::sAreaID + "',@@VALUES"})
    Endif
    Aadd(aParam, {",@@FIELDS",     ""})
    Aadd(aParam, {",@@VALUES",     ""})
    Aadd(aParam, {",@@UPDTVALUES", ""})
  Endif

  IF bOk
    IF ::bCachingAppends
      Aadd( ::aCachedAppends, conn:assembleStatement(sStmt, aParam) ) 
    Else
      // - - - Normal Operation
      sAux := conn:executeUpdate( conn:assembleStatement(sStmt, aParam) )
      IF Empty(sAux)
        IF ::bAppendPending
          ::clearAllCaches()
          ::setRecordChanged( 0, .F. )
          ::Goto( conn:getAPPRecnoFromArea(::sTableName, ::sAreaID) )
          ::RLock( ::getBufferRecno() )
        Else
          IF bPartOfIndex
            ::clearAllCaches()
            ::Refresh()
          Endif
        Endif
      Else
        raiseError( "Update failed!; FB Error: " + sAux )
      Endif
    Endif
  EnDif
  ::setRecordChanged( 0, .F. )
  ::bAppendPending := .F.
  Return

METHOD flushCachedAppends() CLASS rqRDDFire AS Logical // ??? Not Documented
  Local conn := ::getConnection() AS rqFBConn
  Local bOk := .T.
  Local sAux
  Local iCt

  For iCt := 1 To Len(::aCachedAppends)
    sAux := conn:executeUpdate( ::aCachedAppends[iCt] )
    IF !Empty(sAux)
      bOk := .F.

      // - - - Error
      raiseError( "Flushing Cached Appends failed!;FB Error: " + sAux )
      Exit
    Endif
  Next
  IF bOk
    ::clearAllCaches()
    ::setRecordChanged( 0, .F. )
    ::Goto( conn:getAPPRecnoFromArea(::sTableName, ::sAreaID) )
    ::RLock( ::getBufferRecno() )
  Endif
  ::aCachedAppends  := {}
  ::bCachingAppends := .F.
  Return bOk

PROTECT METHOD reduceAsciiRecno( ln ) CLASS rqRDDFire
  Local sChar := Right(ln,1)
  Local iAux  := Asc(sChar) - 1
  IF iAux < Asc("0")
    sChar := " "
  Else
    sChar := Chr(iAux)
  Endif
  Return Stuff( ln, Len(ln), 1, sChar )

PROTECT METHOD createSkipPadExpression() CLASS rqRDDFire
  Local sRet        := ""
  Local aIdxFields  := {}
  Local sFieldType, iCt

  IF ::iActualOrder > 0
    aIdxFields := ::translateLineToArray( 
                    ::aIndexes[::iActualOrder][RDDIDX_FIELDS], "," ) 
  Else 
    aIdxFields := { RECNO_FIELD }
  Endif
  For iCt = 1 To Len(aIdxFields)
    sFieldType := ::FieldInfo(DBS_TYPE,aIdxFields[iCt])
    Do Case 
      Case sFieldType == "D"
        sRet += '"{{D}}"'
      Case sFieldType == "N"
        IF aIdxFields[iCt] == RECNO_FIELD
          sRet += '"{{R}}"'
        Else
          sRet += '"{{N}}"'
        Endif
      Otherwise
        sRet += '"{{C}}"'
    EndCase
    sRet += " || " + aIdxFields[iCt] + " || "
  Next
  sRet := Substr(sRet, 1, Rat(" || ",sRet) - 1)
  Return sRet

PROTECT METHOD getRemoteSkipPad() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local sRet := ""
  Local aRet := conn:executeUDFSkipPad( ;
                     ::createSkipPadExpression(), ;
                     ::sTableName, ;
                     ::aActualRecord[::FieldPos(RECNO_FIELD)] )
  IF !Empty(aRet)
    sRet := Alltrim(aRet[1])
  Endif
  Return sRet

METHOD GoTop() CLASS rqRDDFire
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local rs   := {}
  Local iCt, sOrder

  ::bEof := ::bBof := ::bFound := .F.
  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif
  IF ::iActualOrder > 0
    sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
  Else
    sOrder := RECNO_FIELD
  Endif    
  rs := conn:getRow( ::sTableName, '1=1', sOrder, "", 0 )
  IF !Empty(rs)
    ::aActualRecord := rs
    ::sNextRecord   := ::aActualRecord[::FieldPos(sOrder)]
    ::bDeleted      := .F.

    IF ::bFiltered 
      Do While !::bEof 
        IF Eval(::blkFilter)
          // - - - Log
          ::writeLog( LOG_LEVEL_07, "[GOTO] Top (Filtered). Table: " ;
                      + ::sTableName + " Alias: " + ::sAlias + " Condition: ";
                      + ::sFilter + " Rec #" ;
                      + Alltrim(Str(::getBufferRecno())) )
          bRet := .T.
          Exit
        Endif
        ::Skip(1)
      Enddo
      IF ::bEof 
        ::bBof := .T.
      Endif
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[GOTO] Top. Table: " + ::sTableName ;
                  + " Alias: " + ::sAlias + " Rec #" ;
                  + Alltrim(Str(::getBufferRecno())) )
      bRet := .T.
    Endif
  Else
    ::bEof := ::bBof := .T.
  Endif
  Return bRet

METHOD GoBottom() CLASS rqRDDFire
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local rs   := {}
  Local iCt, sOrder

  ::bEof := ::bBof := ::bFound := .F.
  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif
  IF ::iActualOrder > 0
    sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
  Else
    sOrder := RECNO_FIELD
  Endif    
  rs := conn:getRow( ::sTableName, '1=1', sOrder + " desc", "", 0 )
  IF !Empty(rs)
    ::aActualRecord := rs
    ::sNextRecord   := ::aActualRecord[::FieldPos(sOrder)]
    ::bDeleted      := .F.

    IF ::bFiltered 
      Do While !::bBof 
        IF Eval(::blkFilter)
          // - - - Log
          ::writeLog( LOG_LEVEL_07, "[GOTO] Bottom (Filtered). Table: " ;
                      + ::sTableName + " Alias: " + ::sAlias + " Condition: ";
                      + ::sFilter + " Rec #" ;
                      + Alltrim(Str(::getBufferRecno())) )
          bRet := .T.
          Exit
        Endif
        ::Skip(-1)
      Enddo
      IF ::bBof
        ::bEof := .T.
      Endif
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[GOTO] Bottom. Table: " + ::sTableName ;
                  + " Alias: " + ::sAlias + " Rec #" ;
                  + Alltrim(Str(::getBufferRecno())) )
      bRet := .T.
    Endif
  Else
    ::bEof := ::bBof := .T.
  Endif
  Return bRet

PROTECT METHOD isTableCompatibleWithRDD() CLASS rqRDDFire
  Local bRet := .F.
  Local bRecno, bAreaID
  Local iAux
  bRecno := bAreaID := .F.
  iAux := AScan(::aStru, {|x| x[DBS_RDDFIELDTYPE] == RDDFIELDTYPE_RECNO}) 
  IF iAux > 0
    bRecno := .T.
  Endif
  iAux := AScan(::aStru, {|x| x[DBS_RDDFIELDTYPE] == RDDFIELDTYPE_AREAID}) 
  IF iAux > 0
    bAreaID := .T.
  EndiF
  IF bRecno .and. bAreaID
    bRet := .T.
  Endif
  Return bRet

PROTECT METHOD refreshTableStructure() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local aAux := conn:setRddFields( conn:getTableStruct(::sTableName) )
  ::aStru := aAux[1]
  Return

PROTECT METHOD setTableUsedbyThisArea( bol ) CLASS rqRDDFire
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  IF bol
    IF !conn:isTableInExclusiveMode( ::sTableName )
      IF ::bShared
        bRet := conn:setTableUsedbyArea( bol, ::sTableName, ::sAreaID, .T. ) 
      Else
        IF !conn:isTableInUse( ::sTableName )
          bRet := conn:setTableUsedbyArea( bol, ::sTableName, ::sAreaID, .F. ) 
        Else
          // - - - Log
          ::writeLog( LOG_LEVEL_04, "[WARN] Exclusive mode NOT allowed " ;
                      + "on opened tables. Table: " + ::sTableName )
        Endif
      Endif
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_04, "[WARN] Exclusive by another area/process. " ;
                  + "Table: " + ::sTableName )
    Endif
  Else
    bRet := conn:setTableUsedbyArea( bol, ::sTableName, ::sAreaID, .T. ) 
  Endif
  Return bRet

PROTECT METHOD openTable() CLASS rqRDDFire
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local aAux := {}
  IF conn:isTable( ::sTableName )
  IF conn:isUDF( "SKIPPAD", "rqrddudfs" )
    aAux := conn:setRddFields( conn:getTableStruct(::sTableName) )
    ::aStru := aAux[1]
    ::bLimboFields := aAux[2]
    IF !::bLimboFields
      IF ::isTableCompatibleWithRDD()
        // - - - Sucesso
        bRet := .T.
      Else
        // - - - Log
        ::writeLog( LOG_OFF, "[ERROR] Table " + ::sTableName + " does not " ;
                    + "have a compatible structure" )

        raiseError( "Table doesn't have a compatible structure!;Table: " ;
                    + ::sTableName )
      Endif
    Else
      // - - - Log
      ::writeLog( LOG_OFF, "[ERROR] Table " + ::sTableName + " with " ;
                  + "Limbo Fields!" )

      raiseError( "Table has Limbo Fields, probably created by an " ;
                  + "unsuccessful; structure update; Table: " + ::sTableName )
    Endif
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] UDF SkipPad does not exists. Table: " ;
                         + ::sTableName )

    raiseError( "UDF SkipPad is highly needed for tables opening;" ;
                + "Table: " + ::sTableName)
  Endif
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] Table " + ::sTableName + " does not exists" )

    raiseError( "Table does not exists!;Table: " + ::sTableName )
  Endif
  Return bRet 

METHOD DbStruct() CLASS rqRDDFire
  Local aRet := {}
  Local iCt

  For iCt := 1 To Len(::aStru)
    IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
      Aadd(aRet, {::aStru[iCt,DBS_NAME], ::aStru[iCt,DBS_TYPE], ;
                  ::aStru[iCt,DBS_LEN],  ::aStru[iCt,DBS_DEC]} )
    Endif
  Next
  Return aRet

PROTECT METHOD clearAllCaches() CLASS rqRDDFire
  ::aSkCacheAsc  := {}
  ::aSkCacheDesc := {}
  ::aSeekCache   := {}
  Return

PROTECT METHOD addSeekCache( cond, rs ) CLASS rqRDDFire
  IF !Empty(rs)
    IF Empty(::aSeekCache)
      ::iSeekChAge := Seconds()
    Endif
    AAdd(::aSeekCache, {cond, rs})
  Endif
  Return

PROTECT METHOD setSkipCacheAsc( rs ) CLASS rqRDDFire
  Local aAux := {}
  IF Len(rs) >= 2
    aAux := ARRAY( Len(rs) - 1 )
    ::aSkCacheAsc    := ACopy( rs, aAux, 2 )
    ::iSkCacheAscAge := Seconds()

    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[CACHE] CacheAsc Sz: " ;
                              + Alltrim(Str(Len(::aSkCacheAsc))) )
  Endif
  Return

PROTECT METHOD setSkipCacheDesc( rs ) CLASS rqRDDFire
  Local aAux := {}
  IF Len(rs) >= 2
    aAux := ARRAY( Len(rs) - 1 )
    ::aSkCacheDesc    := ACopy( rs, aAux, 2 )
    ::iSkCacheDescAge := Seconds()

    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[CACHE] CacheDesc Sz: " ;
                              + Alltrim(Str(Len(::aSkCacheDesc))) )
  Endif
  Return

PROTECT METHOD isSkCacheAscValid() CLASS rqRDDFire
  Local bRet := .F.
  Local iAge := Abs(Seconds() - ::iSkCacheAscAge)
  IF iAge <= ::getSkipCacheTimeOut() 
    bRet := .T.
  Else
    ::aSkCacheAsc := {}
    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[TIMEOUT] CacheAsc: " + Alltrim(Str(iAge)) ;
                + "s Max.: " + Alltrim(Str(::getSkipCacheTimeOut())) )
  Endif
  Return bRet
PROTECT METHOD isSkCacheDescValid() CLASS rqRDDFire
  Local bRet := .F.
  Local iAge := Abs(Seconds() - ::iSkCacheDescAge)
  IF iAge <= ::getSkipCacheTimeOut() 
    bRet := .T.
  Else
    ::aSkCacheDesc := {}
    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[TIMEOUT] CacheDesc: " + Alltrim(Str(iAge)) ;
                + "s Max.: " + Alltrim(Str(::getSkipCacheTimeOut())) )
  Endif
  Return bRet

PROTECT METHOD isSkipCached( ln, field_nro, ascen, sk ) CLASS rqRDDFire
  Local bRet   := .F.
  Local sPlace := ""
  Local iPos   := 0 
  IF ::isSkCacheAscValid()
    iPos := AScan(::aSkCacheAsc, {|x| x[field_nro] == ln})
    IF iPos > 0 
      IF ascen 
        IF (iPos + sk) <= Len(::aSkCacheAsc)
          bRet := .T.
          sPlace := "CacheAsc"
        Endif
      Else
        IF (iPos - sk) >= 1
          bRet := .T.
          sPlace := "CacheAsc"
        Endif
      Endif
    Endif
  Endif
  IF !bRet .and. ::isSkCacheDescValid()
    iPos := AScan(::aSkCacheDesc, {|x| x[field_nro] == ln})
    IF iPos > 0
      IF ascen
        IF (iPos - sk) >= 1
          bRet := .T.
          sPlace := "CacheDesc"
        Endif
      Else
        IF (iPos + sk) <= Len(::aSkCacheDesc)
          bRet := .T.
          sPlace := "CacheDesc"
        Endif
      Endif
    Endif
  Endif

  IF bRet
    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[CACHED] Skip said yes. Ascen: " ;
                + IIF(ascen, "True", "False") + " Place: " + sPlace ;
                + " Pos: " + Alltrim(Str(iPos)) )
  Endif
  Return bRet

PROTECT METHOD isGotoCached( sRec, recno_fldnro ) CLASS rqRDDFire
  Local bRet := .F.
  Local iPos := 0 
  IF ::isSkCacheAscValid()
    iPos := AScan(::aSkCacheAsc, {|x| Val(x[recno_fldnro]) == Val(sRec)})
    IF iPos > 0 
      bRet := .T.
    Endif
  Endif
  IF !bRet .and. ::isSkCacheDescValid()
    iPos := AScan(::aSkCacheDesc, {|x| Val(x[recno_fldnro]) == Val(sRec)})
    IF iPos > 0 
      bRet := .T.
    Endif
  Endif

  IF bRet
    // - - - Log
    ::writeLog( LOG_LEVEL_08, "[CACHED] Goto said yes" )
  Endif
  Return bRet
   
PROTECT METHOD getSkipCache( ln, field_nro, ascen, sk ) CLASS rqRDDFire AS Array
  Local aRet := {}
  Local iPos := 0 
  iPos := AScan(::aSkCacheAsc, {|x| x[field_nro] == ln})
  IF iPos > 0 
    IF ascen
      IF (iPos + sk) <= Len(::aSkCacheAsc)
        aRet := ::aSkCacheAsc[iPos + sk] 
      Endif
    Else
      IF (iPos - sk) >= 1
        aRet := ::aSkCacheAsc[iPos - sk] 
      Endif
    Endif
  Endif
  IF Empty(aRet)
    iPos := AScan(::aSkCacheDesc, {|x| x[field_nro] == ln})
    IF iPos > 0 
      IF ascen
        IF (iPos - sk) >= 1
          aRet := ::aSkCacheDesc[iPos - sk] 
        Endif
      Else
        IF (iPos + sk) <= Len(::aSkCacheDesc)
          aRet := ::aSkCacheDesc[iPos + sk] 
        Endif
      Endif
    Endif
  Endif

  IF Empty(aRet)
    // - - - Log
    ::writeLog( LOG_OFF, "[FATAL] SkipCache sent empty record!" )
  Endif
  Return aRet

PROTECT METHOD getGotoCache( sRec, recno_fldnro ) CLASS rqRDDFire AS Array
  Local aRet := {}
  Local iPos := 0 
  iPos := AScan(::aSkCacheAsc, {|x| Val(x[recno_fldnro]) == Val(sRec)})
  IF iPos > 0 
    aRet := ::aSkCacheAsc[iPos] 
  Else
    iPos := AScan(::aSkCacheDesc, {|x| Val(x[recno_fldnro]) == Val(sRec)})
    IF iPos > 0 
      aRet := ::aSkCacheDesc[iPos] 
    Endif
  Endif

  IF Empty(aRet)
    // - - - Log
    ::writeLog( LOG_OFF, "[FATAL] GotoCache sent empty record!" )
  Endif
  Return aRet

PROTECT METHOD getSeekCached( cond ) CLASS rqRDDFire AS Array
  Local aRet := {}
  Local iPos := 0 
  Local iAge := Abs(Seconds() - ::iSeekChAge)
  IF iAge <= ::iSeekChTmOut 
    iPos := AScan(::aSeekCache, {|x| (x[1]) == cond})
    IF iPos > 0 
      aRet := ::aSeekCache[iPos,2] 

      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[SEEK] Cached sWh: " + cond + " Tbl: " ;
                  + ::sTableName + " Als: " + ::sAlias )
    Endif
  Else
    ::aSeekCache := {}
  Endif
  Return aRet

METHOD Goto( vl ) CLASS rqRDDFire AS Logical
  Local bRet    := .F.
  Local bCached := .F.
  Local conn    := ::getConnection() AS rqFBConn
  Local rs      := {}
  Local aAux    := {}
  Local recno   := Alltrim(Str(vl))
  Local iCt, sWhere, sOrder
  Local iFldIdx

  IF vl == ::getBufferRecno() .or. vl == 0
    // - - - Log
    ::writeLog( LOG_LEVEL_07, "[GOTO] Ineffective! Rec #" + recno ;
                + " Tbl: " + ::sTableName + " Als: " + ::sAlias )
    Return .T.
  Endif

  ::bEof := ::bBof := ::bFound := .F.
  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif
  IF ::iActualOrder > 0
    sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
  Else
    sOrder := RECNO_FIELD
  Endif    
  iFldIdx := ::FieldPos(sOrder) 
  Do Case
    Case ::iSkipMode == SKIP_MODE_0 .or. ;
         ::iSkipMode == SKIP_MODE_1		// - - Mode 0 or 1

      sWhere := RECNO_FIELD + " = '" + recno + "'"
      rs     := conn:getRow( ::sTableName, sWhere, RECNO_FIELD, "", 0 )
      IF !Empty(rs)
        ::aActualRecord := rs
        ::sNextRecord   := ::aActualRecord[::FieldPos(sOrder)]
        ::bDeleted      := .F.
      Else
        ::bEof := ::bBof := .T.
      Endif
    Case ::iSkipMode == SKIP_MODE_2             // - - Mode 2
      IF ::isGotoCached( recno, ::FieldPos(RECNO_FIELD) )
        aAux    := ::getGotoCache( recno, ::FieldPos(RECNO_FIELD) )
        rs      := { {}, aAux }
        bCached := .T.
      Else
        sWhere := RECNO_FIELD + " = '" + recno + "'"
        rs     := conn:getFirstRows( ::sTableName, sWhere, RECNO_FIELD, "", ;
                                     1, 0 )
      Endif
      IF !Empty(rs)
        ::aActualRecord := rs[2]
        ::sNextRecord   := ::aActualRecord[::FieldPos(sOrder)]
        ::bDeleted      := .F.
      Else
        ::bEof := ::bBof := .T.
      Endif
  EndCase
  IF !::bEof .and. !::bBof
    IF bCached 
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[GOTO] Cached Goto Rec #" + recno ;
                  + ". Tbl: " + ::sTableName + " Als: " + ::sAlias )
      bRet := .T.
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[GOTO] Goto Rec #" + recno ;
                  + ". Tbl: " + ::sTableName + " Als: " + ::sAlias )
      bRet := .T.
    Endif
  Endif
  Return bRet

METHOD Skip( vl ) CLASS rqRDDFire
  // - - - Chamado tambem entre troca de areas
  Local conn       := ::getConnection() AS rqFBConn
  Local bRet       := .T.
  Local rs         := {}
  Local aAux       := {}
  Local sWhere     := ""
  Local bNewCache  := .F.
  Local iCount     := IIF(vl == NIL, 1, vl)
  Local sk         := 0                        // Skip para o FB
  Local sk_flt     := 0                        // Acum. for filtered table
  Local iCt, iCt2, sOrder
  Local iFldIdx
      
  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif

  IF ::bEof .and. ::bBof
    Return .F.
  Endif
  IF ::bEof .and. iCount < 0
    ::GoBottom()
    iCount += 1
  Endif
  //::bEof := ::bBof := ::bFound := .F.
  ::bFound := .F.

  IF ::iActualOrder > 0
    sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
  Else
    sOrder := RECNO_FIELD
  Endif   
  iFldIdx := ::FieldPos(sOrder) 
  IF iCount > 0 .and. !::bEof                              // - - - Forward
    ::bEof := ::bBof := .F.
    IF ::bFiltered
      sk := 0
    Else
      sk := iCount - 1 
    Endif
    Do Case
      Case ::iSkipMode == SKIP_MODE_0		// - - Mode 0
          Do While !::bEof
            sWhere := sOrder + " > '" + ::sNextRecord + "'"
            rs := conn:getRow( ::sTableName, sWhere, sOrder, "", sk )
            IF !Empty(rs)
              ::aActualRecord := rs
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == iCount
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bEof := .T.     // Eof
            Endif
          Enddo
      Case ::iSkipMode == SKIP_MODE_1		// - - Mode 1 (Default)
          Do While !::bEof
            IF ::isSkipCached( ::sNextRecord, iFldIdx, .T., sk+1 )
              aAux      := ::getSkipCache( ::sNextRecord, iFldIdx, .T., sk+1 )
              sWhere    := sOrder + " = '" + aAux[iFldIdx] + "'"
              bNewCache := .F.
            Else
              sWhere    := sOrder + " > '" + ::sNextRecord + "'"
              bNewCache := .T.
            Endif
            IF bNewCache
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder, "", ;
                                       ::getSkipCacheSize(), sk ) // Asc 
            Else
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder, "", 1, 0 )
            Endif
            IF !Empty(rs)
              ::aActualRecord := rs[2]
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.
              IF Len(rs) >= 2 .and. bNewCache
                ::setSkipCacheAsc( rs )
              Endif

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == iCount
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bEof := .T.     // Eof
            Endif
          Enddo
      Case ::iSkipMode == SKIP_MODE_2		// - - Mode 2
          Do While !::bEof
            IF ::isSkipCached( ::sNextRecord, iFldIdx, .T., sk+1 )
              aAux      := ::getSkipCache( ::sNextRecord, iFldIdx, .T., sk+1 )
              rs        := { {}, aAux }
              bNewCache := .F.
            Else
              sWhere    := sOrder + " > '" + ::sNextRecord + "'"
              bNewCache := .T.
            Endif
            IF bNewCache
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder, "", ;
                                       ::getSkipCacheSize(), sk )     // Asc 
            Endif
            IF !Empty(rs)
              ::aActualRecord := rs[2]
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.
              IF Len(rs) >= 2 .and. bNewCache
                ::setSkipCacheAsc( rs )
              Endif

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == iCount
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bEof := .T.     // Eof
            Endif
          Enddo
    EndCase
  ElseIF iCount < 0 .and. !::bBof                      // - - - Backwards
    ::bEof := ::bBof := .F.
    IF ::bFiltered
      sk := 0
    Else
      sk := Abs(iCount) - 1
    Endif
    Do Case
      Case ::iSkipMode == SKIP_MODE_0		// - - Mode 0
          Do While !::bBof
            sWhere := sOrder + " < '" + ::sNextRecord + "'"
            rs := conn:getRow( ::sTableName, sWhere, sOrder + " desc", "", sk )
            IF !Empty(rs)
              ::aActualRecord := rs
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == Abs(iCount)
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bBof := .T.     // Bof
            Endif
          Enddo
      Case ::iSkipMode == SKIP_MODE_1		// - - Mode 1 (Default)
          Do While !::bBof
            IF ::isSkipCached( ::sNextRecord, iFldIdx, .F., sk+1 )
              aAux      := ::getSkipCache( ::sNextRecord, iFldIdx, .F., sk+1 )
              sWhere    := sOrder + " = '" + aAux[iFldIdx] + "'"
              bNewCache := .F.
            Else
              sWhere    := sOrder + " < '" + ::sNextRecord + "'"
              bNewCache := .T.
            Endif
            IF bNewCache
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder + " desc", ;
                                       "", ::getSkipCacheSize(), sk )   // Desc
            Else
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder, "", 1, 0 )
            Endif
            IF !Empty(rs)
              ::aActualRecord := rs[2]
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.
              IF Len(rs) >= 2 .and. bNewCache
                ::setSkipCacheDesc( rs )
              Endif

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == Abs(iCount)
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bBof := .T.     // Bof
            Endif
          Enddo
      Case ::iSkipMode == SKIP_MODE_2		// - - Mode 2
          Do While !::bBof
            IF ::isSkipCached( ::sNextRecord, iFldIdx, .F., sk+1 )
              aAux      := ::getSkipCache( ::sNextRecord, iFldIdx, .F., sk+1 )
              rs        := { {}, aAux }
              bNewCache := .F.
            Else
              sWhere    := sOrder + " < '" + ::sNextRecord + "'"
              bNewCache := .T.
            Endif
            IF bNewCache
              rs := conn:getFirstRows( ::sTableName, sWhere, sOrder + " desc",;
                                       "", ::getSkipCacheSize(), sk )  // Desc 
            Endif
            IF !Empty(rs)
              ::aActualRecord := rs[2]
              ::sNextRecord   := ::aActualRecord[iFldIdx]
              ::bDeleted      := .F.
              IF Len(rs) >= 2 .and. bNewCache
                ::setSkipCacheDesc( rs )
              Endif

              IF ::bFiltered
                IF Eval(::blkFilter)
                  sk_flt += 1
                  IF sk_flt == Abs(iCount)
                    Exit
                  Endif
                Endif
              Else
                Exit
              Endif
            Else
              ::bBof := .T.     // Bof
            Endif
          Enddo
    EndCase
    IF ::bFiltered .and. ::bBof .and. !::bEof
      ::GoTop()
      ::bBof := .T. // Forced! 'cos gotop resets bof
    Endif
  Endif
  Return bRet

PROTECT METHOD commitSuicide() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  IF !::bUsed
    IF conn:dropTable(::sTableName)
      // - - - Log
      ::writeLog( LOG_LEVEL_08, "[DROP$$$] Inside of this log remains the " ;
                  + "ashes of: " + ::sTableName + " Alias: " + ::sAlias ) 
    Else
      // - - - Log
      ::writeLog( LOG_OFF, "[ERROR] Could not drop temporary Table: " ;
                  + ::sTableName + " Alias: " + ::sAlias ) 
    Endif
  Endif
  Return

METHOD Close() CLASS rqRDDFire AS Logical
  ::closeRDD()
  Return .T.

METHOD ClearIndex() CLASS rqRDDFire
  // - - - Zerar matriz de indices
  ::aIndexes     := {}
  ::iActualOrder := 0
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[INDX] Cleared. Alias: " + ::sAlias )
  Return .T.

METHOD SetIndex( ind_name, p1, p2 ) CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local bRet := .F.
  Local aIndexFields := {}
  Local sIndexFields := ""
  Local sIndexExpr   := ""
  Local iCt

  ind_name := ::removeFileNameExt(::getFileName(ind_name))
  IF conn:isIndexOfTable(::sTableName, ind_name) 
    aIndexFields := conn:getIndexFields(ind_name)
    sIndexExpr   := Alltrim((conn:getIndexDescription(ind_name))[1])
    IF !Empty(aIndexFields)
      For iCt := 1 To Len(aIndexFields)
        sIndexFields += Alltrim(Upper(aIndexFields[iCt,1])) + ","
      Next
      sIndexFields := Substr(sIndexFields, 1, Rat(",",sIndexFields) - 1)
      Aadd(::aIndexes, {ind_name, sIndexFields, sIndexExpr})
      NetErr( .F. )

      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[INDX] Opened. Name: " + ind_name ;
                  + " Key: " + sIndexFields + " FS Expr: " + sIndexExpr)

      IF Len(::aIndexes) == 1 // Active first one
        ::SetOrder(1)
        ::GoTop()
      Endif
    Endif
  Else
    NetErr( .T. )
    raiseError( "Unknown Index!;Index: " + ind_name + " Table: " ;
                + ::sTableName + " Alias: " + ::sAlias )
  Endif
  Return bRet

METHOD SetOrder( ind_nroname, p1 ) CLASS rqRDDFire
  Local sRet   := ""
  Local nRecno := ::getBufferRecno()
  IF ValType(ind_nroname) == "N"
    IF ind_nroname <= Len(::aIndexes)
      IF ind_nroname <> ::iActualOrder
        ::iActualOrder := ind_nroname

        IF ind_nroname == 0  
          // - - - Log
          ::writeLog( LOG_LEVEL_07, "[INDX] Honouring: Natural Order" ;
                      + " Rec #" + Alltrim(Str(nRecno)) ;
                      + IIF(::bBof," (Bof)","") + IIF(::bEof," (Eof)","") )
        Else
          // - - - Log
          ::writeLog( LOG_LEVEL_07, "[INDX] Honouring: " ;
                      + ::aIndexes[::iActualOrder][RDDIDX_NAME] + " Order:" ;
                      + Str(ind_nroname,2) + "/" + Alltrim(Str(Len(;
                        ::aIndexes))) + " Rec #" + Alltrim(Str(nRecno)) ;
                      + IIF(::bBof," (Bof)","") + IIF(::bEof," (Eof)","") )
        Endif

        ::clearAllCaches()
        ::Refresh()
      Endif
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[ERROR] Order " + Str(ind_nroname,2) + " " ;
                  + "out of range! Tbl: " + ::sTableName + " Als: " + ::sAlias )

      raiseError( "Order " + Str(ind_nroname,2) + " out of range!;Table: " ;
                  + ::sTableName + " Alias: " + ::sAlias ) 
    Endif
  Endif
  // - - - Retorna o nome da ultima ordem possivel ou ""
  IF !Empty(::aIndexes)
    sRet := ::aIndexes[Len(::aIndexes)][RDDIDX_NAME]
  Endif
  Return sRet 

PROTECT METHOD isolateFieldName( ln ) CLASS rqRDDFire
  Do While At("(",ln) > 0
    ln := Substr(ln, At("(",ln) + 1, Rat(")",ln) - At("(",ln) - 1)
  Enddo
  IF At(">",ln) > 0
    ln := Substr(ln, At(">",ln) + 1)
  Endif
  IF At(",",ln) > 0
    ln := Substr(ln, 1, At(",",ln) - 1)
  Endif
  Return Alltrim(Upper(ln))

METHOD CreateOrder( ind_name, p1, expr, p2, unq ) CLASS rqRDDFire
  Local conn          := ::getConnection() AS rqFBConn
  Local bRet          := .T.
  Local aCampos       := {}
  Local aExp          := ::translateLineToArray( expr, "+" )
  Local iExpSize      := IIF(ValType(expr) == "C", Len(&expr))
  Local bUnique       := .T. // unq auto Unique
  Local sFields       := ""
  Local ind_name_desc := ""
  Local iCt, sAux
  Local sTrigger, sTrigger_updt, sTrigger_isrt
  Local sExp, sFuncName, sFuncParam
  Local sFieldName, sStmt

  IF Empty(bUnique)
    bUnique := .F.
  Endif
  IF Empty(ind_name)
    bRet := .F.
  Else
    ind_name := ::removeFileNameExt(::getFileName(ind_name))
    IF !Empty(ind_name)
      ind_name_desc := ind_name + "_DESC"
      sTrigger_updt := "T_" + ::sTableName + "_" + ind_name + "_U"
      sTrigger_isrt := "T_" + ::sTableName + "_" + ind_name + "_I"
    Else
      bRet := .F.
    Endif
  Endif 
  // - - - Size of Expression
  IF iExpSize <= 0
    raiseError( "Invalid evaluated expression size (0)!" )
    bRet := .F.
  Else
    iExpSize += RECNO_FIELD_SIZE // Firebird Recno Field
  Endif
  // - - - Drop Index Field which means dropping real FB indexes for Asc/Desc 
  //       and any related Trigger in this case Update/Insert
  IF bRet
    IF conn:isFieldOfTable(::sTableName, RDDIDX_PREFIX+ind_name)
      IF !conn:dropField(::sTableName, {RDDIDX_PREFIX+ind_name, "", 0, 0})
        raiseError( "Error on dropping Index Field; Field: " + RDDIDX_PREFIX ;
                    + ind_name )
        bRet := .F.
      Endif
    Endif
  Endif
  // - - - Add Index Field
  IF bRet
    sAux := conn:executeUpdate( "ALTER TABLE " + ::sTableName + " ADD " ;
            + RDDIDX_PREFIX + ind_name + " VARCHAR(" + Alltrim(Str(iExpSize));
            + ")")
    IF Empty(sAux)
      ::refreshTableStructure()
    Else
      raiseError( "Impossible to add Index Field;Field: " ;
                  + RDDIDX_PREFIX + ind_name + ";FB Error: " + sAux )
      bRet := .F.
    Endif
  Endif
  // - - - Isolate fields
  IF bRet
    For iCt := 1 To Len(aExp)
      sFieldName := ::isolateFieldName(aExp[iCt])
      IF ::FieldPos(sFieldName) == 0
        raiseError( "Invalid field or unrecognized field specification.;" ; 
                    + "Field: " + sFieldName + ";Try the following " ;
                    + "formats:;'field->your-fieldname' or just " ;
                    + "'your-fieldname'" )
        bRet := .F.
      Else
        AAdd(aCampos, sFieldName)
      Endif
    Next
  Endif
  // - - - E X P R E S S I O N  A S S E M B L I N G
  IF bRet
    For iCt := 1 To Len(aExp)
      sExp       := aExp[iCt]
      sFuncName  := Alltrim(Upper(Substr(sExp, 1, At("(",sExp) - 1)))
      sFuncParam := Substr(sExp, At(",",sExp))
      sFieldName := aCampos[iCt]
      Do Case
        Case sFuncName == "DTOS"
          sExp := "rqDTOS(@@FIELD" + sFieldName + ")"
        Case sFuncName == "UPPER"
          sExp := "rqUPPER(@@FIELD" + sFieldName + ")"
        Case sFuncName == "SUBSTR"
          sExp := "rqSUBSTR(@@FIELD" + sFieldName + sFuncParam
        Case sFuncName == "PADL"
          sExp := "rqPADL(@@FIELD" + sFieldName + sFuncParam
        Case sFuncName == "PADR"
          sExp := "rqPADR(@@FIELD" + sFieldName + sFuncParam
        Case sFuncName == "STR"
          sExp := "rqSTR(@@FIELD" + sFieldName + sFuncParam
        Case sFuncName == "STRZERO"
          sExp := "rqSTRZERO(@@FIELD" + sFieldName + sFuncParam
        Case sFuncName == ""
          sExp := "@@FIELD" + sFieldName
        Otherwise
          raiseError("Function not supported for Index creation;Function:" ;
                      + sFuncName)
          sExp := "@@FIELD" + sFieldName
      EndCase
      IF ::FieldInfo(DBS_RDDFIELDTYPE,sFieldName) == RDDFIELDTYPE_NORMAL
        aExp[iCt] := sExp
      Else
        aExp[iCt] := ""
      Endif
    Next
    Aadd(aExp, "SkipPad('{{R}}' || @@FIELDRECNO)") // Recno auto added 
  Endif
  // - - - Populate index field
  IF bRet
    sExp := ""
    For iCt := 1 To Len(aExp)
      IF !Empty(aExp[iCt])
        sExp += aExp[iCt] + " || "
      Endif
    Next
    sExp  := Substr(sExp, 1, Rat(" || ",sExp) - 1)
    sStmt := 'UPDATE @@TABELA SET @@FIELDIDX = @@CONTEUDO '
    sStmt := StrTran(sStmt, "@@TABELA",   ::sTableName)
    sStmt := StrTran(sStmt, "@@FIELDIDX", RDDIDX_PREFIX + ind_name)
    sStmt := StrTran(sStmt, "@@CONTEUDO", Strtran(sExp,"@@FIELD"))
    // Test
    sAux := conn:executeUpdate( sStmt + 'WHERE RECNO = 0')
    IF !Empty(sAux)
      raiseError( "Invalid Index expression!; Index creation aborted;" ;
                  + "Expression: " + StrTran(sExp,"@@FIELD"))
      bRet := .F.
    Else
      // Propaga update por toda tabela
      sAux := conn:executeUpdate( sStmt )
      IF !Empty(sAux) 
        raiseError( "Unsuccessful Index update!;Index: " + ind_name )
        bRet := .F.
      Endif
    Endif
  Endif

  // - - - F I R E B I R D  I N D E X E S
  IF bRet 
    // - - Ascendente
    sAux := conn:executeUpdate('CREATE' + IIF(bUnique," UNIQUE "," ") + 'ASC ';
                 + 'INDEX ' + ind_name + ' ON ' + ::sTableName + ' (' ;
                 + RDDIDX_PREFIX + ind_name + ')')
    IF !Empty(sAux)
      raiseError( sAux )
      bRet := .F.
    Endif

    // - - Descendente
    sAux := conn:executeUpdate('CREATE' + IIF(bUnique," UNIQUE "," ") + 'DESC ';
                 + 'INDEX ' + ind_name_desc + ' ON ' + ::sTableName + ' (' ;
                 + RDDIDX_PREFIX + ind_name + ')')
    IF !Empty(sAux)
      raiseError( sAux )
      bRet := .F.
    Endif
  Endif

  // - - - I N D E X  D E S C R I P T I O N
  IF bRet 
    sAux := conn:setIndexDescription( ind_name, expr )  // Asc
    IF !Empty(sAux)
      raiseError( sAux )
      bRet := .F.
    Endif
  Endif

  // - - - T R I G G E R S
  IF bRet 
    // - - Update
    sStmt := ;
       'CREATE TRIGGER @@TNAME FOR @@TABELA ' ;
     + 'ACTIVE BEFORE UPDATE POSITION 0 AS ' ;
     + 'BEGIN ' ;
     + '  IF( OLD.@@FIELDIDX != NEW.@@FIELDIDX OR @@CONDICAO ) THEN ' ;
     + '  BEGIN ';
     + '    NEW.@@FIELDIDX = @@CONTEUDO ;' ;
     + '  END ';
     + 'END '

    sStmt := StrTran(sStmt, "@@TNAME",    sTrigger_updt)
    sStmt := StrTran(sStmt, "@@TABELA",   ::sTableName)
    sStmt := StrTran(sStmt, "@@FIELDIDX", RDDIDX_PREFIX + ind_name)
    sStmt := StrTran(sStmt, "@@CONTEUDO", StrTran(sExp,"@@FIELD","NEW."))
    sAux  := ""
    For iCt := 1 To Len(aCampos)
      IF ::FieldInfo(DBS_RDDFIELDTYPE,aCampos[iCt]) == RDDFIELDTYPE_NORMAL
        sAux += 'OLD.' + aCampos[iCt] + ' != NEW.' + aCampos[iCt] + ' OR ' ;
                + 'OLD.' + aCampos[iCt] + ' IS NULL OR '
      Endif
    Next
    sAux := Substr(sAux, 1, Rat(" OR ",sAux) - 1)
    sStmt := StrTran(sStmt, "@@CONDICAO", sAux)
    sAux := conn:executeUpdate( sStmt )
    IF !Empty(sAux)
      raiseError( "Unknown error on creating Trigger U; FB Error: " + sAux )
      bRet := .F.
    Endif

    // - - Insert
    sStmt := ;
       'CREATE TRIGGER @@TNAME FOR @@TABELA ' ;
     + 'ACTIVE BEFORE INSERT POSITION 1 AS ' ;
     + 'BEGIN ' ;
     + '  NEW.@@FIELDIDX = @@CONTEUDO; ' ;
     + 'END '

    sStmt := StrTran(sStmt, "@@TNAME",    sTrigger_isrt)
    sStmt := StrTran(sStmt, "@@TABELA",   ::sTableName)
    sStmt := StrTran(sStmt, "@@FIELDIDX", RDDIDX_PREFIX + ind_name)
    sStmt := StrTran(sStmt, "@@CONTEUDO", StrTran(sExp,"@@FIELD","NEW."))
    sAux := conn:executeUpdate( sStmt )
    IF !Empty(sAux)
      raiseError( "Unknown error on creating Trigger I; FB Error: " + sAux )
      bRet := .F.
    Endif
  Endif

  IF !bRet
    raiseError( "Error on creating Index!;Index: " + ind_name )
  Endif
  Return bRet

METHOD Seek( expr, soft, p1 ) CLASS rqRDDFire
  Local conn        := ::getConnection() AS rqFBConn
  Local bGlobalSoft := Set(_SET_SOFTSEEK) 
  Local bSoftOn     := IIF(!Empty(soft), soft, bGlobalSoft) 
  Local rs          := {}
  Local sWhere, sOrder
  Local iFldIdx
 
//  IF ::bEof .and. ::bBof
//    Return .F.
//  Endif

  ::bEof := ::bBof := ::bFound := .F.
  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif
  IF ::iActualOrder > 0
    sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
  Else
    sOrder := RECNO_FIELD
  Endif
  iFldIdx := ::FieldPos(sOrder) 
  IF bSoftOn
    sWhere := sOrder + " >= '" + expr + "'"
    rs := ::getSeekCached( sWhere )
    IF Empty(rs)
      rs := conn:getRow( ::sTableName, sWhere, sOrder, "", 0 )
      IIF(!Empty(rs), ::addSeekCache(sWhere,rs))
    Endif
    IF !Empty(rs)
      ::aActualRecord := rs
      ::sNextRecord   := ::aActualRecord[iFldIdx]
      ::bDeleted      := .F.

      IF Substr(::sNextRecord, 1, ;
                Len(::sNextRecord) - RECNO_FIELD_SIZE) == expr
        // - - - Exact match
        IF ::bFiltered
          Do While !::bEof
            IF Eval(::blkFilter)
              IF Substr(::sNextRecord, 1, ;
                        Len(::sNextRecord) - RECNO_FIELD_SIZE) == expr
                ::bFound := .T.
              Else
                ::bFound := .F.
              Endif
              Exit
            Endif
            ::Skip(1)
          Enddo
        Else
          ::bFound := .T.
        Endif
      Else
        // - - - Greater than ...
        ::bFound := .F.
        IF ::bFiltered
          Do While !::bEof
            IF Eval(::blkFilter)
              Exit
            Endif
            ::Skip(1)
          Enddo
        Endif
      Endif
    Else
      ::bFound := .F.
      ::bEof   := .T. 
    Endif
  Else
    sWhere := sOrder + " STARTING '" + expr + "'"
    rs := ::getSeekCached( sWhere )
    IF Empty(rs)
      rs := conn:getRow( ::sTableName, sWhere, sOrder, "", 0 )
      IIF(!Empty(rs), ::addSeekCache(sWhere,rs))
    Endif
    IF !Empty(rs)
      ::aActualRecord := rs
      ::sNextRecord   := ::aActualRecord[iFldIdx]
      ::bDeleted      := .F.

      IF Substr(::sNextRecord, 1, Len(expr)) == expr
        IF ::bFiltered
          IF Eval(::blkFilter)
            ::bFound := .T.
          Else
            ::bFound := .F.
            ::bEof   := .T. 
          Endif
        Else
          ::bFound := .T.
        Endif
      Else
        ::bFound := .F.
        ::bEof   := .T. 
      Endif
    Else
      ::bFound := .F.
      ::bEof   := .T. 
    Endif
  Endif
  Return ::bFound

METHOD Append( p1 ) CLASS rqRDDFire AS Logical
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local sFieldType
  Local sValue
  Local iCt

  IF ::isRecordChanged()
    ::flushRecordBuffer()
  Endif
  IF !conn:existsAnotherUserFLock( ::sTableName, ::sAreaID )
    ::UnLock()
    ::clearRecordBuffer()
    bRet := ::bAppendPending := .T.
    NetErr( .F. )
  
    IF ::bCachingAppends
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[APPND] Append Cached Tbl: " + ::sTableName ;
                  + " Alias: " + ::sAlias )
    Else
      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[APPND] Append Tbl: " + ::sTableName ;
                  + " Alias: " + ::sAlias )
    Endif
  Else
    NetErr( .T. )
  Endif
  Return bRet

METHOD Refresh() CLASS rqRDDFire AS Logical
  Local bRet  := .F.
  Local conn  := ::getConnection() AS rqFBConn
  Local rs    := {}
  Local recno := Alltrim(Str(::getBufferRecno()))
  Local iCt, sWhere, sOrder

  IF Val(recno) > 0
    IF ::iActualOrder > 0
      sOrder := ::aIndexes[::iActualOrder][RDDIDX_FIELDS]
    Else
      sOrder := RECNO_FIELD
    Endif    
    sWhere := RECNO_FIELD + " = '" + recno + "'"
    rs     := conn:getRow( ::sTableName, sWhere, RECNO_FIELD, "", 0 )
    IF !Empty(rs)
      ::aActualRecord := rs
      ::sNextRecord   := ::aActualRecord[::FieldPos(sOrder)]
      ::bDeleted      := .F.

      // - - - Log
      ::writeLog( LOG_LEVEL_08, "[REFRESH] Refresh Rec #" + recno ;
                  + " Table: " + ::sTableName + " Alias: " + ::sAlias )
      bRet := .T.
    Endif
  Endif
  Return bRet

METHOD Commit() CLASS rqRDDFire AS Logical
  Return ::Skip(0)

METHOD UnLock( vl ) CLASS rqRDDFire AS Logical
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local recno
  ::Commit()  
  IF Empty(vl) 
    IF ::bOwnerOfLocks
      conn:unlockAllRecords( ::sTableName, ::sAreaID )
      conn:setTableAsUnLocked( ::sTableName, ::sAreaID )
      ::bOwnerOfLocks := .F.

      // - - - Log
      ::writeLog( LOG_LEVEL_06, "[UNLOCK] Unlock Global. Table: " ;
                  + ::sTableName + " Alias: " + ::sAlias )
    Endif
    bRet := .T.
  Else
    recno := Alltrim(Str(vl))
    IF conn:setRecordAsUnLocked( ::sTableName, ::sAreaID, recno )

      // - - - Log
      ::writeLog( LOG_LEVEL_06, "[UNLOCK] Unlock Record. Table: " ;
                  + ::sTableName + " Alias: " + ::sAlias + " Rec #" + recno )
      bRet := .T.
    Endif
  Endif
  Return bRet

METHOD RLock( vl ) CLASS rqRDDFire AS Logical
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  Local recno
  IF Empty(vl) 
    ::UnLock()
    recno := Alltrim(Str(::getBufferRecno()))
  Else
    recno := Alltrim(Str(vl))
  Endif
  IF Val(recno) > 0
    IF conn:setRecordAsLocked( ::sTableName, ::sAreaID, recno )

      // - - - Log
      ::writeLog( LOG_LEVEL_06, "[RLOCK] Record Locked. Table: " ;
                  + ::sTableName + " Alias: " + ::sAlias + " Rec #" + recno )
      ::Refresh()
      ::bOwnerOfLocks := .T. // Only UnLock() changes this status
      bRet := .T.
    Else

      // - - - Log
      ::writeLog( LOG_LEVEL_04, "[WARN] RLock conflict. Claimer: " + ::sAreaID ;
                  + " Owner: " + conn:getRLockAreaID(::sTableName, recno) ;
                  + " Rec #" + recno )
    Endif
  Else
    bRet := .T.
  Endif
  Return bRet

METHOD FLock() CLASS rqRDDFire AS Logical
  Local bRet := .F.
  Local conn := ::getConnection() AS rqFBConn
  conn:unlockAllRecords( ::sTableName, ::sAreaID )
  IF conn:setTableAsLocked( ::sTableName, ::sAreaID )

    // - - - Log
    ::writeLog( LOG_LEVEL_06, "[FLOCK] Table Locked. Table: " + ::sTableName ;
                + " Alias: " + ::sAlias + " Owner: " + ::sAreaID )
    ::Refresh()
    ::bOwnerOfLocks := .T.  // Only UnLock() changes this status
    bRet := .T.
  Else

    // - - - Log
    ::writeLog( LOG_LEVEL_04, "[WARN] FLock conflict. Claimer: " + ::sAreaID ;
                + " Owner: " + conn:getFLockAreaID(::sTableName) )
  Endif
  Return bRet

METHOD ClearRelation() CLASS rqRDDFire // vd. TODO list
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[RELATION] Cleared. Alias: " + ::sAlias )
  Return .T.

METHOD RelationObject( vl ) CLASS rqRDDFire // vd. TODO list
  Return NIL

METHOD SetFilter( blk, expr ) CLASS rqRDDFire
  Local bRet := .T.
  IF ValType(blk) == VALTYPE_CODEBLOCK
    ::blkFilter := blk
    ::sFilter   := IIF(Empty(expr), "", expr)
    ::bFiltered := .T.

    // - - - Log
    ::writeLog( LOG_LEVEL_06, "[FILTER] Actived. Alias: " + ::sAlias ;
                + " Condition: " + ::sFilter )
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] SetFilter needs a codeblock! Alias: " ;
                + ::sAlias )
    raiseError( "Filter needs a codeblock! " )
  Endif
  Return bRet

METHOD ClearFilter() CLASS rqRDDFire
  ::sFilter   := ""
  ::blkFilter := {|| .T.}
  ::bFiltered := .F.
  // - - - Log
  ::writeLog( LOG_LEVEL_06, "[FILTER] Cleared. Alias: " + ::sAlias )
  Return .T.

METHOD Pack() CLASS rqRDDFire AS Logical
  Return .T.

METHOD OrderInfo( nro_info, p1, p2, p3) CLASS rqRDDFire
  Local uValue
  Do Case
    Case nro_info == DBOI_EXPRESSION
      uValue := ::IndexKey()
    Otherwise
      uValue := NIL
  EndCase
  Return uValue

METHOD Info( nro_info, vl ) CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local uValue
  Do Case
    Case nro_info == DBI_ISFLOCK
      uValue := conn:isTableLockedByUser( ::sTableName, ::sAreaID )
    Case nro_info == DBI_FULLPATH
      uValue := ::sTableName
    Case nro_info == DBI_ALIAS
      uValue := ::sAlias
    Otherwise
      uValue := NIL
  EndCase
  Return uValue

METHOD dropViews() CLASS rqRDDFire AS Logical
  Local conn := ::getConnection() AS rqFBConn
  Return conn:dropViewsFromTable(::sTableName)

METHOD createView( vwname, aflds, afldsalias ) CLASS rqRDDFire AS Logical
  Local conn := ::getConnection() AS rqFBConn
  Local bRet := .F.
  Local aParam := {}
  Local viewname := IIF(Empty(vwname), ::sAlias, Upper(vwname)) 
  Local sFieldAls := ""
  Local sStmt
  Local iCt

  sStmt := ;
      'SELECT @@FIELDS FROM @@TABLENAME'

  IF !conn:isView(viewname)
    IF ValType(aflds) == VALTYPE_ARRAY && ValType(afldsalias) == VALTYPE_ARRAY
      For iCt := 1 To Len(aflds)
        sFieldAls += afldsalias[iCt] + ","
        Aadd(aParam, {"@@FIELDS", aflds[iCt] + ",@@FIELDS"})
      Next
      Aadd(aParam, {"@@TABLENAME", ::sTableName})
      Aadd(aParam, {",@@FIELDS", ""})
      sFieldAls := Substr(sFieldAls, 1, Rat(",",sFieldAls) - 1)
      bRet := conn:createView( viewname, sFieldAls, ;
                               conn:assembleStatement(sStmt,aParam) )
      conn:grantAccess( "SELECT", viewname, "PUBLIC" )

      // - - - Log
      ::writeLog( LOG_LEVEL_06, "[VIEW] Created for Als:" + ::sAlias )
    Else

      raiseError( "Wrong Parameters Type for View. Als:" + ::sAlias )
    Endif
  Else

    // - - - Log
    ::writeLog( LOG_OFF, "[WARN] VIEW already exists: " + viewname)
  Endif
  Return bRet

METHOD CreateDB( tblname, struct, p1 ) CLASS rqRDDFire AS Logical
  Local conn AS rqFBConn
  Local bRet := .F.
  IF ::isConnected()
    conn := ::getConnection()
    bRet := conn:createTable( tblname, struct )
  Endif
  Return bRet

METHOD Delete( p1, p2, p3 ) CLASS rqRDDFire AS Logical
  Local conn := ::getConnection() AS rqFBConn
  IF ::bUsed
    ::bDeleted := conn:deleteRecord( ::sTableName, ;
                                     Alltrim(Str(::getBufferRecno())) )
    ::clearAllCaches()  // ???? Temporary 
  Endif
  Return ::bDeleted

METHOD Zap() CLASS rqRDDFire AS Logical
  Local conn := ::getConnection() AS rqFBConn
  IF ::bUsed
    ::bDeleted := conn:deleteAllRecords( ::sTableName )
    ::clearAllCaches()
  Endif
  Return ::bDeleted

METHOD Eval( blk, evfor, evwhile, p1 ) CLASS rqRDDFire // ???? Not Documented
  Local bRet     := .T.
  Local blkfor   := {}
  Local blkwhile := {}

  IF !Empty(p1)
    // - - - Log
    ::writeLog( LOG_OFF, "[WARN] Support only for the first three params " ;
                + "of DBEval. Tbl:" + ::sTableName + " Als:" + ::sAlias )
  Endif

  // - - - Log
  ::writeLog( LOG_LEVEL_07, "[DBEVAL] Started. " + "Tbl:" + ::sTableName ;
                            + " Als:" + ::sAlias )

  IF ValType(evfor) == VALTYPE_CHARACTER .and. !Empty(evfor)
    blkfor := &( evfor )
  ElseIF ValType(evfor) == VALTYPE_CODEBLOCK
    blkfor := evfor
  Else
    blkfor := {|| .T.}
  Endif 

  IF ValType(evwhile) == VALTYPE_CHARACTER .and. !Empty(evwhile)
    blkwhile := &( evwhile )
  ElseIF ValType(evwhile) == VALTYPE_CODEBLOCK
    blkwhile := evwhile
  Else
    ::GoTop() // whole table
    blkwhile := {|| .T.}
  Endif 

  IF ValType(blk)            == VALTYPE_CODEBLOCK .and. ;
     ValType(Eval(blkfor))   == VALTYPE_LOGICAL   .and. ; 
     ValType(Eval(blkwhile)) == VALTYPE_LOGICAL

    ::clearAllCaches()
    ::Refresh()
    Do While Eval(blkwhile) .and. !::bEof
      IF Eval(blkfor)
        Eval(blk)
      Endif
      ::Skip(1)
    Enddo
  Else
    // - - - Log
    ::writeLog( LOG_OFF, "[ERROR] DBEval parameters error! Als:" + ::sAlias )

    raiseError( "DBEval parameters error! Als:" + ::sAlias )
    bRet := .F.
  Endif

  // - - - Log
  ::writeLog( LOG_LEVEL_07, "[DBEVAL] Ended. " + "Tbl:" + ::sTableName ;
                            + " Als:" + ::sAlias )
  Return bRet

PROTECT METHOD reccountTable() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  IF ::bUsed 
    IF (Abs(Seconds() - ::iRecCountAge)) <= ::getRecCountTimeOut() 
      // Still valid
    Else
      //::iRecCount    := conn:getTableRecCount( ::sTableName )
      ::iRecCount    := conn:getTableNumberOfInserts( ::sTableName )
      ::iRecCountAge := Seconds()

      // - - - Log
      ::writeLog( LOG_LEVEL_07, "[COUNT] RecCount. Tbl: " + ::sTableName ;
                                + " Als: " + ::sAlias + " CT: " ;
                                + Alltrim(Str(::iRecCount)) )
    Endif
  Endif
  Return ::iRecCount

PROTECT METHOD closeRDD() CLASS rqRDDFire
  Local conn := ::getConnection() AS rqFBConn
  Local iCt

  IF ::bUsed
    IF ::isRecordChanged()
      ::flushRecordBuffer()
    Endif
    IF ::isConnected()
      ::UnLock()
      conn:removeAlias( ::sAlias )
      ::setTableUsedbyThisArea( .F. )
    Endif
    ::bUsed := .F.
    ::bConn := .F.

    // - - - Log
    ::writeLog( LOG_LEVEL_07, "[CLOS] Table closed. Table: " + ::sTableName ;
                + " Alias: " + ::sAlias )

    IF ::isTemporaryTable()
      ::commitSuicide()
    Endif
    For iCt := 1 to Len(::aStru)
      IF ::aStru[iCt,DBS_RDDFIELDTYPE] == RDDFIELDTYPE_NORMAL
        RDDretField( SELF, ::aStru[iCt,DBS_NAME] )
      Endif
    Next
    RDDretAlias( SELF )
  Endif
  Return .T.

Static Function raiseError( msg )
  Local oErr := Error{} AS Error
  oErr:DESCRIPTION := "[RDD Error] ;" + msg
  oErr:GENCODE     := -1
  oErr:SEVERITY    := ES_RTERR
  Eval(ErrorBlock(),oErr)
  Return oErr

Static Function wrtLog( level, msg, actuallevel )
  Local iAux := actuallevel
#Cinline
{
  int level    = VAR_INT(VAR_NAME_LOCPAR(level));
  char *msg    = VAR_CHR(VAR_NAME_LOCPAR(msg));
  int loglevel = VAR_INT(VAR_NAME_LOCAL(iaux));

  if( level <= loglevel )
  {
    FILE *fp;
    long time(), now;
    char errfile[100];
    time( &now );
    sprintf( errfile, "/var/log/rqrddfire.log" );
    fp = fopen( errfile, "a" );
    if( fp != NULL )
    {
      fprintf( fp, "\t[%i] %s ", level, getenv("USER") );
      fprintf( fp, "%s ", ctime(&now) );
      fprintf( fp, "%s\n", msg );
      fclose( fp );
    }
  }
}
#endCinline
  Return

