/*
RqRDDFire - A Free FlagShip RDD for Firebird RDBMS
v.0.9.0, September 2007, Brazil
Ricardo Delamar Roque (rroque6428@terra.com.br)

License
=======
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed AS IS in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

http://www.gnu.org/licenses/lgpl-2.1.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <ibase.h>
#include <ib_util.h>
#include "example.h"

#define MAXLEN                  1024
#define LOG_OFF                 0
#define LOG_LEVEL_01            1
#define LOG_LEVEL_02            2
#define LOG_LEVEL_03            3
#define LOG_LEVEL_04            4
#define LOG_LEVEL_05            5
#define LOG_LEVEL_06            6
#define LOG_LEVEL_07            7
#define LOG_LEVEL_08            8
#define LOG_LEVEL_09            9
#define LOG_LEVEL_10            10

static void rddlog( int level, char *string );
char        *replicate( char *buff, char *s, int vezes );
char        *rtrim( char *buff, char *s );
char        *padl( char *buff, char *s, int tam, char *f );
char        *padr( char *buff, char *s, int tam, char *f );
char        *parse_date( char *buff, char *s );
char        *parse_recno( char *buff, char *s );
char        *parse_skip_token( char *buff, char *s );
char        *substr( char *buff, char *s, short m, short n);
int         at( char *p, char *s );
char        *upper( char *buff, char *s );
char        *lower( char *buff, char *s );
char        *descend( char *buff, char *s );
char        *strzero( char *buff, char *s, int t, int d );
char        *str( char *buff, char *s, int t, int d );
char        *skippad( char *buff, char *s );

static int  	_rqfbconn_loglevel_ = LOG_OFF;
static char 	_rqfbconn_event_[4000];

typedef struct blob {
  short   (*blob_get_segment) ();
  void    *blob_handle;
  long    blob_number_segments;
  long    blob_max_segment;
  long    blob_total_length;
  void    (*blob_put_segment) ();
} *BLOB;

static void rddlog( int level, char *string )
{
  if( level <= _rqfbconn_loglevel_ )
  {
    FILE *fp;
    long time(), now;
    char errfile[100];
    time( &now );
    sprintf( errfile, "/var/log/rqrddfire.log" );
    fp = fopen( errfile, "a" );
    if( fp != NULL )
    {
      fprintf( fp, "\t%s ", getenv("USER") );
      fprintf( fp, "%s ", ctime(&now) );
      fprintf( fp, "%s\n", string );
      fclose( fp );
    }
  }
  return;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Replicate
char *replicate( char *buff, char *s, int v )
{
  buff[0] = '\0';
  if( (strlen(s)*v) < 256 ) while( v-- ) strcat( buff, s );
  else strcpy( buff, "Replicate Overflow Max.256" );
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Rtrim
char *rtrim( char *buff, char *s )
{
  char *r = buff;

  buff[0] = '\0';
  strcpy( buff, s );
  if( *r )
  {
    r += strlen(buff) - 1;
    while( r >= buff && *r == ' ' ) r--;
  }  
  *++r = '\0' ;
  return buff;
} 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Padl
char *padl( char *buff, char *s, int tam, char *f )
{
  char lbuff1[MAXLEN];
  char *p  = buff; 
  
  buff[0] = '\0';
  strcpy( buff, replicate(lbuff1, f, tam) );
  strcat( buff, rtrim(lbuff1, s) );
  p += strlen(buff) - tam;
  return p;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Padr
char *padr( char *buff, char *s, int tam, char *f )
{
  char lbuff1[MAXLEN];
 
  buff[0] = '\0';
  if( !*f ) *f = " ";
  if( tam > 0 )
  {
    strcpy( buff, rtrim(lbuff1, s) );
    strcat( buff, replicate(lbuff1, f, tam) );
    strcpy( buff, substr(lbuff1, buff, 1,tam) );
  }
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Parse_Date
char *parse_date( char *buff, char *s )
{
  struct tm *t = (struct tm *)malloc(sizeof(struct tm));
  char      *p;
 
  buff[0] = '\0';
  memset( t, '\0', sizeof(*t) );
  //p = strptime( s, "%d-%B-%C%y", t ); // FB1.0.3
  p = strptime( s, "%C%y-%m-%d", t ); // FB2.0.3
  if( p != NULL )
  {
    strftime( buff, MAXLEN, "%C%y%m%d", t );
  }
  free( t );
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - parse_recno
char *parse_recno( char *buff, char *s )
{
  char lbuff1[MAXLEN];

  buff[0] = '\0';
  strcpy( buff, s );
  buff[at(".",s) - 1] = '\0';
  strcpy( buff, padl(lbuff1, buff, 15, "0") );
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Parse_Skip_Token
char *parse_skip_token( char *buff, char *s )
{
  char lbuff1[MAXLEN];
  char *p;

  buff[0] = '\0';
  strcpy( buff, s );
  p = strtok( buff, "}}" );
  if( p )
  {
    switch( *p )
    {
      case 'R': 
                p = strtok( NULL, "}}" );
                p = parse_recno(lbuff1, p);
        	break;
      case 'D': 
                p = strtok( NULL, "}}" );
                p = parse_date(lbuff1, p);
        	break;
      case 'C': 
                p = strtok( NULL, "}}" );
        	break;
      case 'N': 
                p = strtok( NULL, "}}" );
                p = padl( lbuff1, p, strlen(p), "0" );
        	break;
      default:
        	p = s;
    }
  }
  return p; 
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Substr
char *substr( char *buff, char *s, short m, short n )
{
  short i = 0;

  if( *s && m >= 1 )
  {
    if( n < 1 ) n = 0; 
    s += m - 1;
    while( *s && (i < n || n == 0) ) buff[i++] = *s++;
  }
  buff[i] = '\0';

  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - At
int at( char *p, char *s )
{
  int z = strlen(s);
  int a = 0;
  char *e;
  
  e = strstr(s, p);
  if( e )
  {
    a = strlen(e);
    if( a > 0 ) a = z - a + 1;
    if( a > z ) a = 0;
  }
  return a;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - upper
char *upper( char *buff, char *s )
{
  char *r = buff;

  while( *s )
  {
    if( *s >= 'a' && *s <= 'z' )
    {
      *r++ = *s++ + 'A' - 'a'; 	     
    }
    else *r++ = *s++;
  }
  *r = '\0';
  
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - lower
char *lower( char *buff, char *s )
{
  char *r = buff;

  while( *s )
  {
    if( *s >= 'A' && *s <= 'Z' )
    {
      *r++ = *s++ - 'A' + 'a';
    }
    else *r++ = *s++;
  }
  *r = '\0';

  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - descend
char *descend( char *buff, char *s )
{
  char *r = buff;
 
  while( *s )
  {
    *r++ = 256 - (unsigned char) *s++;
  }
  *r = '\0';
  
  return buff;
}
  
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - strzero
char *strzero( char *buff, char *s, int t, int d )
{
  char lbuff1[MAXLEN];
  char lbuff2[MAXLEN];
  int  ptpos;
  
  buff[0] = '\0';
  ptpos = at(".", s);
  if( ptpos == 1 ) strcpy( buff, "0" );
  else strcpy( buff, rtrim(lbuff2, substr(lbuff1, s, 1, ptpos - 1)) );
  if( d ) strcat( buff, "." );
  if( d && ptpos) ptpos++;
  strcat( buff, padr(lbuff2, substr(lbuff1, s, ptpos, 0), d, "0") );
  strcpy( buff, padl(lbuff1, buff, t, "0") );

  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - str
char *str( char *buff, char *s, int t, int d )
{
  char lbuff1[MAXLEN];
  char lbuff2[MAXLEN];
  int  ptpos;
  
  buff[0] = '\0';
  ptpos = at(".", s);
  if( ptpos == 1 ) strcpy( buff, " " );
  else strcpy( buff, rtrim(lbuff2, substr(lbuff1, s, 1, ptpos - 1)) );
  if( d ) strcat( buff, "." );
  if( d && ptpos) ptpos++;
  strcat( buff, padr(lbuff2, substr(lbuff1, s, ptpos, 0), d, "0") );
  strcpy( buff, padl(lbuff1, buff, t, " ") );
	      
  return buff;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - SkipPad
char *skippad( char *buff, char *s )
{
  char lbuff1[MAXLEN];
  char token[256];
  char resto[256];
  char *r = resto;
  char *p;

  strcpy( token, s);
  strcpy( resto, s);

  buff[0] = '\0';
  if( at("{{",token) ) do
                       {
                         p = strtok( token, "{{" );
                         if( p )
                         {
			   if( at("}}",p) ) strcat(buff, parse_skip_token(lbuff1, p));
                           r += strlen( p ) + 2;
                           strcpy( token, r );
                         }
                       }
                       while( p );
  return buff;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_padl(s, tam, f) - Returns the padl value of 's' filled with 'f'. 

 DECLARE EXTERNAL FUNCTION rqpadl
   VARCHAR(256), SMALLINT, VARCHAR(1)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_padl' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_padl(char* s, short* tam, char* f)
/* VARCHAR(256) string original   
   SMALLINT     tamanho   
   VARCHAR(1)   caracter de preenchimento */
{
  //char *buffer = (char *)malloc(256); //FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  char *aux;

  s += 2;
  f += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i tam f=%i :", strlen(s), strlen(f) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 && strlen(f) > 0 )
  { 
	  
    aux = padl( lbuff1, s, *tam, f );
//    buffer[0] = (unsigned short) strlen(aux) + 2;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
    strcat( buffer, aux );
  }
  else
  {
//    buffer[0] = (unsigned short)3;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
  } 
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Padl (%s) with (%s) resulting in (%s) "
	    "which is %i characters long ", s, f, buffer, *tam );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_padr(s, tam, f) - Returns the padr value of 's' filled with 'f'. 

 DECLARE EXTERNAL FUNCTION rqpadr
   VARCHAR(256), SMALLINT, VARCHAR(1)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_padr' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_padr(char* s, short* tam, char* f)
/* VARCHAR(256) string original   
   SMALLINT     tamanho   
   VARCHAR(1)   caracter de preenchimento */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  char *aux;

  s += 2;
  f += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i tam f=%i :", strlen(s), strlen(f) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 && strlen(f) > 0 )
  { 
	  
    aux = padr( lbuff1, s, *tam, f );
//    buffer[0] = (unsigned short) strlen(aux) + 2;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
    strcat( buffer, aux );
  }
  else
  {
//    buffer[0] = (unsigned short)3;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
  } 
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Padr (%s) with (%s) resulting in (%s) "
	    "which is %i characters long ", s, f, buffer, *tam );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_strzero(s, tam, d) - Returns 's' filled with left zeroes

 DECLARE EXTERNAL FUNCTION rqstrzero
   VARCHAR(256), SMALLINT, SMALLINT
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_strzero' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_strzero(char* s, short* tam, short* d)
/* VARCHAR(256) string original   
   SMALLINT     tamanho   
   SMALLINT     casas decimais */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];

  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, strzero(lbuff1, s, *tam, *d) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Strzero (%s) resulting in (%s) "
	    "which is %i characters long ", s, buffer, *tam );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_str(s, tam, d) - Returns 's' filled with left blanks

 DECLARE EXTERNAL FUNCTION rqstr
   VARCHAR(256), SMALLINT, SMALLINT
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_str' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_str(char* s, short* tam, short* d)
/* VARCHAR(256) string original   
   SMALLINT     tamanho   
   SMALLINT     casas decimais */
{
  // char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];

  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, str(lbuff1, s, *tam, *d) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Str (%s) resulting in (%s) "
	    "which is %i characters long ", s, buffer, *tam );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_skippad(s) - Returns 's' in a form that is useful to Skip FS Command

 DECLARE EXTERNAL FUNCTION skippad
   VARCHAR(256)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_skippad' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_skippad(char* s)
/* VARCHAR(256) string */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  char *aux;
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 )
  { 
    aux = skippad( lbuff1, s );
//    buffer[0] = (unsigned short) strlen(aux) + 2;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
    strcat( buffer, aux );
  }
  else
  {
//    buffer[0] = (unsigned short)3;
//    buffer[1] = ' ';
//    buffer[2] = '\0';
  } 
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] SkipPad (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_dtos(s) - Returns 's' as YYYYMMDD date

 DECLARE EXTERNAL FUNCTION rqdtos
   VARCHAR(30)
   RETURNS CSTRING(8) FREE_IT
   ENTRY_POINT 'rq_udf_dtos' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_dtos(char* s)
/* VARCHAR(30) string */
{
  //char *buffer = (char *)malloc(9); // FB203
  char *buffer = (void *)ib_util_malloc(9); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, parse_date(lbuff1, s) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] DtoS (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(8) FREE IT*/
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_substr(s,f,n) - Returns substring of 's' from 'f' plus 'n' characters 

 DECLARE EXTERNAL FUNCTION rqsubstr
   VARCHAR(256), SMALLINT, SMALLINT
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_substr' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_substr(char* s, short* f, short*n)
/* VARCHAR(256) string   
   SMALLINT     from   
   SMALLINT     number of chars */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, substr(lbuff1, s, *f, *n) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Substr (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT */
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_upper(s) - Returns upper of 's'  

 DECLARE EXTERNAL FUNCTION rqupper
   VARCHAR(256)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_upper' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_upper(char* s)
/* VARCHAR(256) string */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, upper(lbuff1, s) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Upper (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT */
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_lower(s) - Returns lower of 's'  

 DECLARE EXTERNAL FUNCTION rqlower
   VARCHAR(256)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_lower' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_lower(char* s)
/* VARCHAR(256) string */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, lower(lbuff1, s) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Lower (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT */
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_rtrim(s) - Remove trailing spaces of 's'  

 DECLARE EXTERNAL FUNCTION rqtrim
   VARCHAR(256)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_rtrim' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_rtrim(char* s)
/* VARCHAR(256) string */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, rtrim(lbuff1, s) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] RTrim (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT */
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by FB Team
 * Funcao original de udflib.c fornecida pelo pacote de inst do FB
 * 
 rq_udf_blobsubstr(b,f,n) - Returns substring of 'b' from 'f' to 'n' 

 DECLARE EXTERNAL FUNCTION rqblobsubstr
   BLOB, INTEGER, INTEGER
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_blobsubstr' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_blobsubstr(BLOB b, long* m, long* n)
{
	char *buf, *p, *q;
	long i = 0;
	long curr_bytecount = 0;
	long begin, end; 
	short length, actual_length;
 
	//char *buffer = (char *)malloc(256); // FB203
	char *buffer = (void *)ib_util_malloc(256); // FB251

	buffer[0] = '\0';

	if (!b->blob_handle)
		return buffer;
	if (*m > *n || *m < 1L || *n < 1L) 
		return buffer;
	if (b->blob_total_length < (long)*m) 
		return buffer;

	length = b->blob_max_segment + 1L;
	buf = (char *) malloc (length); 

	begin = *m;				/* beginning position */

	if (b->blob_total_length < (long)*n) 
		end = b->blob_total_length;	/* ending position */
	else
		end = *n;

	/* Limit the return string to 255 bytes */
	if (end - begin + 1L > 255L)
		end = begin + 254L;

	q = buffer;

    while ((*b->blob_get_segment) (b->blob_handle, buf, length, &actual_length))
		{    
		buf [actual_length] = 0;

		p = buf;
		while (*p && (curr_bytecount <= end))
			{
			curr_bytecount++;
			if (*p == '\n')
				*p = ' ';
			if (curr_bytecount >= begin)
				*q++ = *p;
			p++;
			}
		if (curr_bytecount >= end) 
			{
			*q = 0;
			break;
			}
		}
 
	free (buf);
	return buffer;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - by Roque
 rq_udf_descend(s) - Returns ASCII complement of 's'  

 DECLARE EXTERNAL FUNCTION rqdescend
   VARCHAR(256)
   RETURNS CSTRING(256) FREE_IT
   ENTRY_POINT 'rq_udf_descend' MODULE_NAME 'rqrddudfs';
                                                                        */
char* EXPORT rq_udf_descend(char* s)
/* VARCHAR(256) string */
{
  //char *buffer = (char *)malloc(256); // FB203
  char *buffer = (void *)ib_util_malloc(256); // FB251
  char lbuff1[MAXLEN];
  
  s += 2;

  // - - - Log
  sprintf( _rqfbconn_event_, ": tam s=%i :", strlen(s) );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  buffer[0] = '\0';
  if( strlen(s) > 0 ) strcat( buffer, descend(lbuff1, s) );
  
  // - - - Log
  sprintf( _rqfbconn_event_, "[UDF] Descend (%s) resulting in (%s) ", 
		             s, buffer );
  rddlog( LOG_LEVEL_09,_rqfbconn_event_ );

  return buffer; /* CSTRING(256) FREE IT */
}

int main( int argc, char *argv[] )
{	
  char lbuff1[MAXLEN];

  fprintf( stderr, "Rtrim Teste:%s:\n", rtrim(lbuff1, "Teste   ") );
  fprintf( stderr, "Rtrim Teste:%s:\n", rtrim(lbuff1, " Teste") );
  fprintf( stderr, "Rtrim Teste:%s:\n", rtrim(lbuff1, " ") );
  fprintf( stderr, "Rtrim Teste:%s:\n", rtrim(lbuff1, "") );
  fprintf( stderr, "Replicate Teste:%s:\n", replicate(lbuff1, ".",30) );
  fprintf( stderr, "Padl Teste:%s:\n", padl(lbuff1, "GI LAGO SULddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd                                       ",30,".") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "{{D}}2003-09-12 {{C}}GI 302 SUL  {{N}}200.00  {{R}}35.000000000000000      ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "{{D}}14-MAR-2003{{C}}GI 302 SUL  {{N}}20.00   {{R}}5.0000000000000000 ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "{{D}}1-JAN-2003 {{C}}GI 302 SUL  {{N}}2.00    {{R}}235.00000000000000 ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "{{D}}1-JAN-2003 ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "GI SAO SEB37.00000000000000     {892.14}                ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "  ") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, " {{ola-}}") );
  fprintf( stderr, "SkipPad Teste:%s:\n", skippad(lbuff1, "{{     ") );
  // FB1.0.3 format
  fprintf( stderr, "ParseDate Test:%s:\n", parse_date(lbuff1, "1-APR-2002 ddsds ") );
  // FB2.0.3 format
  fprintf( stderr, "ParseDate Test:%s:\n", parse_date(lbuff1, "2006-08-28 00:00:00.0000") );
  fprintf( stderr, "Substr Test:%s:\n", substr(lbuff1, "Testando ",3,10) );
  fprintf( stderr, "Substr Test:%s:\n", substr(lbuff1, "",3,10) );
  fprintf( stderr, "Substr Test:%s:\n", substr(lbuff1, "Testando ",3,1) );
  fprintf( stderr, "Substr Test:%s:\n", substr(lbuff1, "Testando ",3,0) );
  fprintf( stderr, "Upper Test:%s:\n", upper(lbuff1, "abcdefghijklmnopqrstuvxzyw") );
  fprintf( stderr, "Lower Test:%s:\n", lower(lbuff1, "ABCDEFGHIJKLMNOPQRSTUVXZYW") );
  fprintf( stderr, "Padr Test:%s:\n", padr(lbuff1, "",2,"0") );
  fprintf( stderr, "Padr Test:%s:\n", padr(lbuff1, " ",2,"0") );
  fprintf( stderr, "Padr Test:%s:\n", padr(lbuff1, "2",2,"0") );
  fprintf( stderr, "Padr Test:%s:\n", padr(lbuff1, "22",2,"0") );
  fprintf( stderr, "Padr Test:%s:\n", padr(lbuff1, "234",2,"0") );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, "",10,0) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, " ",10,0) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, "234",10,2) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, "178.2",10,4) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, "178.8",10,0) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, "234.0000000     ",10,0) );
  fprintf( stderr, "StrZero Test:%s:\n", strzero(lbuff1, ".8",5,1) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, "",10,0) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, " ",10,0) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, "234",10,2) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, "178.2",10,4) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, "178.8",10,0) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, "234.0000000     ",10,0) );
  fprintf( stderr, "Str     Test:%s:\n", str(lbuff1, ".8",5,1) );
  fprintf( stderr, "Descend Test:%s:\n", descend(lbuff1, "20040504") );
  fprintf( stderr, "Descend Test:%s:\n", descend(lbuff1, "TESTANDO") );
  exit( 0 );
}
